/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package distributions;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Collections;
import java.util.HashMap;
import java.util.regex.Matcher;

import hyleak.Interval;
import hyleak.HyleakPatterns;
import support.Contexter;
import support.Logger;

/**
 *
 * @author  
 */
public class MapDistribution extends HashMap<BigInteger,BigDecimal> implements ProbabilityDistribution{
    private static final Contexter contexter = Contexter.getInstance();
    private static final Logger logger = Logger.getInstance();
        
    private final BigInteger complexityLimit = new BigInteger("10000");
    
    public MapDistribution(String expr) {
        log("The regular expression for a distribution is "+HyleakPatterns.pattern_constant_distribution.pattern(),4);
        Matcher matcher = HyleakPatterns.pattern_constant_distribution.matcher(expr);
        if (matcher.matches()){
            log("Creating distribution from description: " + matcher.group(1),2);
        }else{
            log("ERROR: string "+expr+" does not match regular expression.",1);
            System.exit(1);
        }
        String internalstring = matcher.group(1);
        
        
        
        while (internalstring.contains(";")){
            int semicolonindex = internalstring.indexOf(';');
            String element = internalstring.substring(0,semicolonindex);
            internalstring = internalstring.substring(semicolonindex+1);
            log("element: "+element+", rest = "+internalstring,2);
            BigInteger value = new BigInteger(element.substring(0,element.indexOf('-')));
            BigDecimal prob = new BigDecimal(element.substring(element.indexOf('>')+1));
            this.put(value, prob);
        }
        
        log("Created the distribution "+this.toString()+"\n",4);
    }
    
    public MapDistribution(Interval interval) {
        BigInteger size = interval.getSize();
        log("Creating distribution for interval "+interval.toString(),4 );
        BigDecimal eachprob =  contexter.getDecimalOne().divide(new BigDecimal(size), contexter.getContext());
        for (BigInteger i = interval.getBottom(); i.compareTo(interval.getTop())<=0; i=i.add(BigInteger.ONE)){
            this.put(i, eachprob);
            
        }
        log("Created the distribution "+this.toString()+"\n",4);
    }
    
    public MapDistribution(BigInteger value) {
        
        this.put(value, BigDecimal.ONE);
        log("Created the distribution "+this.toString()+"\n",4);
   
    }
    
   
    
    public MapDistribution(BigInteger lowbound, BigInteger uppbound) {
        this(new Interval(lowbound,uppbound));
        
    }
    
    public MapDistribution(HashMap<BigInteger,BigDecimal> map) {
        this.putAll(map);
        
    }
    
    public MapDistribution(BigInteger[] values, BigDecimal[] probs) {//TODO: rifarlo con min, max e probs?
        if (values.length!=probs.length){
            log("ERROR: creating a distribution with "+values.length+" values and "+probs.length+" probabilities?",1);
            System.exit(1);
        }
        for (int i=0; i<values.length; i++){
            this.put(values[i], probs[i]);
        }
        //log("Internal is "+internal[0].toString(),4);
    }

    public BigInteger getMinimum() {
        return (BigInteger)Collections.min(this.keySet());
    }
    
    public BigInteger getMaximum() {
        return (BigInteger)Collections.max(this.keySet());
    }
   

    public BigInteger getSize() {
        return new BigInteger(""+(this.keySet().size()));
    }
    
    @Override
    public String toString(){
        String toreturn = "{";
        for (BigInteger i : this.keySet()){
            toreturn += i.toString()+"->"+this.get(i).toString()+";";
        }
        return toreturn+"}";
    }

    public boolean isDirac(){
        return this.getSize().compareTo(BigInteger.ONE)==0;
    }

        
    public BigDecimal getProbabilityOf(BigInteger index){
        return this.get(index);
    }
    
    public BigDecimal getProbabilityOfFalse(){
        return this.getProbabilityOf(BigInteger.ZERO);
    }
    
    public BigDecimal getProbabilityOfTrue(){
        return contexter.getDecimalOne().subtract(this.getProbabilityOfFalse(),contexter.getContext());
    }
    
    private static void log(Object aObject, int level){
        logger.log(String.valueOf(aObject),level);
    }

    public MapDistribution add(MapDistribution val2) {
        MapDistribution toreturn;
        if (estimateBinaryComplexity(val2).compareTo(this.complexityLimit)>0){
            toreturn = addStatistically(val2);
        }else{
//            if (this.isDirac()){
//                if (val2.isDirac()){
//                    toreturn = new HybridDistribution(this.getMinimum().add(val2.getMinimum()));
//
//                }else{
//                    toreturn = val2.addConstant(this.getMinimum());
//                }
//            }else{
//                if (val2.isDirac()){
//                    toreturn = this.addConstant(val2.getMinimum());
//                }else{
                    // warning: this is a naive implementation and may take a lot of time to terminate. avoid if possible.
                    HashMap map = new HashMap();
                    for (BigInteger i : this.keySet()){
                        for (BigInteger j : val2.keySet()){
                            BigInteger key = i.add(j);
                            BigDecimal prob = this.get(i).multiply(val2.get(j), contexter.getContext());
                            if (map.containsKey(key)){
                                map.put(key, prob.add((BigDecimal)map.get(key), contexter.getContext()));
                            }else{
                                map.put(key,prob);
                            }
                            
                        }
                    }
                    toreturn = new MapDistribution(map);
                }
//            }
//            
//        }
        return toreturn;
    }

    public MapDistribution subtract(MapDistribution val2) {
        MapDistribution toreturn;
        if (estimateBinaryComplexity(val2).compareTo(this.complexityLimit)>0){
            toreturn = subtractStatistically(val2);
        }else{
            HashMap map = new HashMap();
            for (BigInteger i : this.keySet()){
                for (BigInteger j : val2.keySet()){
                    BigInteger key = i.subtract(j);
                    BigDecimal prob = this.get(i).multiply(val2.get(j), contexter.getContext());
                    if (map.containsKey(key)){
                        map.put(key, prob.add((BigDecimal)map.get(key), contexter.getContext()));
                    }else{
                        map.put(key,prob);
                    }

                }
            }
            toreturn = new MapDistribution(map);
        }
        return toreturn;
    }

    public MapDistribution mod(MapDistribution val2) {
        MapDistribution toreturn;
        if (estimateBinaryComplexity(val2).compareTo(this.complexityLimit)>0){
            toreturn = modStatistically(val2);
        }else{
            HashMap map = new HashMap();
            for (BigInteger i : this.keySet()){
                for (BigInteger j : val2.keySet()){
                    BigInteger key = i.mod(j);
                    BigDecimal prob = this.get(i).multiply(val2.get(j), contexter.getContext());
                    if (map.containsKey(key)){
                        map.put(key, prob.add((BigDecimal)map.get(key), contexter.getContext()));
                    }else{
                        map.put(key,prob);
                    }

                }
            }
            toreturn = new MapDistribution(map);
        }
        return toreturn;
    }

    public MapDistribution xor(MapDistribution val2) {
        BigDecimal mefalse = this.getProbabilityOfFalse();
        BigDecimal hefalse = val2.getProbabilityOfFalse();
        BigDecimal metrue = this.getProbabilityOfTrue();
        BigDecimal hetrue = val2.getProbabilityOfTrue();
        BigInteger[] values = new BigInteger[2];
        values[0] = BigInteger.ZERO;
        values[1] = BigInteger.ONE;
               
        BigDecimal proboffalse = (mefalse.multiply(hetrue, contexter.getContext())).add(metrue.multiply(hefalse, contexter.getContext()), contexter.getContext());
        BigDecimal[] probs = new BigDecimal[2];
        probs[0] = proboffalse;
        probs[1] = contexter.getDecimalOne().subtract(proboffalse, contexter.getContext());
        return new MapDistribution(values,probs);
    }

    public MapDistribution divide(MapDistribution val2) {
        MapDistribution toreturn;
        if (estimateBinaryComplexity(val2).compareTo(this.complexityLimit)>0){
            toreturn = divideStatistically(val2);
        }else{
            HashMap map = new HashMap();
            for (BigInteger i : this.keySet()){
                for (BigInteger j : val2.keySet()){
                    BigInteger key = i.divide(j);
                    BigDecimal prob = this.get(i).multiply(val2.get(j), contexter.getContext());
                    if (map.containsKey(key)){
                        map.put(key, prob.add((BigDecimal)map.get(key), contexter.getContext()));
                    }else{
                        map.put(key,prob);
                    }

                }
            }
            toreturn = new MapDistribution(map);
        }
        return toreturn;
    }

    public MapDistribution multiply(MapDistribution val2) {
        MapDistribution toreturn;
        if (estimateBinaryComplexity(val2).compareTo(this.complexityLimit)>0){
            toreturn = multiplyStatistically(val2);
        }else{
            HashMap map = new HashMap();
            for (BigInteger i : this.keySet()){
                for (BigInteger j : val2.keySet()){
                    BigInteger key = i.multiply(j);
                    BigDecimal prob = this.get(i).multiply(val2.get(j), contexter.getContext());
                    if (map.containsKey(key)){
                        map.put(key, prob.add((BigDecimal)map.get(key), contexter.getContext()));
                    }else{
                        map.put(key,prob);
                    }

                }
            }
            toreturn = new MapDistribution(map);
        }
        return toreturn;
    }

    public MapDistribution and(MapDistribution val2) {
        BigDecimal metrue = this.getProbabilityOfTrue();
        BigDecimal hetrue = val2.getProbabilityOfTrue();
        BigInteger[] values = new BigInteger[2];
        values[0] = BigInteger.ZERO;
        values[1] = BigInteger.ONE;
               
        BigDecimal proboftrue = metrue.multiply(hetrue, contexter.getContext());
        
        BigDecimal[] probs = new BigDecimal[2];
        probs[0] = contexter.getDecimalOne().subtract(proboftrue, contexter.getContext());
        probs[1] = proboftrue;
        return new MapDistribution(values,probs);
    }

    public MapDistribution or(MapDistribution val2) {
        BigDecimal mefalse = this.getProbabilityOfFalse();
        BigDecimal hefalse = val2.getProbabilityOfFalse();
        BigInteger[] values = new BigInteger[2];
        values[0] = BigInteger.ZERO;
        values[1] = BigInteger.ONE;
               
        BigDecimal proboffalse = (mefalse.multiply(hefalse, contexter.getContext()));
        BigDecimal[] probs = new BigDecimal[2];
        probs[0] = proboffalse;
        probs[1] = contexter.getDecimalOne().subtract(proboffalse, contexter.getContext());
        return new MapDistribution(values,probs);
    }

    public MapDistribution negate() {
        BigDecimal metrue = this.getProbabilityOfTrue();
        BigInteger[] values = new BigInteger[2];
        values[0] = BigInteger.ZERO;
        values[1] = BigInteger.ONE;
               
        BigDecimal proboffalse = metrue;
        BigDecimal[] probs = new BigDecimal[2];
        probs[0] = proboffalse;
        probs[1] = contexter.getDecimalOne().subtract(proboffalse, contexter.getContext());
        return new MapDistribution(values,probs);
    }

    
    
    

    @Override
    public BigDecimal getProbability(String key) {
        return this.get(new BigInteger(key));
    }

    @Override
    public BigDecimal getEntropy() {
        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
    }

    @Override
    public String printNice() {
        return this.toString();
    }

    private BigInteger estimateBinaryComplexity(MapDistribution val2) {
        return this.getSize().multiply(val2.getSize());
    }

    private MapDistribution addStatistically(MapDistribution val2) {
        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
    }

    private MapDistribution subtractStatistically(MapDistribution val2) {
        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
    }

    private MapDistribution modStatistically(MapDistribution val2) {
        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
    }

    private MapDistribution divideStatistically(MapDistribution val2) {
        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
    }

    private MapDistribution multiplyStatistically(MapDistribution val2) {
        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
    }
    
    @Override
    public MapDistribution clone(){
        return new MapDistribution(this);
    }
}
