/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package hyleak;

import java.util.ArrayList;
import support.Edge;
import support.EdgeList;
import support.Node;
import support.NodeList;

/**
 *
 * @author  
 */
class CFGExtractor {

    ArrayList<String> codelines;

    public CFGExtractor(ArrayList<String> code) {
        codelines = code;
    }

    public ControlFlowGraph getCFG() {
        NodeList nodes = new NodeList();
        EdgeList edges = new EdgeList();

        for (int i = 1; i < codelines.size(); i++) {
            Node n = new Node("" + i);
            String codeline = codelines.get(i);
            n.setCodeline(codeline.trim());
            nodes.add(n);

            if (codeline.startsWith("goto")) {
                int lastspace = codeline.lastIndexOf(" ");
                int lastsemicolumn = codeline.lastIndexOf(";");
                String target = codeline.substring(lastspace, lastsemicolumn);
                Edge e = new Edge("" + (i), target);
                //edges.removeFromTo(""+(i-1),""+i);
                edges.add(e);
            } else if (codeline.startsWith("if")) {
                // do nothing, edges will be added by then and else
            } else if (codeline.startsWith("then")) {
                // edge will start from the if node
                int lastspace = codeline.lastIndexOf(" ");
                int lastsemicolumn = codeline.lastIndexOf(";");
                String target = codeline.substring(lastspace, lastsemicolumn);
                Edge e = new Edge("" + (i - 1), target, "then");
                edges.add(e);
            } else if (codeline.startsWith("else")) {
                // edge will start from the if node
                int lastspace = codeline.lastIndexOf(" ");
                int lastsemicolumn = codeline.lastIndexOf(";");
                String target = codeline.substring(lastspace, lastsemicolumn);
                Edge e = new Edge("" + (i - 2), target, "else");
                edges.add(e);
            } else if (codeline.startsWith("return")) {
                // I'm putting a loop on the return, but I could also put nothing
                Edge e = new Edge("" + i, "" + i);
                edges.add(e);
            } else {
                Edge e = new Edge("" + i, "" + (i + 1));
                edges.add(e);
            }
        }

        NodeList toremove = new NodeList();
        // remove goto nodes for cleanliness
        for (Node n : nodes) {
            if (n.getCodeline().startsWith("goto")) {
                EdgeList elist = edges.getEdgesFrom(n);
                assert (elist.size() == 1); //a goto node should have only 1 outgoing edge
                Edge e = elist.get(0);
                String from = e.getFrom(); // address of the goto
                String dest = e.getTo(); //destination of the goto
                for (Edge eache : edges) {
                    if (eache.getTo().contentEquals(from)) {
                        eache.setTo(dest);
                    }
                }
                edges.remove(e);
                toremove.add(n);
            }
        }
        if (!toremove.isEmpty()) {
            for (Node n : toremove) {
                nodes.remove(n);
            }
        }

        //experimental: compress CFG (very unoptimized)
        boolean removednode = false; // has a node being tagged for removal?
        while (!removednode) {
            Node nodetoremove = null;
            for (Node n : nodes) {
                if (n.getCodeline().startsWith("if")) {
                    continue;
                }
                EdgeList efrom = edges.getEdgesFrom(n);
                EdgeList eto = edges.getEdgesTo(n);
                if ((efrom.size() == 1) && (eto.size() < 2)) {
                    Edge tosucc = efrom.get(0);
                    Node succ = nodes.getNodeByName(tosucc.getTo());
                    EdgeList succfrom = edges.getEdgesFrom(succ);
                    EdgeList succto = edges.getEdgesTo(succ);
                    if ((succfrom.size() == 1) && (succto.size() == 1)) {
                        if (eto.size() > 0) {
                            Edge fromprec = eto.get(0);
                            Node prec = nodes.getNodeByName(fromprec.getFrom());
                            edges.removeFromTo(prec.getElement(), n.getElement());
                            edges.add(new Edge(prec.getElement(), succ.getElement()));
                        }

                        edges.removeFromTo(n.getElement(), succ.getElement());

                        succ.setCodeline(n.getCodeline() + "\n" + succ.getCodeline());
                        nodetoremove = n;
                        removednode = true;
                        break;
                    }
                }
            }
            if (removednode) {
                nodes.remove(nodetoremove);
                removednode = false;
            } else {
                break;
            }
        }

        return new ControlFlowGraph(nodes, edges);
    }

}
