/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/

package hyleak;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.regex.Matcher;

/**
 *
 * @author  
 * @author  
 */
class CodePriorDistributionManager extends PriorDistributionManager{
    private ArrayList<String> codelines = null;
    


    public CodePriorDistributionManager(ArrayList<String> code) {
        super();
        codelines = code;
    }
    
    
    
    
    
    @Override
    public void analyze(){
        secretsize = new BigInteger("0");
        for (String line : codelines) {
            if (line.startsWith("secret")){
                if (logger.getVerbosity()>=4){
                    log("Testing if it matches regular expression "+HyleakPatterns.pattern_secret.pattern(),4);
                }
                Matcher matcher = HyleakPatterns.pattern_secret.matcher(line );
                if ( matcher.find() ) {
                    //log("Num groups: " + matcher.groupCount(),4);
                    log("Variable length: " + matcher.group(1),4);
                    log("Variable name: " + matcher.group(2),4);
                    Integer varlength = Integer.valueOf(matcher.group(1));
                    String varname = matcher.group(2);
                    BigInteger seclength = BigInteger.valueOf(2).pow(varlength);
                    String varinterval="[0,"+seclength.toString()+"]";
                    initialsecretintervals.put(varname, new Interval(varinterval));
                    log("Secret length: " + seclength,4);
                    if (secretsize.compareTo(BigInteger.ZERO)==0){
                        secretsize = seclength;
                    }else{
                        secretsize = secretsize.multiply(seclength);
                    }
                }else{
                    
                    if (logger.getVerbosity()>=4){
                        log("Testing if it matches regular expression "+HyleakPatterns.pattern_secret_with_assignment.pattern(),4);
                    }
                    
                    matcher = HyleakPatterns.pattern_secret_with_assignment.matcher(line );
                    if ( matcher.find() ) {
                        //log("Num groups: " + matcher.groupCount(),4);
                        //log("Variable length: " + matcher.group(1),4);
                        log("Variable name: " + matcher.group(2),4);
                        log("Assigned interval: "+matcher.group(3),4);
                        //varlength = Integer.valueOf(matcher.group(1));
                        String varname = matcher.group(2);
                        Interval varinterval = new Interval(matcher.group(3));
   
                        initialsecretintervals.put(varname, varinterval);
                        BigInteger seclength = varinterval.getSize();
                        log("Secret length: " + seclength,4);
                        if (secretsize.compareTo(BigInteger.ZERO)==0){
                            secretsize = seclength;
                        }else{
                            secretsize = secretsize.multiply(seclength);
                        }
                    }
                }
                
            }
           
        }
    }
}
