/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014 - INRIA Rennes (FR)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/
    
package hyleak;
import java.io.*;
import java.util.ArrayList;
import java.util.Scanner;
import support.Logger;
/**
 *
 * @author  
 */
public class CodeReader {
  private final File fFile;
  private final ArrayList<String> linelist;
  private int linenumber;
  private boolean hasRandom;
  
  private static void log(Object aObject, int level){
        Logger logger = Logger.getInstance();
        logger.log(String.valueOf(aObject),level);
    }
    
  public CodeReader(String aFileName){
    fFile = new File(aFileName);  
    log("Reading from file \""+fFile+"\"",2);
    linenumber = 1;
    linelist = new ArrayList<>();
    linelist.add(aFileName);
    hasRandom = false;
  }
  
  public final ArrayList<String> readLines() throws FileNotFoundException {
    
    try (Scanner scanner = new Scanner(new FileReader(fFile))) {
      //first use a Scanner to get each line
        log("--- Begin input code ---",4);
      while ( scanner.hasNextLine() ){
          String line = scanner.nextLine();
          log(linenumber+":   "+line,4);
          if (line.startsWith("random ")||line.startsWith("private")){
              this.hasRandom = true;
          }
          linelist.add(linenumber, line);
          linenumber ++;
      }
    }
    log("--- End input code ---",4);
    log("Read "+linenumber+" lines from input file "+linelist.get(0)+"\n",2);
    
    return linelist;
    
  }
  
  public boolean isRandomized(){
      return this.hasRandom;
  }
 
}
