/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

package hyleak;

import java.math.BigDecimal;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.locks.ReentrantLock;
import support.Contexter;
import support.Logger;
import theory.BigFunctions;

/**
 *
 * @author  
 */
public class ConcurrentSimplePosteriorDistributionManager{
    protected static final Contexter contexter = Contexter.getInstance();
    protected static final Logger logger = Logger.getInstance();
    
    protected final ConcurrentHashMap<String,SimpleMultiIntervalProbabilityDistribution> conditionaldistmap = new ConcurrentHashMap();
    protected final ConcurrentSimpleProbabilityDistribution outputdist = new ConcurrentSimpleProbabilityDistribution();
    private final ReentrantLock lock = new ReentrantLock();
    private boolean normalized = false;

    public void inferInformation(MarkovChainState finalstate) {
        VariableList secrets = new VariableList();
        VariableList observables = new VariableList();
        VariableList vars = finalstate.getVariables();
        normalized = false;
        
        for (Variable var : vars){
            if (var.isPrivate()){
                PrivateVariable privvar =(PrivateVariable)var;
                if (privvar.isSecret()){
                    secrets.add(privvar);            
                }
                //newsecretsize = newsecretsize.add(asecret.getSize());
            }else{
                PublicVariable pubvar =(PublicVariable)var;
                if (pubvar.isObservable()){
                    observables.add(pubvar);            
                }
            }
        }
        log("",3);
        log("Observation: "+observables.getIdentifier()+"\nSecret: "+secrets.getIdentifier(),3);
        
        // use the name and value of the observables as key. the identifier is a string like "o=17;"
        String key = observables.getIdentifier();
        BigDecimal stateprob = finalstate.getProbability();
        String varidentifier = vars.getIdentifier();
        
        log("Updating probability distribution after observing "+key+" with variables "+varidentifier+" and probability "+stateprob.toPlainString(),3);
        
        // outputdist is the probability distributions over all observed outputs
        outputdist.addProbability(key, stateprob);
        if (logger.getVerbosity()>2){
            log("The new distribution on the observables is:\n"+outputdist.printNice(),3);
            log("with total probability "+outputdist.getTotalProbability(),3);
        }
        
        log("Updating conditional posterior distribution for "+key,3);
        lock.lock();
        // outputs maps each possible output to the corresponding posterior distributions over the secrets
        if (conditionaldistmap.containsKey(key)){ //the output has already appeared; update the corresponding prob dists with the new information about the secrets
             //try{
              
                lock.unlock();
                SimpleMultiIntervalProbabilityDistribution conditionaldistribution =(SimpleMultiIntervalProbabilityDistribution) conditionaldistmap.get(key);
                log("Locking posterior distribution of output "+key,3);
                synchronized (conditionaldistribution){
                    log("Old distribution on the secret if the observable is "+key+":\n"+conditionaldistribution.printNice(),3); 
                    conditionaldistribution.update(secrets,stateprob);
                    log("New distribution on the secret if the observable is "+key+":\n"+conditionaldistribution.printNice(),3); 
            //    }
                }
                log("Unlocking posterior distribution of output "+key,3);
            //}finally{
            //    log("Releasing lock",3);
            //    lockmap.get(key).unlock();
            //}
            
        }else{ // the output has been observed for the first time; create a new prob dist
            SimpleMultiIntervalProbabilityDistribution newmap = new SimpleMultiIntervalProbabilityDistribution();
            newmap.initialize(secrets,stateprob);
            //ReentrantLock newlock = new ReentrantLock();
            conditionaldistmap.put(key, newmap); // maps this output to the new prob dist 
            lock.unlock();
            log("Created new distributions for output "+key,3);
        }

        //TEMPORARY - TESTING ONLY
        //logger.setVerbosity(2);
    }
   
    private static void log(Object aObject, int level){
        Logger loggy = Logger.getInstance();
        loggy.log(String.valueOf(aObject),level);
    }
    
    BigDecimal getTotalProbability() {
        return outputdist.getTotalProbability();
    }

    ConcurrentSimpleProbabilityDistribution getDistributionOnOutputs() {
        return outputdist;
    }
    
    public BigDecimal computePosteriorShannonEntropy() {
        BigDecimal toreturn = contexter.getDecimalZero();
        //outputdist.normalize();
        BigDecimal totobsprob = outputdist.getTotalProbability();

        if (!normalized){
            this.normalize();
        }
        
        for (String aout : outputdist.keySet()){
            BigDecimal prob = outputdist.getProbability(aout);
            log("Output "+aout+" has probability "+prob+".\nIts conditional probability distribution is",3);
            BigDecimal entropy = contexter.getDecimalZero();
            SimpleMultiIntervalProbabilityDistribution dist = conditionaldistmap.get(aout);
                       
            log(dist.printNice(),3);
            
            entropy = entropy.add(dist.getEntropy(),contexter.getContext());
            log("Its conditional entropy is "+entropy,3);
                    
            BigDecimal contr = prob.multiply(entropy, contexter.getContext());
            log("Output "+aout+" contributes "+contr+" bits to the posterior entropy estimate\n",3);
            toreturn = toreturn.add(contr);
        }
        // assume that the unobserved part of the system has the same leakage as the observed part
        // totobsprob = totobsprob.multiply(new BigDecimal(100));
        log("Conditional Shannon entropy after observing "+(totobsprob.multiply(new BigDecimal(100))).toEngineeringString()+"% of the probability space is "+toreturn.toEngineeringString(),3);
        //toreturn = toreturn.divide(totobsprob, contexter.getContext());
        
        return toreturn;
    }

    private void normalize() {
        log("I observed "+outputdist.size()+" outputs for a total observed probability of "+outputdist.getTotalProbability().toEngineeringString(),2);
        for (String aout : outputdist.keySet()){
            BigDecimal prob = outputdist.getProbability(aout);
            log("Output "+aout+" has probability "+prob+".\nIts conditional probability distribution is",4);
            //BigDecimal entropy = contexter.getDecimalZero();
            SimpleMultiIntervalProbabilityDistribution dist = conditionaldistmap.get(aout);
            
            log("Posterior distribution on the secrets before normalizing:",4);
         
            if (logger.getVerbosity()>=4){
                log(dist.printNice(),4);
            }

            //normalize
            dist.normalize();
            
            log("Posterior distribution on the secrets after normalizing:",4);

            if (logger.getVerbosity()>=4){
                log(dist.printNice(),4);
            }
        }
        normalized = true;
    }

    BigDecimal computePosteriorMinEntropy() {
        BigDecimal toreturn = contexter.getDecimalZero();
        //outputdist.normalize();
        BigDecimal totobsprob = outputdist.getTotalProbability();

        //log("I observed "+outputdist.size()+" outputs for a total observed probability of "+totobsprob.toPlainString(),3);

        if (!normalized){
            this.normalize();
        }
        
        for (String aout : outputdist.keySet()){
            BigDecimal prob = outputdist.getProbability(aout);
            log("Output "+aout+" has probability "+prob+".\nIts conditional probability distribution is",3);
            BigDecimal minentropy = contexter.getDecimalZero();
            SimpleMultiIntervalProbabilityDistribution dist = conditionaldistmap.get(aout);
                       
            log(dist.printNice(),3);
            BigDecimal highestprob = dist.getHighestProbability();
            minentropy = contexter.getDecimalZero().subtract(log2(highestprob),contexter.getContext());
            log("Its conditional entropy is "+minentropy,3);
                    
            BigDecimal contr = prob.multiply(minentropy, contexter.getContext());
            log("Output "+aout+" contributes "+contr+" bits to the posterior entropy estimate\n",3);
            toreturn = toreturn.add(contr);
        }
        // assume that the unobserved part of the system has the same leakage as the observed part
        // totobsprob = totobsprob.multiply(new BigDecimal(100));
        log("Conditional min-entropy after observing "+(totobsprob.multiply(new BigDecimal(100))).toEngineeringString()+"% of the probability space is "+toreturn.toEngineeringString(),3);
        //toreturn = toreturn.divide(totobsprob, contexter.getContext());
        
        return toreturn;
    }
    
    private BigDecimal log2(BigDecimal value) {
        if (value.compareTo(contexter.getDecimalOne())==0){
            return contexter.getDecimalZero();
        }
        log("Computing the logarithm in base 2 of "+value.toPlainString()+" with scale "+value.scale(),4);
        BigDecimal toreturn = BigFunctions.ln(value,contexter.getContext().getPrecision());
        log("The natural logarithm is "+toreturn,4);
        toreturn = toreturn.divide(BigFunctions.ln(new BigDecimal("2",contexter.getContext()),contexter.getContext().getPrecision()),contexter.getContext());
        log("The logarithm in base 2 is "+toreturn,4);
        
        return toreturn;
    }
}
