/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package hyleak;

import java.math.BigDecimal;
import java.math.BigInteger;

import distributions.MapDistribution;
import support.Logger;

/**
 *
 * @author  
 */
class DiracGuard extends ProbabilisticGuard{

    public DiracGuard(MarkovStateDistributionVariable variable, VariableList thenvars, MapDistribution argdist, String operator, MapDistribution expval) {
        super(variable, thenvars, argdist, operator, expval);
        if (!argdist.isDirac()||!expval.isDirac()){
          log("ERROR: trying to construct a Dirac guard with non-Dirac distributions. Terminating",1);
          System.exit(1);
        } 
    }

    
    private void solve() {
        BigInteger arg = argument.getMinimum();
        BigInteger expr = expvalue.getMinimum();
        log("Evaluating Dirac guard "+arg+' '+operator+' '+expr,4);
        switch (operator) {
            case "==":
                // processing equality
                if (arg.compareTo(expr)==0){
                    isTrue();
                }else{
                    isFalse();
                }
                break;
            case "!=":
                // processing inequality
                if (arg.compareTo(expr)==0){
                    isFalse();
                }else{
                    isTrue();
                }
                break;
            case ">=":
                // processing greater or equal
                if (arg.compareTo(expr)>=0){
                    isTrue();
                }else{
                    isFalse();
                }
                break;
            case "<=":
                // processing lesser or equal
                if (arg.compareTo(expr)<=0){
                    isTrue();
                }else{
                    isFalse();
                }
                break;
            case "<":
                // processing lesser
                if (arg.compareTo(expr)<0){
                    isTrue();
                }else{
                    isFalse();
                }
                break;
            case ">":
                // processing greater
                if (arg.compareTo(expr)>0){
                    isTrue();
                }else{
                    isFalse();
                }
                break;
        }
    }
    
    private static void log(Object aObject, int level){
        Logger loggy = Logger.getInstance();
        loggy.log(String.valueOf(aObject),level);
    }

    private void isTrue() {
        log("The Dirac guard is true",4);
        truevars = vars;
        falsevars = new VariableList();
        proboftrue = contexter.getDecimalOne();
    }

    private void isFalse() {
        log("The Dirac guard is false",4);
        truevars = new VariableList();
        falsevars = vars;
        proboftrue = contexter.getDecimalZero();
    }
   
    @Override
    VariableList getVarsIfTrue() {
        if (proboftrue.compareTo(contexter.getDecimalZero())<0){
            this.solve();
        }
        return truevars;
    }

    @Override
    VariableList getVarsIfFalse() {
        if (proboftrue.compareTo(contexter.getDecimalZero())<0){
            this.solve();
        }
        return falsevars;
    }

    @Override
    BigDecimal getProbOfTrue() {
        if (proboftrue.compareTo(contexter.getDecimalZero())<0){
            this.solve();
        }
        return proboftrue;
    }
    
}
