/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package hyleak;

import java.math.BigInteger;
import java.text.ParseException;
import java.util.regex.Matcher;

import distributions.MapDistribution;
import support.Contexter;
import support.Logger;

/**
 *
 * @author  
 */
public class ExpressionEvaluator {
    private static final Contexter contexter = Contexter.getInstance();
    private static final Logger logger = Logger.getInstance();
    
    private static void log(Object aObject, int level){
        Logger loggy = Logger.getInstance();
        loggy.log(String.valueOf(aObject),level);
    }
    
     public DiracDistribution evaluateExpressionToDirac(String expr, VariableList vars) throws ParseException{
       boolean found = false;
       while (found == false){
           // look for a closed parenthesis
           int clospos =-1;
           for (int i=0;i<expr.length();i++){
               if (expr.charAt(i)==')'){
                   clospos = i;
                   break;
               }
           }
       
           // if there are no closed parenthesis the expression is a constant; return its value
           if (clospos==-1){
               return this.evaluateInnerExpressionToDirac(expr, vars);
           
           }else{
               //else find the last opened parenthesis before the closed one
               String smallexpr = expr.substring(0, clospos+1);
               log ("Found closed parenthesis: "+smallexpr,4);
               int openpos = smallexpr.lastIndexOf('(');
               smallexpr = smallexpr.substring(openpos);
               log("Found inner expression "+smallexpr,4);
               // found an inner expression; evaluate it (without the parenthesis)
               DiracDistribution innerval = new DiracDistribution("0");
               try{
                   innerval = this.evaluateInnerExpressionToDirac(smallexpr.substring(1, smallexpr.length()-1), vars);
               }catch (ParseException e){
                   throw e;
               }
               
               String smallexprval = ""+innerval;
               expr = expr.replace(smallexpr, smallexprval);
               log("Replaced inner expression "+smallexpr+" with its value "+smallexprval+" in the string; now the string to evaluate is "+expr,4);
           
           }
       }
       
       
       return new DiracDistribution("-1");
   }

    
    public DiracDistribution evaluateInnerExpressionToDirac (String expr, VariableList vars)throws ParseException {
        log("Evaluating inner expression "+expr+" with variables "+vars.toString(),4);
        if (expr.startsWith("(")){
            expr = expr.substring(1,expr.lastIndexOf(")"));
            log("Removing parenthesis; the expression is now "+expr,4);
        }
        if (logger.getVerbosity()>=4){
            log("The regular expression for a constant is "+HyleakPatterns.pattern_expression_constant.pattern(),4);
        }
        
        Matcher matcher = HyleakPatterns.pattern_expression_constant.matcher(expr );
        if ( matcher.matches() ) {
            log("It is the constant " + matcher.group(1)+"; returning it.",4);
            return new DiracDistribution(matcher.group(1));
        }else{
            log("It is not a constant, checking if it is a constant distribution",4);
            if (logger.getVerbosity()>=4){
                log("The regular expression for a constant distribution is "+HyleakPatterns.pattern_constant_distribution.pattern(),4);
            }
            
            matcher = HyleakPatterns.pattern_constant_distribution.matcher(expr );
            if ( matcher.matches() ) {
                log("It is the constant distribution " + matcher.group(1)+", returning it...",4);
                return new DiracDistribution(matcher.group(1));
                
                
                
            }
            log("It is not a constant distribution, checking if it is a variable",4);
            if (logger.getVerbosity()>=4){
                log("The regular expression for a variable is "+HyleakPatterns.pattern_expression_variable.pattern(),4);
            }
            
            matcher = HyleakPatterns.pattern_expression_variable.matcher(expr );
            if ( matcher.matches() ) {
                log("It is the variable " + matcher.group(1)+"; checking its value...",4);
                String varname = this.getVariableName(matcher.group(1), vars);
                MarkovStateDistributionVariable var = vars.getDistributionVariable(varname);
                if (var==null){
                    log("ERROR: variable "+varname+" not found.",4);
                    throw new ParseException("ERROR: variable "+varname+" not found.",0);
                }else if (!var.isDirac()){
                    log("ERROR: variable "+varname+" is not Dirac.",4);
                    throw new ParseException("ERROR: variable "+varname+" is not Dirac.",0);
                }else{
                    BigInteger val = var.getMaximum();
                    log("The value is "+val+"; returning it.",4);
                    return new DiracDistribution(val);
                }
                
                
            }else{
                log("It is not a variable, checking if it is an operation",4);
                if (logger.getVerbosity()>=4){
                    log("The regular expression for an operation is "+HyleakPatterns.pattern_expression_operation.pattern(),4);
                }
                
                matcher = HyleakPatterns.pattern_expression_operation.matcher(expr );
                if ( matcher.matches() ) {
                    String op1 = matcher.group(1);
                    String operator = matcher.group(2);
                    String op2 = matcher.group(3);
                    log("It is the operation "+op1+" "+operator+" "+op2+"; evaluating it...",4);
                    switch (operator) {
                        case "+":
{
                                log("It is a sum; evaluating the operands",4);
                                BigInteger val1 = new BigInteger(op1);
                                BigInteger val2 = new BigInteger(op2);
                                BigInteger result = val1.add(val2);
                                log("The sum result is "+result,4);
                                return new DiracDistribution(result);
                            }
                        case "-":
{
                                log("It is a subtraction; evaluating the operands",4);
                             
                                BigInteger val1 = new BigInteger(op1);
                                BigInteger val2 = new BigInteger(op2);
                                BigInteger result = val1.subtract(val2);
                                log("The subtraction result is "+result,4);
                                return new DiracDistribution(result);
                            }
                        case "/":
{
                                log("It is a division; evaluating the operands",4);
                                
                                BigInteger val1 = new BigInteger(op1);
                                BigInteger val2 = new BigInteger(op2);
                                BigInteger result = val1.divide(val2);
                                log("The division result is "+result,4);
                                return new DiracDistribution(result);
                            }
                        case "*":
{
                                log("It is a product; evaluating the operands",4);
                               
                                BigInteger val1 = new BigInteger(op1);
                                BigInteger val2 = new BigInteger(op2);
                                BigInteger result = val1.multiply(val2);
                                log("The product result is "+result,4);
                                return new DiracDistribution(result);
                            }
                        case "%":
{
                                log("It is a modulo; evaluating the operands",4);
                               
                                BigInteger val1 = new BigInteger(op1);
                                BigInteger val2 = new BigInteger(op2);
                                BigInteger result = val1.mod(val2);
                                log("The modulo result is "+result,4);
                                return new DiracDistribution(result);
                            }
                        case "&&":
{
                                log("It is a logical AND; evaluating the operands",4);
                               
                                BigInteger val1 = new BigInteger(op1);
                                BigInteger val2 = new BigInteger(op2);
                                BigInteger result;
                                if (val1.compareTo(BigInteger.ZERO)==0 && val2.compareTo(BigInteger.ZERO)==0){
                                    result = BigInteger.ZERO;
                                }else{
                                    result = BigInteger.ONE;
                                }
                                log("The AND result is "+result,4);
                                return new DiracDistribution(result);
                            }
                        case "||":
{
                                log("It is a logical OR; evaluating the operands",4);
                               
                                BigInteger val1 = new BigInteger(op1);
                                BigInteger val2 = new BigInteger(op2);
                                BigInteger result;
                                if (val1.compareTo(BigInteger.ZERO)==0 || val2.compareTo(BigInteger.ZERO)==0){
                                    result = BigInteger.ZERO;
                                }else{
                                    result = BigInteger.ONE;
                                }
                                log("The OR result is "+result,4);
                                return new DiracDistribution(result);
                            }
                        case "^":
{
                                log("It is a logical XOR; evaluating the operands",4);
                                
                                BigInteger val1 = new BigInteger(op1);
                                BigInteger val2 = new BigInteger(op2);
                                BigInteger result;
                                if ((val1.compareTo(BigInteger.ZERO)==0 && val2.compareTo(BigInteger.ZERO)==0) || (val1.compareTo(BigInteger.ZERO)!=0 && val2.compareTo(BigInteger.ZERO)!=0)){
                                    result = BigInteger.ZERO;
                                }else{
                                    result = BigInteger.ONE;
                                }
                                log("The XOR result is "+result,4);
                                return new DiracDistribution(result);
                            }
                    }
                    log("ERROR: operator not recognized.",1);
                    throw new ParseException("ERROR: operator not recognized;",0);
                }else {
                 log("It is not an operation, checking if it is a negation",4);
                if (logger.getVerbosity()>=4){
                    log("The regular expression for a negation is "+HyleakPatterns.pattern_negated_expression.pattern(),4);
                
                }
                matcher = HyleakPatterns.pattern_negated_expression.matcher(expr );
                if ( matcher.matches() ) {
                    String negexpr = matcher.group(1);
                   
                    log("It is the negated operation "+negexpr+"; negating it...",4);
                    BigInteger negval = new BigInteger(negexpr.substring(1));
                    
                    if (negval.compareTo(BigInteger.ZERO)==0){
                        negval = BigInteger.ONE;
                    }else{
                        negval = BigInteger.ZERO;
                    }
                    return new DiracDistribution(negval);
                }else {
                    log("ERROR: string "+expr+" does not match any known inner expression, cannot convert to Dirac distribution.",1);
                    throw new ParseException("ERROR: string "+expr+" does not match any known inner expression, cannot convert to Dirac distribution.",0);
                }
            }
            }


        }
    }
    
    
    public MapDistribution evaluateExpressionToDistribution(String expr, VariableList vars) throws ParseException{
        boolean found = false;
       while (found == false){
           // look for a closed parenthesis
           int clospos =-1;
           for (int i=0;i<expr.length();i++){
               if (expr.charAt(i)==')'){
                   clospos = i;
                   break;
               }
           }
       
           // if there are no closed parenthesis the expression is inner; return its value
           if (clospos!=-1){
               //else find the last opened parenthesis before the closed one
               String smallexpr = expr.substring(0, clospos+1);
               log ("Found closed parenthesis: "+smallexpr,4);
               int openpos = smallexpr.lastIndexOf('(');
               smallexpr = smallexpr.substring(openpos);
               log("Found inner expression "+smallexpr,4);
               // found an inner expression; evaluate it (without the parenthesis)
               MapDistribution innerdist = new MapDistribution(BigInteger.ZERO);
               try{
                   innerdist = this.evaluateInnerExpressionToDistribution(smallexpr.substring(1, smallexpr.length()-1), vars);
               }catch (ParseException e){
                   throw e;
               }
               
               String smallexprval = innerdist.toString();
               expr = expr.replace(smallexpr, smallexprval);
               log("Replaced inner expression "+smallexpr+" with its value "+smallexprval+" in the string; now the string to evaluate is "+expr,4);
           
           }else{
               
               found = true;
           }
       }
       return this.evaluateInnerExpressionToDistribution(expr, vars);
    }
       
       
   

    public MapDistribution evaluateInnerExpressionToDistribution(String expr, VariableList vars) throws ParseException {
        log("Evaluating inner expression "+expr+" with variables "+vars.toString(),4);
        if (expr.startsWith("(")){
            expr = expr.substring(1,expr.lastIndexOf(")"));
            log("Removing parenthesis; the expression is now "+expr,4);
        }
        if (logger.getVerbosity()>=4){
            log("The regular expression for a constant is "+HyleakPatterns.pattern_expression_constant.pattern(),4);
        }
        
        Matcher matcher = HyleakPatterns.pattern_expression_constant.matcher(expr );
        if ( matcher.matches() ) {
            log("It is the constant " + matcher.group(1)+"; returning it as a Dirac distribution.",4);
            return new MapDistribution(new BigInteger(matcher.group(1)));
        }else{
            
            if (logger.getVerbosity()>=4){
                log("It is not a constant, checking if it is a distribution",4);
            }
                log("The regular expression for a distribution is "+HyleakPatterns.pattern_constant_distribution.pattern(),4);
            }

            matcher = HyleakPatterns.pattern_constant_distribution.matcher(expr );
            if ( matcher.matches() ) {
                log("It is the distribution " + expr+"; returning it.",4);
                return new MapDistribution(expr);
            }else{

                log("It is not a distribution, checking if it is a variable",4);
                if (logger.getVerbosity()>=4){
                    log("The regular expression for a variable is "+HyleakPatterns.pattern_expression_variable.pattern(),4);
                }

                matcher = HyleakPatterns.pattern_expression_variable.matcher(expr );
                if ( matcher.matches() ) {
                    log("It is the variable " + matcher.group(1)+"; checking its value...",4);
                    String varname = this.getVariableName(matcher.group(1), vars);
                    Variable var = vars.getVariable(varname);
                    if (var==null){
                        log("ERROR: variable "+varname+" not found.",4);
                        throw new ParseException("ERROR: variable "+varname+" not found.",0);

                    }else{
                        MapDistribution val = ((MarkovStateDistributionVariable)var).getDistribution();
                        log("The value is "+val+"; returning it.",4);
                        return val;
                    }


                }else{
                    log("It is not a variable, checking if it is an operation",4);
                    if (logger.getVerbosity()>=4){
                        log("The regular expression for an operation is "+HyleakPatterns.pattern_expression_operation.pattern(),4);
                    }

                    matcher = HyleakPatterns.pattern_expression_operation.matcher(expr );
                    if ( matcher.matches() ) {
                        String op1 = matcher.group(1);
                        String operator = matcher.group(2);
                        String op2 = matcher.group(3);
                        log("It is the operation "+op1+" "+operator+" "+op2+"; evaluating it...",4);

                        MapDistribution result = new MapDistribution(BigInteger.ZERO);
                        MapDistribution val1 = new MapDistribution(new BigInteger(op1));
                        MapDistribution val2 = new MapDistribution(new BigInteger(op2));
                        switch (operator) {
                            case "+":
    {
                                    log("It is a sum; evaluating the operands",4);

                                    result = val1.add(val2);
                                    log("The sum result is "+result,4);
                                    break;
                                }
                            case "-":
    {
                                    log("It is a subtraction; evaluating the operands",4);

                                    result = val1.subtract(val2);
                                    log("The subtraction result is "+result,4);
                                    break;
                                }
                            case "/":
    {
                                    log("It is a division; evaluating the operands",4);

                                    result = val1.divide(val2);
                                    log("The division result is "+result,4);
                                    break;
                                }
                            case "*":
    {
                                    log("It is a product; evaluating the operands",4);

                                    result = val1.multiply(val2);
                                    log("The product result is "+result,4);
                                    break;
                                }
                            case "%":
    {
                                    log("It is a modulo; evaluating the operands",4);

                                    result = val1.mod(val2);
                                    log("The modulo result is "+result,4);
                                    break;
                                }
                            case "&&":
    {
                                    log("It is a logical AND; evaluating the operands",4);

                                    result = val1.and(val2);

                                    log("The AND result is "+result,4);
                                    break;
                                }
                            case "||":
    {
                                    log("It is a logical OR; evaluating the operands",4);

                                    result = val1.or(val2);

                                    log("The OR result is "+result,4);
                                    break;
                                }
                            case "^":
    {
                                    log("It is a logical XOR; evaluating the operands",4);

                                    result = val1.xor(val2);

                                    log("The XOR result is "+result,4);
                                    break;
                                }

                            default:
                                log("ERROR: operator not recognized.",1);
                                throw new ParseException("ERROR: operator not recognized;",0);
                                }


                        return result;

                    }else {
                     log("It is not an operation, checking if it is a negation",4);
                    if (logger.getVerbosity()>=4){
                        log("The regular expression for a negation is "+HyleakPatterns.pattern_negated_expression.pattern(),4);

                    }
                    matcher = HyleakPatterns.pattern_negated_expression.matcher(expr );
                    if ( matcher.matches() ) {
                        String negexpr = matcher.group(1);

                        log("It is the negated operation "+negexpr+"; negating it...",4);
                        MapDistribution negval = new MapDistribution(negexpr.substring(1));
                        negval = negval.negate();
                        return negval;
                    }else {
                        log("ERROR: string "+expr+" does not match any known inner expression, cannot convert to distribution.",1);
                        throw new ParseException("ERROR: string "+expr+" does not match any known inner expression.",0);
                    }
                }
            }


        }
    }

    private String getVariableName(String varname, VariableList variables) {
        String toreturn =varname;
        if (varname.contains("[")){
            String temp = varname.substring(varname.indexOf('['));
            while (temp.contains("[")){
                int closedpar = temp.indexOf(']');
                String index = temp.substring(1, closedpar);
                log("Evaluating index "+index,4);
                index = this.getVariableName(index, variables);
                BigInteger val = new BigInteger("0");
                try {
                    val = this.evaluateInnerExpressionToDirac(index, variables).toBigInteger();
                } catch (ParseException e){
                    log(e.getMessage(),1);
                }
                toreturn = toreturn.replace("["+index+"]", "["+val.toString()+"]");
                temp = temp.substring(closedpar+1);
            }
        }
        return toreturn;
    }

}
