/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/

package hyleak;

import java.math.BigDecimal;
import support.Contexter;
import support.Logger;

/**
 *
 * @author  
 */
class FrequencyAndProbability{
        private Integer frequency;
        private BigDecimal probability;

        protected static final Contexter contexter = Contexter.getInstance();
        protected static final Logger logger = Logger.getInstance();

    public FrequencyAndProbability(Integer frequency, BigDecimal probability) {
        if (frequency<1){
            log("ERROR: Frequency lesser than 1?",1);
            System.exit(1);
        }
        if (probability.compareTo(contexter.getDecimalZero())<=0){
            log("ERROR: Probability zero?",1);
            System.exit(1);
        }
        
        this.frequency = frequency;
        this.probability = probability;
    }
        
    public Integer getFrequency() {
        return frequency;
    }

    public void setFrequency(Integer frequency) {
        if (frequency<1){
            log("ERROR: Frequency lesser than 1?",1);
            System.exit(1);
        }
        this.frequency = frequency;
    }

    public BigDecimal getProbability() {
        
        return probability;
    }

    public void setProbability(BigDecimal probability) {
        if (probability.compareTo(contexter.getDecimalZero())<=0){
            log("ERROR: Probability set to zero?",1);
            System.exit(1);
        }
        this.probability = probability;
    }
        
    public void multiplyProbability(BigDecimal tomult) {
        if (tomult.compareTo(contexter.getDecimalZero())<=0){
            log("ERROR: Probability multiplied by zero?",1);
            System.exit(1);
        }
        this.probability = this.probability.multiply(tomult, contexter.getContext());
    }
    
    
    public void updateProbability(BigDecimal newprob){
        if (this.probability.compareTo(newprob)==0){
            this.frequency = this.frequency + 1;
            return;
        }
        BigDecimal oldcount = new BigDecimal(frequency,contexter.getContext());
        BigDecimal oldcountplusone = oldcount.add(contexter.getDecimalOne(), contexter.getContext());
        BigDecimal normalizedoldprob = probability.divide(oldcountplusone, contexter.getContext()).multiply(oldcount,contexter.getContext());
        BigDecimal normalizednewprob = newprob.divide(oldcountplusone, contexter.getContext());
        this.probability = normalizedoldprob.add(normalizednewprob, contexter.getContext());
        if (this.probability.compareTo(contexter.getDecimalZero())<=0){
            log("ERROR: Probability zero after update?",1);
            System.exit(1);
        }
        this.frequency = this.frequency + 1;
    }  
    
        @Override
    public String toString(){
        return "frequency: "+frequency+", probability: "+probability.toEngineeringString();
    }
    
    protected static void log(Object aObject, int level){
        logger.log(String.valueOf(aObject),level);
    }
}
