/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/

package hyleak;

import java.math.BigDecimal;
import java.util.Collection;
import java.util.HashMap;
import java.util.Set;
import support.Contexter;
import support.Logger;

/**
 *
 * @author  
 */
public class HashFreqProb extends HashMap<String, FrequencyAndProbability> {

    protected static final Contexter contexter = Contexter.getInstance();
    protected static final Logger logger = Logger.getInstance();
    
    public HashFreqProb(String varidentifier, BigDecimal stateprob) {
        this.put(varidentifier, new FrequencyAndProbability(1,stateprob));
    }
    
    public BigDecimal getTotalProbability() {
        Collection<FrequencyAndProbability> allvalues = this.values();
        BigDecimal toreturn = contexter.getDecimalZero();
        for (FrequencyAndProbability value : allvalues){
            toreturn = toreturn.add(value.getProbability(), contexter.getContext());
        }
        return toreturn;
    }
    
    public Integer getTotalFrequency() {
        Collection<FrequencyAndProbability> allvalues = this.values();
        Integer toreturn = 0;
        for (FrequencyAndProbability value : allvalues){
            toreturn = toreturn + value.getFrequency();
        }
        return toreturn;
    }
    
    
    public String printNice() {
        String toreturn = "";
        Set<String> keys = this.keySet();
        for (String key : keys){
            toreturn +="hash "+key+", value "+this.getTotalFrequency()+" and probability "+this.getTotalProbability().toPlainString()+"\n";
        }
        return toreturn;
    }

    public void updateProbability(String variables, BigDecimal probability) {
        FrequencyAndProbability freqprob = this.get(variables);
        if (freqprob == null){
            this.put(variables, new FrequencyAndProbability(1,probability));
        }else{
            freqprob.updateProbability(probability);
        }
        
    }

    public void create(String variables, BigDecimal probability) {
        this.put(variables, new FrequencyAndProbability(1,probability));
    }
    
    

    public void multiplyProbabilities(BigDecimal tomult) {
        for (FrequencyAndProbability fp : this.values()){
            fp.multiplyProbability(tomult);
        }
    }
}
