/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/

package hyleak;

import java.math.BigInteger;
import java.security.SecureRandom;
import java.util.Iterator;
import java.util.regex.Matcher;
import support.Logger;

/**
 *
 * @author  
 */
public class Interval implements Iterable, Comparable<Interval>{
    private String interval;
    private static final Logger logger = Logger.getInstance();
    
    
    
    public Interval(String inter){
        if (inter.isEmpty()){
            interval="";
            return;
        }
        if (isValidInterval(inter)){
            interval = inter;
        }else{
            interval = "";
            log("ERROR: not an interval "+inter,1);
            System.exit(1);
        }
        interval = inter;
        if (!this.isBottomLeqTop()){
            log("ERROR: bottom of interval greater than top: "+inter,1);
            System.exit(1);
        }
        
    }
    
    public Interval (BigInteger bottom, BigInteger top){
        this("["+bottom.toString()+','+top.toString()+']');
    }

    public Interval (BigInteger singleVal){
        this(singleVal,singleVal);
    }
    
    private boolean isBottomLeqTop(){
        return (this.getBottom().compareTo(this.getTop())<=0);
    }

    public BigInteger getSize() {
        log("Computing size of interval "+this.interval,4);
        if (this.isEmpty()) {
            log("The interval is empty, its size is 0",4);
            return BigInteger.ZERO;
        }
        if (this.isSingle()){
            
            BigInteger toreturn = this.getTop().add(BigInteger.ONE).subtract(this.getBottom());
            log("The size of the interval is "+toreturn.toString(),4);
            return toreturn;
            
        }
        
        BigInteger toreturn = this.getHead().getSize().add(this.getTail().getSize());
        log("The size of the interval is "+toreturn.toString(),4);
        return toreturn;
        
//       String intervals = this.interval;
//       while (intervals.compareTo("")!=0){
//           // look for a closed parenthesis
//           int clospos =-1;
//           for (int i=0;i<intervals.length();i++){
//               if (intervals.charAt(i)==']'){
//                   clospos = i;
//                   break;
//               }
//           }
//       
//           // if there are no closed parenthesis we are done; return
//           if (clospos==-1){
//               break;
//           
//           }else{
//               //else find the last opened parenthesis before the closed one
//               String oneinterval = intervals.substring(0, clospos+1);
//               log ("Found closed parenthesis: "+oneinterval,4);
//               int openpos = oneinterval.lastIndexOf('[');
//               oneinterval = oneinterval.substring(openpos);
//               log("Found single interval "+oneinterval,4);
//               // found an inner expression; evaluate it (without the parenthesis)
//               int commapos = oneinterval.lastIndexOf(',');
//               String bottomstring = oneinterval.substring(1, commapos);
//               log("Bottom string: "+bottomstring,4);
//               BigInteger bottom = new BigInteger(bottomstring);
//               String topstring = oneinterval.substring(commapos+1, oneinterval.length()-1);
//               log("Top string: "+topstring,4);
//               BigInteger top = new BigInteger(topstring);
//               toreturn = toreturn.add(top).subtract(bottom).add(BigInteger.ONE);
//               if (intervals.lastIndexOf('[', oneinterval.length())==-1){
//                   intervals = "";
//               }else{
//                   intervals = intervals.substring(oneinterval.length());
//               }
//               
//           }
//       }
//       log("The size of the interval is "+toreturn,4);
//       return toreturn;
    }
    
    
    
    public BigInteger getBottom(){
        int startpos = 1;
        int endpos = interval.indexOf(',');
        String firstvalue = interval.substring(startpos, endpos);
        //log("The bottom is "+firstvalue,4);
        return new BigInteger(firstvalue);
    }
    
    public BigInteger getTop(){
        int startpos = interval.lastIndexOf(',')+1;
        int endpos = interval.length()-1;
        String lastvalue = interval.substring(startpos, endpos);
        //log("The top is "+lastvalue,4);
        return new BigInteger(lastvalue);
    }
    
    private static void log(Object aObject, int level){
        logger.log(String.valueOf(aObject),level);
    }
    
    @Override
    public String toString(){
        if (this.isEmpty()) return "[]";
        return this.interval;
    }
    
    public Interval intersect(Interval tointersect){
        log("Computing intersection of "+this.interval+" and "+tointersect.toString(),4);
        Interval toreturn=this.clone();
        Interval myhead = toreturn.getHead();
        Interval hishead = tointersect.getHead();
        
        Interval oneintersec = toreturn.getSingleIntersection(myhead, hishead);
        
        Interval meleft;
        Interval himleft;
        if (myhead.getTop().compareTo(hishead.getTop())<=0){
            log("Discarding my head "+myhead.toString(),4);
            meleft = toreturn.getTail();
        }else{
            meleft = toreturn;
        }
        
        if (myhead.getTop().compareTo(hishead.getTop())>=0){
            log("Discarding his head "+hishead.toString(),4);
            himleft = tointersect.getTail();
        }else{
            himleft = tointersect;
        }
        //log("Left of me: "+meleft.toString()+"\nLeft of him: "+himleft.toString(),4);
        if (meleft!=null && himleft!=null){
            toreturn = oneintersec.concatenate(meleft.intersect(himleft));
        }else{
            toreturn = oneintersec;
        }
        return toreturn;
        
    }
    
    public void compress(){
        log("Compressing interval "+this.interval,4);        
        if (this.isEmpty() || this.isSingle()){
            log("Interval is empty or single. Nothing to compress.",4);
            return;
        }
        Interval head = this.getHead();
        Interval tail = this.getTail();
        
        tail.compress();
        
        BigInteger headtop = head.getTop();
        BigInteger tailbottom = tail.getBottom();
        
        if (headtop.add(BigInteger.ONE).compareTo(tailbottom)>=0){
            // compression needed
            Interval tailhead = tail.getHead();
            Interval tailtail = tail.getTail();
            String newstring = "["+head.getBottom().toString()+","+tailhead.getTop().toString()+"]";
            if (tailtail!=null){
                newstring = newstring + tailtail.toString();
            }
            this.interval=newstring;
        }else{
            this.interval=head.toString()+tail.toString();
        }
        log("Compressed to "+this.interval,4);
    }

    private boolean isValidInterval(String inter) {
        Matcher matcher = HyleakPatterns.pattern_intervals.matcher(inter);
        return matcher.matches();
    }
    
    public void reduce(String operator, BigInteger expval) {
        //take the single intervals and reduce them according to the operator and value
        String newintervals = "";
       
       while (this.interval.compareTo("")!=0){
           // look for a closed parenthesis
           
           int clospos = this.interval.indexOf(']');
       
           // if there are no closed parenthesis we are done; return
           if (clospos==-1){
               break;
           
           }else{
               //else find the last opened parenthesis before the closed one
               String oneintervalstring = this.interval.substring(0, clospos+1);
               Interval oneinterval = new Interval(oneintervalstring);
               log("Found single interval "+oneinterval,4);
               
               
               Interval onenewinterval = this.reduceSingleInterval(oneinterval,operator,expval);
               newintervals = newintervals.concat(onenewinterval.toString());
               
               if (this.interval.lastIndexOf('[', oneintervalstring.length())==-1){
                   this.interval = "";
               }else{
                   this.interval = this.interval.substring(oneintervalstring.length());
               }
               
           }
       }
       this.interval=newintervals;
    
    }
    
    private Interval reduceSingleInterval(Interval interv, String operator, BigInteger expval) {
        log("Reducing "+interv+" assuming that "+operator+expval+" holds",4);
        String toreturn = "";
        BigInteger bottom = interv.getBottom();
        BigInteger top = interv.getTop();
      switch (operator) {
          case "==":
              // processing equality
              // if the value is outside the range return empty; else return singleton interval
              if ((expval.compareTo(bottom)>=0)&&(expval.compareTo(top)<=0)){
                  toreturn = "["+expval+","+expval+"]";
              }else{
                  toreturn = "";
              }
              break;
          case "!=":
              // processing inequality
              // if the value is less than bottom or more than top, nothing to do
              if ((expval.compareTo(bottom)<0)||(expval.compareTo(top)>0)){
                  toreturn = "["+bottom+","+top+"]";
              }else if ((expval.compareTo(bottom)==0)&&(expval.compareTo(top)==0)){
                  // return empty string
                  toreturn = "";
              }else if (expval.compareTo(bottom)==0){
                  // cut the bottom value
                  toreturn = "["+(bottom.add(BigInteger.ONE))+","+top+"]";
              }else if (expval.compareTo(top)==0){
                  // cut the top value
                  toreturn = "["+bottom+","+(top.subtract(BigInteger.ONE))+"]";
              }else{
                  // the value is inside; we need to split the interval in two
                  toreturn = "["+bottom+","+(expval.subtract(BigInteger.ONE))+"]["+(expval.add(BigInteger.ONE))+","+top+"]";
              }
              break;
          case ">=":
              // processing greater or equal
              // if the value is lesser or equal to bottom, nothing to do
              if (expval.compareTo(bottom)<=0){
                  toreturn = "["+bottom+","+top+"]";
              }else if (expval.compareTo(top)>0){
                  // if the value is greater that top, return empty string
                  toreturn = "";
              }else {
                  // else move the bottom to expval
                  toreturn = "["+expval+","+top+"]";
              }
              break;
          case "<=":
              // processing lesser or equal
              // if the value is greater or equal to top, nothing to do
              if (expval.compareTo(top)>=0){
                  toreturn = "["+bottom+","+top+"]";
              }else if (expval.compareTo(bottom)<0){
                  // if the value is lesser that bottom, return empty string
                  toreturn = "";
              }else {
                  // else move the top to expval
                  toreturn = "["+bottom+","+expval+"]";
              }
              break;
          case "<":
              // processing lesser
              // if the value is greater than top, nothing to do
              if (expval.compareTo(top)>0){
                  toreturn = "["+bottom+","+top+"]";
              }else if (expval.compareTo(bottom)<=0){
                  // if the value is lesser or equal that bottom, return empty string
                  toreturn = "";
              }else {
                  // else move the top to expval-1
                  toreturn = "["+bottom+","+(expval.subtract(BigInteger.ONE))+"]";
              }
              break;
          case ">":
              // processing greater
              // if the value is lesser than bottom, nothing to do
              if (expval.compareTo(bottom)<0){
                  toreturn = "["+bottom+","+top+"]";
              }else if (expval.compareTo(top)>=0){
                  // if the value is greater or equal than top, return empty string
                  toreturn = "";
              }else {
                  // else move the bottom to expval+1
                  toreturn = "["+(expval.add(BigInteger.ONE))+","+top+"]";
              }
              break;
      }
        log("Reduced to "+toreturn,4);
        return new Interval(toreturn);
    }
    
    @Override
    public Interval clone(){
        return new Interval(this.interval);
    }

    public boolean isSingle(){
        return (interval.indexOf(']')==interval.length()-1);
    }
    
    public boolean isSingleton(){
        return (this.getBottom().compareTo(this.getTop())==0);
    }
    
    public String getHeadString(){
        return this.interval.substring(0, this.interval.indexOf(']')+1);
    }
    
    public Interval getHead(){
        String head = this.getHeadString();
        return new Interval(head);
    }
    
    public String getTailString(){
        return this.interval.substring(this.interval.indexOf(']')+1);
    }
    
    public Interval getTail(){
        String tail = this.getTailString();
        if (tail.isEmpty()==false){
            return new Interval(tail);
        }
        return null;
    }
    
    public boolean isEmpty(){
        return this.interval.isEmpty();
    }
    
    public Interval concatenate(Interval toconcat){
        if (this.isEmpty()){
            if (toconcat.isEmpty()){
                return new Interval("");
            }else{
                return toconcat.clone();
            }
        }else if (toconcat.isEmpty()){
            return this.clone();
        }
        Interval toreturn = this.clone();
        BigInteger mytop = toreturn.getTop();
        BigInteger hisbottom = toconcat.getBottom();
        
        if (mytop.compareTo(hisbottom)>0){
            log("ERROR: cannot concatenate intervals "+this.interval+" and "+toconcat.toString()+" because top of the first is greater than bottom of the second.",1);
            System.exit(1);
        }
        toreturn.interval = toreturn.interval.concat(toconcat.toString());
        toreturn.compress();
        return toreturn;
    }
    
    public Interval subtract(Interval tosubtract) {
        log("Subtracting "+tosubtract.toString()+" from "+this.interval,4);
        Interval toreturn = this.clone();
        if (tosubtract.isEmpty()){
            return toreturn;
        }
        
        BigInteger mybottom = toreturn.getBottom();
        BigInteger mytop = toreturn.getTop();
        BigInteger subbottom = tosubtract.getBottom();
        BigInteger subtop = tosubtract.getTop();
        
        if (mybottom.compareTo(subtop)>0){
            //no intersection; nothing to do
            return toreturn;
        }
        if (mytop.compareTo(subbottom)<0){
            //no intersection; nothing to do
            return toreturn;
        }
        String newinterval = "";
        if (toreturn.isSingle()){ // I have a single interval (no tail)
            Interval itshead = tosubtract.getHead();
            BigInteger headbottom = itshead.getBottom();
            BigInteger headtop = itshead.getTop();
            boolean visittail = true;
           // log("Subtracting "+itshead.toString()+" from "+toreturn.interval,4);
            
            if (mybottom.compareTo(headtop)>0){
                    /*
                    case ME:             [      ]
                         HIM:  [     ]
                    */
                log("The intervals do not intersect, nothing to do but checking the tail",4);
                newinterval = toreturn.interval;
            }else if (mytop.compareTo(headbottom)<0){
                    /*
                    case ME:   [      ]
                         HIM:             [     ]
                    */
                log("The intervals do not intersect, nothing to do and ignoring the tail",4);
                visittail = false;
                newinterval = toreturn.interval;
            }else{
            
                if (mybottom.compareTo(headbottom)>=0){
                    if (mytop.compareTo(headtop)>=0){
                        /*
                        case ME:     [      ]
                             HIM:  [     ]
                        */
                        log("case 1",4);
                        log("My top is "+mytop.toString()+" , his top is "+headtop.toString(),4);
                        if (mytop.compareTo(headtop)!=0){
                            newinterval ="["+headtop.add(BigInteger.ONE).toString()+","+mytop.toString()+"]"; 
                        }else{
                            newinterval = "";
                            visittail = false;
                        }
                    }else{

                        /*
                        case ME:     [      ]
                             HIM:  [         ]
                        */
                        //nothing else to do, this interval will be deleted
                        log("case 2",4);
                        visittail = false;
                    }

                }else{
                    if (mytop.compareTo(headtop)<=0){
                        /*
                        case ME:   [     ]
                             HIM:    [      ]
                        */
                        log("case 3",4);
                        if (mybottom.compareTo(headbottom)!=0){
                            newinterval ="["+mybottom.toString()+","+headbottom.subtract(BigInteger.ONE)+"]"; 
                        }
                        visittail = false;
                    }else{
                        /*
                        case ME:   [         ]
                             HIM:    [     ]
                        */
                        log("case 4",4);
                        newinterval ="["+mybottom.toString()+","+headbottom.subtract(BigInteger.ONE)+"]["+headtop.add(BigInteger.ONE).toString()+","+mytop.toString()+"]";
                    }
                }
            }
            if (newinterval.isEmpty()){
                log("The interval is now empty",4); 
            }else{
                log("The interval is now "+newinterval,4);
            }
            toreturn.interval = newinterval;
            // I compared myself to its head, now recursively to the rest if needed
          
            if (visittail){
                Interval itstail = tosubtract.getTail();
                if (itstail!=null){
                    toreturn = toreturn.subtract(tosubtract.getTail());
                }
            }
         
            
        }else{ // I have more than one interval (head and tail)
            Interval myhead = toreturn.getHead();
            Interval mytail = toreturn.getTail();
            Interval newhead = myhead.subtract(tosubtract);
            newinterval = newhead.toString();
//            if (myhead.isEmpty()){
//                newinterval = "";
//            }else{
//                newinterval = myhead.toString();
//            }
            log("Now my head is "+newinterval,4);
            if (mytail != null){
                log("I have the tail "+mytail.toString()+", subtracting "+tosubtract.toString()+" from it",4);
                Interval newtail = mytail.subtract(tosubtract);
                if (!newtail.isEmpty()){
                    log("My tail is now "+newtail.toString(),4);
                    newinterval = newinterval + newtail.toString();
                }else{
                    log("My tail is now empty",4);
                }
            }
            log("Now I am "+newinterval,4);
            toreturn = new Interval(newinterval);
        }
        
        return toreturn;
    }

    private Interval getSingleIntersection(Interval myhead, Interval hishead) {
        log("Computing single intersection of "+myhead.toString()+" and "+hishead.toString(),4);
        if (myhead.isEmpty() || hishead.isEmpty()){
            return new Interval ("");
        }
        
        BigInteger mybottom = myhead.getBottom();
        BigInteger hisbottom = hishead.getBottom();
        BigInteger newbottom;
        if (mybottom.compareTo(hisbottom)>=0){
            newbottom = mybottom;
        }else{
            newbottom = hisbottom;
        }
        
        BigInteger mytop = myhead.getTop();
        BigInteger histop = hishead.getTop();
        BigInteger newtop;
        if (mytop.compareTo(histop)<=0){
            newtop = mytop;
        }else{
            newtop = histop;
        }
        Interval toreturn;
        if (newtop.compareTo(newbottom)<0){
            toreturn= new Interval ("");
        }else{
            toreturn= new Interval("["+newbottom.toString()+","+newtop.toString()+"]");
        }
        log("The single intersection is "+toreturn.toString(),4);
        return toreturn;
    }
    
    public boolean equals(Interval that){
        return this.toString().equalsIgnoreCase(that.toString());
    }

    BigInteger getFirstValueAfter(BigInteger current) {
        Interval head = this.getHead();
        BigInteger headtop = head.getTop();
        BigInteger headbottom = head.getBottom();
        if (current.compareTo(headbottom)<0){
            return headbottom;
        }
        if (current.compareTo(headtop)<0){
            return current.add(BigInteger.ONE);
        }
        Interval tail = this.getTail();
        if (tail.isEmpty()){
            // current is the highest value of the interval or more, there is no value after
            log("ERROR: no value after "+current+" in interval "+this.toString(),1);
            System.exit(1);
        }
        return tail.getFirstValueAfter(current);
    }

    @Override
    public Iterator<BigInteger> iterator() {
        return new IntervalIterator(this);
    }

    @Override
    public int compareTo(Interval otherInterv) {
        return this.toString().compareTo(otherInterv.toString());
    }

    Interval RandomlyChoose() {
        BigInteger mysize = this.getSize();
        SecureRandom srand = new SecureRandom();
        BigInteger pos;
        do {
        pos = new BigInteger(mysize.bitLength(), srand);
        } while (pos.compareTo(mysize) >= 0); 
        BigInteger elementAtPos = this.getElementAtPosition(pos);
        return new Interval(elementAtPos);
    }

    private BigInteger getElementAtPosition(BigInteger pos) {
        if (this.getSize().compareTo(pos)<0){
            log("ERROR: unable to return value at position "+pos+", interval only has "+this.getSize()+" elements",1);
            System.exit(1);
        }
        Interval iterInterv = this;
        Interval head = iterInterv.getHead();
        BigInteger headsize = head.getSize();
        while (headsize.compareTo(pos)<0){
            pos = pos.subtract(headsize);
            iterInterv = iterInterv.getTail();
            head = iterInterv.getHead();
            headsize = head.getSize();
        }
        return head.getBottom().add(pos);
    }
}
