/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package hyleak;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.Set;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;
//import bham.leakiest.ProbDist;





import distributions.SingleDistribution;
import distributions.JointDistribution;
import distributions.JointFrequencyDistribution;
import theory.BigFunctions;
import theory.EstimateMI;
//import support.ConvertLeakiEstObjects;
import support.Contexter;
import support.Logger;

/**
 * @author  
 * @author  
 */
public class JointPosteriorDistributionManager {
    private static final Contexter contexter = Contexter.getInstance();
    private static final Logger logger = Logger.getInstance();

    /////////////////////////////////////////////////////////////
    // Whether the prior is known to the analyst or not  
    protected static final boolean priorIsKnown = true;
    
    /////////////////////////////////////////////////////////////
    // Prior distribution manager
    private PriorDistributionManager priorDistMan = null;

    // Prior distribution on secrets
    private SingleDistribution priorDistribution = null;

    /////////////////////////////////////////////////////////////
    // list of component ID's (each simulated statistically)
    private Signature[] listComponentIDs = null;

    /////////////////////////////////////////////////////////////
    // Exact sub-prior (obtained by trace analysis)
    //private SingleDistribution exactSubPrior = null;
    // List of empirical sub-priors
    //private HashMap<Signature, SingleDistribution> mapSubPriors = new HashMap<>();

    // List of all sub-priors
    // (the first element is the exact one obtained by trace analysis
    //  and the rest of elements is obtained by statistical analysis.)
    private SingleDistribution[] listSubPriors = null;
    
    // Exact sub-prior (obtained by trace analysis)
    //private SingleDistribution exactSubPrior = null;
    
    // List of the probabilities of all components
    private BigDecimal[] listExecProbs = null;

    // List of the sizes of supports
    private int[] listSizesOfSupports = null;
    
    /////////////////////////////////////////////////////////////
    // Joint distribution obtained by combining all anaylses
    private JointDistribution totalJointDist = null;

    // Exact joint sub-distribution (obtained by trace analysis)
    private JointDistribution exactSubJointDist = null;
    // List of frequency sub-distribution (obtained by statistical analyses)
    private HashMap<Signature, JointFrequencyDistribution> mapFreqSubJoints = new HashMap<>();
    // List of empirical joint sub-distribution (obtained by statistical analyses)
    private HashMap<Signature, JointDistribution> mapSubJoints = new HashMap<>();

    // List of all joint sub-distributions
    // (the first element is the exact one obtained by trace analysis
    //  and the rest of elements is obtained by statistical analysis.)
    private JointDistribution[] listSubJoints = null;

    /////////////////////////////////////////////////////////////
    // List of sample sizes (used by statistical analyses of components)
    private HashMap<Signature, BigInteger> mapUsedSampleSizes = new HashMap<>();
    private BigInteger[] usedSampleSizes = null;
    // Total number of simulations of all components
    private BigInteger totalNoOfSim = null;

    /////////////////////////////////////////////////////////////
    // Exact total probability (obtained by trace analysis)
    private BigDecimal exactTotalProb = null;
    // List of sums of probabilities (obtained by statistical analyses)
    private HashMap<Signature, BigDecimal> mapTotalProbs = new HashMap<>();

    /////////////////////////////////////////////////////////////
    // Distribution on secrets
    private SingleDistribution distSecrets = null;
    // Distribution on observables
    private SingleDistribution distObservables = null;

    /////////////////////////////////////////////////////////////
    // Estimated size of the secret domain TODO: to be supported
    private HashMap<Signature, Integer> mapEsitmateNumOfSecrets = new HashMap<>();
    // Estimated size of the observable domain TODO: to be supported
    private HashMap<Signature, Integer> mapEsitmateNumOfObservables = new HashMap<>();
    // List of the sizes of supports
    private int[] listEsitmateNumOfSecrets = null;
    private int[] listEsitmateNumOfObservables = null;
    
    /////////////////////////////////////////////////////////////
    // Estimated posterior Shannon entropy, mutual information and its confidence interval
    private BigDecimal PosteriorEntropyPointEstimate = null;
    private BigDecimal MIPointEstimate = null;
    private BigDecimal MIPointEstimateBias = null;
    private BigDecimal MICorrectedPointEstimate = null;
    private BigDecimal[] confidenceInterval = null;

    /**
     * Constructs the joint posterior distribution manager by giving the prior
     * distribution manager.
     *
     * @param priorDistMan prior distribution manager
     * @param totalNoOfSim the total number of simulations of the code
     */
    public JointPosteriorDistributionManager(PriorDistributionManager priorDistMan, BigInteger totalNoOfSim) {
        this.priorDistMan = priorDistMan;
        this.totalNoOfSim = totalNoOfSim;
        log("Joint distribution manager is constructed with the sample size " + totalNoOfSim, 3);
    }

    /**
     * Constructs the joint posterior distribution manager by giving the prior
     * distribution on secrets. (We use this methond especially when we read
     * files of leakiEst's format.)
     *
     * @param priorDistribution prior distribution (SingleDistribution)
     * @param totalNoOfSim the total number of simulations of the code
     */
    public JointPosteriorDistributionManager(SingleDistribution priorDistribution, BigInteger totalNoOfSim) {
        this.priorDistribution = priorDistribution;
        this.totalNoOfSim = totalNoOfSim;
        log("Joint distribution manager is constructed with the sample size " + totalNoOfSim, 3);
    }

    /////////////////////////////////////////////////////////////
    // Methods for determining parameters for statistical simulations
    /**
     * Computes the (initial) sample sizes for the staistical simulations of
     * components.
     *
     * @param mapExecProbs hash map that maps a componentID
     *        to its probability of executing the component
     * @param roundsLeft the number of rounds left
     * @return the initial sample sizes for the staistical simulations of
     *         components
     */
    public Map<Signature, BigInteger> computeSampleSizesInitially(HashMap<Signature, BigDecimal> mapExecProbs, int roundsLeft) {
    	// set all rations are 1
        HashMap<Signature, Integer> mapRatios = new HashMap<>();
        HashMap<Signature, Integer> mapSizesOfSupports = new HashMap<>();
        for(Signature sig : mapExecProbs.keySet()) {
            mapRatios.put(sig, 1);
            mapSizesOfSupports.put(sig, 1);
        }
    	return computeSampleSizesInitially(mapExecProbs, mapRatios, mapSizesOfSupports, roundsLeft);
    }
    
    /**
     * Computes the (initial) sample sizes for the staistical simulations of
     * components.
     *
     * @param mapExecProbs hash map that maps a componentID
     *        to its probability of executing the component
     * @param mapEsitmateNumOfObservables hash map that maps a componentID
     *        to the number of division of each signature
     * @param mapSizesOfSupports hash map that maps a componentID
     *        to the number of division of each signature
     * @param roundsLeft the number of rounds left
     * @return the initial sample sizes for the staistical simulations of
     *         components
     */
    public Map<Signature, BigInteger> computeSampleSizesInitially(HashMap<Signature, BigDecimal> mapExecProbs, HashMap<Signature, Integer> mapEsitmateNumOfObservables, HashMap<Signature, Integer> mapSizesOfSupports, int roundsLeft) {
        log("INFO(JPDM177): computeSampleSizesInitially is launched.", 3);
        // obtain array listComponentIDs
        listComponentIDs = mapExecProbs.keySet().toArray(new Signature[0]);
        log("Total budget of sample size: " + totalNoOfSim, 3);

        // obtain arrays listExecProbs and listSizesOfSupports
        int numOfComponents = listComponentIDs.length + 1;
        listExecProbs = new BigDecimal[numOfComponents];
        listExecProbs[0] = null;
        listEsitmateNumOfObservables = new int[numOfComponents];
        listEsitmateNumOfObservables[0] = 1; // TODO: to be corrected
        listSizesOfSupports = new int[numOfComponents];
        listSizesOfSupports[0] = 1;
        for (int i = 0; i < listComponentIDs.length; i++) {
        	listExecProbs[i + 1] = mapExecProbs.get(listComponentIDs[i]);
        	listEsitmateNumOfObservables[i + 1] = mapEsitmateNumOfObservables.get(listComponentIDs[i]);
            listSizesOfSupports[i + 1] = mapSizesOfSupports.get(listComponentIDs[i]);
        }
        // compute sample sizes
        BigInteger[] listSampleSizes = EstimateMI.initializeSampleSizes(totalNoOfSim, listExecProbs, listEsitmateNumOfObservables, listSizesOfSupports);
        // construct a hash map
        HashMap<Signature, BigInteger> mapSampleSizes = new HashMap<>();
        for (int i = 0; i < listComponentIDs.length; i++) {
            // divide the sample size by the number of rounds left
            if (listSampleSizes[i + 1] != null) {
                listSampleSizes[i + 1] = listSampleSizes[i + 1].divide(BigInteger.valueOf(roundsLeft));
            } else {
                listSampleSizes[i + 1] = BigInteger.ZERO;
            }
            // print
            if (listExecProbs[i + 1] != null) {
                log("listExecProbs[" + (i + 1) + "]", 3);
                log("listSampleSizes[" + (i + 1) + "] = " + listSampleSizes[i + 1], 3);
            } else {
                log("listExecProbs[" + (i + 1) + "] = null", 3);
            }
            mapSampleSizes.put(listComponentIDs[i], listSampleSizes[i + 1]);
        }
        return mapSampleSizes;
    }
    
    
    /**
     * Computes the sample sizes for the satistical simulations of components
     * adaptively from trial runs.
     *
     * @param roundsLeft the number of rounds left
     * @return the final sample sizes for the satistical simulations of
     * components
     */
    public HashMap<Signature, BigInteger> computeSampleSizesAdaptively(int roundsLeft) {
        log("INFO(JPDM226): computeSampleSizesAdaptively(" + roundsLeft + ") is launched.", 3);
        // check whether listComponentIDs was created
        if (listComponentIDs == null) {
            log("ERROR(JPDM229): listComponentIDs is null.", 1);
            System.exit(1);
        }
        // Check whether all necessary information is prepared
        checkAllInfoPrepared();

        // compute sample sizes
        BigInteger[] listSampleSizes = EstimateMI.updateSampleSizes(totalNoOfSim, usedSampleSizes, totalJointDist, distSecrets, distObservables, listSubJoints, listExecProbs);
    	//printArray(listSampleSizes, "listSampleSizes",2);

        // construct a hash map
        HashMap<Signature, BigInteger> mapSampleSizes = new HashMap<Signature, BigInteger>();
        for (int i = 0; i < listComponentIDs.length; i++) {
            BigInteger listNextSampleSizes = listSampleSizes[i + 1].divide(BigInteger.valueOf(roundsLeft)).max(BigInteger.ZERO);
            if (logger.getVerbosity() >= 3) {
                log("INFO(JPDM244): usedSampleSizes[" + (i + 1) + "]     = " + usedSampleSizes[i + 1], 3);
                log("INFO(JPDM245): listSampleSizes[" + (i + 1) + "]     = " + listSampleSizes[i + 1], 3);
                log("INFO(JPDM246): listNextSampleSizes[" + (i + 1) + "] = " + listNextSampleSizes, 3);
            }
            mapSampleSizes.put(listComponentIDs[i], listNextSampleSizes);
        }
        return mapSampleSizes;
    }

    /**
     * Computes the importance prior distributions.
     *
     * @return the array of importance prior distributions
     */
    public SingleDistribution[] computeImpPriorsInitially() {
        log("ERROR(JPDM259): WE DO NOT USE IMPORTANCE PRIOR ANY MORE.", 1);
        return null;
    }

    /////////////////////////////////////////////////////////////
    // Methods for storing joint sub-distributions, sub-priors,
    // and their total probabilities
    /**
     * Stores an exact joint sub-distribution (obtained by trace analysis), the
     * sub-prior and its total probability to this manager.
     *
     * @param nonNormSubJoint (non-normalized) joint sub-distribution for the
     * component
     */
    synchronized public void storeExactSubDistToManager(JointDistribution nonNormSubJoint) {
        log("Storing an exact sub distribution...", 3);
        if (this.exactSubJointDist != null) {
            //this.exactSubJointDist.print("Exact distribution before storing a sub distribution", 3);
        }
        //nonNormSubJoint.print("Being stored exact sub distribution", 3);
        if (nonNormSubJoint != null) {
            BigDecimal totalProbability = nonNormSubJoint.getSumOfProbabilities();
            if (this.exactSubJointDist == null) {
                this.exactSubJointDist = nonNormSubJoint;
            } else {
        		//nonNormSubJoint.print("nonNormSubJoint");
                //this.exactSubJointDist.print("exactSubJointDist(before)");
                this.exactSubJointDist.add(nonNormSubJoint);
            	//log("----->",1);
                //this.exactSubJointDist.print("exactSubJointDist(after)");
            }
            if (this.exactTotalProb == null) {
                this.exactTotalProb = totalProbability;
            } else {
                this.exactTotalProb = this.exactTotalProb.add(totalProbability);
            }
        }
        //this.exactSubJointDist.print("Exact distribution after storing a sub distribution", 3);
    }

    /**
     * Stores a joint frequency sub-distribution, the number of simulations, the
     * sub-prior and its total probability to this manager.
     *
     * @param componentID ID of the component
     * @param freqSubJoint joint frequency sub-distribution for the component
     * @param numOfSims the number of simulations for the component (This value
     * is zero if the sub-distribution comes from trace analysis.)
     * @param probabilityOfState the probability of executing the component
     * @param abstracted whether the abstraction-then-sampling technique has been used
     */
    synchronized public void storeFreqSubDistToManager(Signature componentID, JointFrequencyDistribution freqSubJoint, int numOfSims, BigDecimal probabilityOfState, boolean abstracted) {
        logger.changeVerbosity("JPDM_STORE1");
        if (logger.getVerbosity() > 2) {
            log("Storing frequency sub distributions (The number of simulations = " + numOfSims + ")", 3);
            freqSubJoint.print("Being stored frequency sub distribution", 3);
        }
        if (mapFreqSubJoints.containsKey(componentID)) {
            log("INFO(JPDM314): Updating the frequency sub-distribution", 4);
            // update the frequency sub-distribution
            JointFrequencyDistribution freqDist = mapFreqSubJoints.get(componentID);
            freqDist.add(freqSubJoint);
            mapFreqSubJoints.put(componentID, freqDist);
        } else {
            log("INFO(JPDM320): Storing the frequency sub-distribution", 4);
            // store the frequency sub-distribution
            JointFrequencyDistribution freqDist = new JointFrequencyDistribution(freqSubJoint);
            mapFreqSubJoints.put(componentID, freqDist);
        }

        if (mapUsedSampleSizes.containsKey(componentID)) {
            // update the sample size
            BigInteger usedSampleSize = mapUsedSampleSizes.get(componentID).add(BigInteger.valueOf(numOfSims));
            mapUsedSampleSizes.put(componentID, usedSampleSize);
        } else {
            // store the sample size
            BigInteger usedSampleSize = BigInteger.valueOf(numOfSims);
            mapUsedSampleSizes.put(componentID, usedSampleSize);
        }
    	//mapFreqSubJoints.get(componentID).print("Updated frequency sub distribution",3);
        // compute and store the joint probability sub-distribution for the joint frequency sub-distribution
        JointDistribution jointProbSubDist = computeProbSubDistFromFreqSubDist(mapFreqSubJoints.get(componentID), mapUsedSampleSizes.get(componentID), probabilityOfState, abstracted);
        //jointProbSubDist.print("jointProbSubDist-----",3);
        mapSubJoints.put(componentID, jointProbSubDist);
    	//mapSubJoints.get(componentID).print("Updated joint sub distribution",3);
        // store the total probability of the component
        mapTotalProbs.put(componentID, probabilityOfState);
        // setup lists to adaptively compupte sample sizes later
        if (listComponentIDs != null){
            updateLists(listComponentIDs.length + 1);
        }
        logger.restoreVerbosity();
    }

    /**
     * Computes the probability distribution from the frequency distribution.
     *
     * @param freqSubJoint joint frequency sub-distribution for the component
     * @param sampleSize the number of simulations for the component
     * @return joint probability distribution computed from frequency distribution
     * @param abstracted whether the abstraction-then-sampling technique has been used
     */
    public static JointDistribution computeProbSubDistFromFreqSubDist(JointFrequencyDistribution freqSubJoint, BigInteger sampleSize, BigDecimal probabilityOfState, boolean abstracted) {
        log("INFO(JPDM358): computeProbSubDistFromFreqSubDist is launched.", 3);
        // get elements of the freqency distribution
        Set<String> secretLabels = freqSubJoint.getSecretLabels();
        Set<String> observableLabels = freqSubJoint.getObservableLabels();
        Map<String, Map<String, Integer>> jointFreqMap = freqSubJoint.getJointFreqMap();
        if(secretLabels == null || observableLabels == null || jointFreqMap == null) {
            log("INFO(JPDM364): the given frequent sub-joint has a null component.", 1);
            return null;
        }
        logger.changeVerbosity("SUB_JOINT_DIST");
        if(logger.getVerbosity() >= 3) {
            freqSubJoint.print("freqSubJoint",1);
        }
        logger.restoreVerbosity();
        
        // compute the sum of frequencies when abstraction-then-sampling
        BigInteger sumOfFrequencies = BigInteger.ZERO;
        if(abstracted) {
        	int counter = 0;
            for(String secretLabel : secretLabels) {
                if(jointFreqMap.get(secretLabel) != null) {
                    for(String observableLabel : jointFreqMap.get(secretLabel).keySet()) {
                        int freq = jointFreqMap.get(secretLabel).get(observableLabel);
                        counter += freq;
                    }
                }
            }
        	sumOfFrequencies = BigInteger.valueOf(counter);
        } else {
        	sumOfFrequencies = sampleSize;
        }
        
        // compute the joint probability distribution
        Map<String, Map<String, BigDecimal>> jointProbMap = new LinkedHashMap<String, Map<String, BigDecimal>>();
        for(String secretLabel : secretLabels) {
            // total counts with this secret
        	/*
            int totalFreqWithSecret = 0;
            for(String observableLabel : jointFreqMap.get(secretLabel).keySet()) {
                totalFreqWithSecret += jointFreqMap.get(secretLabel).get(observableLabel);
            }
            log("totalFreqWithSecret[" + secretLabel + "] = " + totalFreqWithSecret, 4);
            */
            //log("sumOfFrequencies = "+sumOfFrequencies+"  probabilityOfState = "+probabilityOfState, 1);

            // compute the joint probability of having this secret and each observable
            Map<String, BigDecimal> observableProbMap = new LinkedHashMap<String, BigDecimal>();
            if(jointFreqMap.get(secretLabel) != null) {
                for(String observableLabel : jointFreqMap.get(secretLabel).keySet()) {
                    int freq = jointFreqMap.get(secretLabel).get(observableLabel);
                    //log(" freq[" + secretLabel + ", " + observableLabel + "] = " + freq, 1);
                    BigDecimal prob = probabilityOfState.multiply((new BigDecimal(freq)), contexter.getContext());
                    prob = prob.divide((new BigDecimal(sumOfFrequencies)), contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP); //TODO: +10?
                    //log(" prob[" + secretLabel + ", " + observableLabel + "] = " + prob, 1	);
                    observableProbMap.put(observableLabel, prob);
                }
                jointProbMap.put(secretLabel, observableProbMap);
            }
        }
        JointDistribution jointProbSubDist = new JointDistribution(secretLabels, observableLabels, jointProbMap);
        //jointProbSubDist.print("jointProbSubDist",1);
        return jointProbSubDist;
    }

    /**
     * Stores a joint frequency distribution and the number of simulations
     * to this manager.
     *
     * @param componentID ID of the component
     * @param freqSubJoint joint frequency distribution
     * @param numOfSims the number of simulations
     * @param execProb the probability of executing the component
     */
    synchronized public void storeFreqDistFromMatrixToManager(Signature componentID, JointFrequencyDistribution freqSubJoint, int numOfSims, BigDecimal execProb) {
        storeFreqSubDistToManager(componentID, freqSubJoint, numOfSims, execProb, false);
    }

    /////////////////////////////////////////////////////////////
    // Methods for constructing the joint distribution
    /*
     * Convert mapSubJoints to listSubJoints[],
     * mapUsedSampleSizes to usedSampleSizes[],
     * mapTotalProbs to listExecProbs[],
     * for using existing methods.
     */
    synchronized private void setupLists() {
        log("INFO(JPDM440): setupLists is launched.", 3);
        int numOfComponents = 1;
        if(listComponentIDs == null) {
            log("CAUTION(JPDM443): listComponentIDs is null.", 2);
        } else {
            numOfComponents = listComponentIDs.length + 1;
        }
        if(exactSubJointDist == null || exactTotalProb == null) {
            log("CAUTION(JPDM448): exactSubJointDist or exactTotalProb is null.", 2);
        } else {
            //exactSubJointDist.print("exactSubJointDist", 3);
            //log("exactTotalProb = "+exactTotalProb, 3);
        }
        if(mapSubJoints == null || mapUsedSampleSizes == null || mapTotalProbs == null) {
            log("ERROR(JPDM453): The array list mapSubJoints or mapUsedSampleSizes or mapTotalProbs is null.", 1);
            System.exit(1);
        }
        updateLists(numOfComponents);
    }

    /*
     * Convert mapSubJoints to listSubJoints[],
     * mapUsedSampleSizes to usedSampleSizes[],
     * mapTotalProbs to listExecProbs[],
     * for using existing methods.
     * 
     * @param numOfComponents the number of components
     */
    synchronized private void updateLists(int numOfComponents) {
        // array of subJoints, usedSampleSizes, sub-priors
        listSubJoints = new JointDistribution[numOfComponents];
        listSubJoints[0] = exactSubJointDist;
        usedSampleSizes = new BigInteger[numOfComponents];
        usedSampleSizes[0] = BigInteger.ZERO;
        listExecProbs = new BigDecimal[numOfComponents];
        listExecProbs[0] = exactTotalProb;

        // from statistical analyses
        for (int i = 0; i < numOfComponents - 1; i++) {
            listSubJoints[i + 1] = mapSubJoints.get(listComponentIDs[i]);
            usedSampleSizes[i + 1] = mapUsedSampleSizes.get(listComponentIDs[i]);
            listExecProbs[i + 1] = mapTotalProbs.get(listComponentIDs[i]);
        }
    }

    /**
     * Computes the (posterior) joint distribution from joint sub-distributions.
     * (Note: listSubJoints[], usedSampleSizes[] and listSubPriors[] are
     * prepared here.)
     *
     * @return the (posterior) joint distribution
     */
    public JointDistribution computeTotalJointDistribution() {
        // prepare listSubJoints[], usedSampleSizes[] and listSubPriors[]
        log("INFO(JPDM494): computeTotalJointDistribution is launched.", 3);
        setupLists();
        // the number of components
        int numOfComponents = listSubJoints.length;
		//System.out.println("numOfComponents = " + numOfComponents);
        // sum the joint sub-distributions to obtain the total joint
        //listSubJoints[0].print("listSubJoints["+0+"]",2);
        totalJointDist = new JointDistribution(listSubJoints[0]);
        for (int i = 1; i < numOfComponents; i++) {
            /*
             if(listSubJoints[i] != null) {
             listSubJoints[i].print("listSubJoints["+i+"]",2);
             } else {
             log("CAUTION(JPDM507): listSubJoints["+i+"] is null.",2);
             }
             */
            totalJointDist.add(listSubJoints[i]);
            /*
             if(totalJointDist != null) {
             totalJointDist.print("totalJointDist",2);
             } else {
             log("CAUTION(JPDM515): totalJointDist is null.",2);
             }
             */
        }
        // print the total joint distribution
        logger.changeVerbosity("TOTAL_DIST");
        if(logger.getVerbosity() >= 3) {
        	totalJointDist.print("totalJointDist",2);
        }
        logger.restoreVerbosity();

        return totalJointDist;
    }

    /**
     * Computes and stores the (posterior) joint distribution from joint
     * sub-distributions if it is not generated. Returns the stored distribution
     * otherwise.
     *
     * @return the (posterior) joint distribution
     */
    public JointDistribution getTotalJointDistribution() {
        if (totalJointDist == null) {
            totalJointDist = computeTotalJointDistribution();
        }
        return totalJointDist;
    }

    /////////////////////////////////////////////////////////////
    // Methods for computing an empirical mutual inforamtion
    // after obtaining the joint total distribution.
    /**
     * Computes the posterior Shannon entropy (without any assumption on the
     * prior distribution).
     *
     * @return the posterior Shannon entropy
     */
    public BigDecimal computePosteriorShannonEntropy() {
        if (totalJointDist == null) {
            log("ERROR(JPDM547): Cannot compute the posterior Shannon entropy. The total joint distribution is null.", 1);
            return null;
        }
    	distObservables = computeDistributionOnObservables();

        // get the domains of secrets and observables
        Set<String> secretLabels = totalJointDist.getSecretLabels();
        Set<String> observableLabels = totalJointDist.getObservableLabels();
        // compute the posterior Shannon entropy
        BigDecimal result = BigDecimal.ZERO;
        for (String observableLabel : observableLabels) {
            //totalJointDist.print("totalJointDist::");
            //distObservables.print("distObservables::");
            //log("observableLabel = "+observableLabel,3);
            BigDecimal obsProb = distObservables.getProbability(observableLabel);
            for (String secretLabel : secretLabels) {
                //log("(sec, obs) = ("+secretLabel + ", " + observableLabel+")",1);
                BigDecimal jointProb = totalJointDist.getProbability(secretLabel, observableLabel);
                if (jointProb.compareTo(BigDecimal.ZERO) != 0) {
                    BigDecimal jointEntropy = jointProb.multiply(BigFunctions.log2(jointProb.divide(obsProb, contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP), contexter.getContext().getPrecision())); //TODO: +10?
                    result = result.add(jointEntropy);
                }
            }
        }
        PosteriorEntropyPointEstimate = result.negate();
        return PosteriorEntropyPointEstimate;
    }

    /**
     * Computes the posterior Shannon entropy (when the prior distribution is
     * uniform).
     *
     * @return the posterior Shannon entropy
     */
    public BigDecimal computePosteriorShannonEntropyWithUniformPrior() {
        if (totalJointDist == null) {
            log("ERROR(JPDM583): Cannot compute the posterior Shannon entropy. The total joint distribution is null.", 1);
            System.exit(1);
        }
        distSecrets = getDistributionOnSecrets();
        distObservables = getDistributionOnObservables();

        // get the domains of secrets and observables
        String[] secretLabelsNonZeroProbabilities = priorDistMan.getInitialSecretValues();
        Set<String> observableLabels = totalJointDist.getObservableLabels();
        // compute the posterior Shannon entropy
        BigDecimal result = BigDecimal.ZERO;
        for (String observableLabel : observableLabels) {
            BigDecimal obsProb = distObservables.getProbability(observableLabel);
            for (String secretLabel : secretLabelsNonZeroProbabilities) {
                BigDecimal jointProb = totalJointDist.getProbability(secretLabel, observableLabel);
                if (jointProb.compareTo(BigDecimal.ZERO) != 0) {
                    BigDecimal jointEntropy = jointProb.multiply(BigFunctions.log2(jointProb.divide(obsProb, contexter.getContext().getPrecision(), BigDecimal.ROUND_HALF_UP), contexter.getContext().getPrecision())); //TODO: +10?
                    result = result.add(jointEntropy);
                }
            }
        }
        //log("INFO(JPDM604): result = "+result,1);
        PosteriorEntropyPointEstimate = result.negate();
        return PosteriorEntropyPointEstimate;
    }

    /**
     * Computes the mutual information point estimate.
     *
     * @return the posterior Shannon entropy
     */
    public BigDecimal computeMIPointEstimate() {
        if (priorDistMan != null) {
            BigDecimal priorEntropy = priorDistMan.computePriorShannonEntropy();
            if (PosteriorEntropyPointEstimate == null) {
                PosteriorEntropyPointEstimate = computePosteriorShannonEntropyWithUniformPrior();
            }
            MIPointEstimate = priorEntropy.subtract(PosteriorEntropyPointEstimate);
            return MIPointEstimate;
        } else if (priorDistribution != null) {
            BigDecimal priorEntropy = priorDistribution.getEntropy(); // TODO: to be supported
            if (PosteriorEntropyPointEstimate == null) {
                PosteriorEntropyPointEstimate = computePosteriorShannonEntropy();
            }
            MIPointEstimate = priorEntropy.subtract(PosteriorEntropyPointEstimate);
            return MIPointEstimate;
        } else {
            log("CAUTION(JPDM630): No prior is found. We estimate the prior:", 3);
            priorDistribution = getDistributionOnSecrets();
            BigDecimal priorEntropy = priorDistribution.getEntropy(); // TODO: to be supported
            if (PosteriorEntropyPointEstimate == null) {
                PosteriorEntropyPointEstimate = computePosteriorShannonEntropy();
            }
            MIPointEstimate = priorEntropy.subtract(PosteriorEntropyPointEstimate);
            return MIPointEstimate;
        }
    }

    /////////////////////////////////////////////////////////////
    // Methods for computing the distribution on secrets
    // after obtaining the joint total distribution.
    /**
     * Computes the probability distribution on secrets.
     *
     * @return the probability distribution on secrets
     */
    SingleDistribution computeDistributionOnSecrets() {
        if (totalJointDist == null) {
            totalJointDist = computeTotalJointDistribution();
        }
        distSecrets = totalJointDist.getProjectionSecrets();
        // print the distribution on secrets
        logger.changeVerbosity("DIST_SEC");
        if(logger.getVerbosity() >= 3) {
        	distSecrets.print("distSecrets",2);
        }
        logger.restoreVerbosity();
        return distSecrets;
    }

    /**
     * Computes and stores the probability distribution on secrets if it is
     * not generated. Returns the stored distribution otherwise.
     */
    public SingleDistribution getDistributionOnSecrets() {
        if (distSecrets == null) {
            distSecrets = computeDistributionOnSecrets();
        }
        return distSecrets;
    }

    /////////////////////////////////////////////////////////////
    // Methods for computing the distribution on observables
    // after obtaining the joint total distribution.
    /**
     * Computes the probability distribution on observables.
     *
     * @return the probability distribution on observables
     */
    SingleDistribution computeDistributionOnObservables() {
        if (totalJointDist == null) {
            totalJointDist = computeTotalJointDistribution();
        }
        distObservables = totalJointDist.getProjectionObservables();
        // print the distribution on observables
        logger.changeVerbosity("DIST_OBS");
        if(logger.getVerbosity() >= 3) {
        	distObservables.print("distObservables",2);
        }
        logger.restoreVerbosity();
        return distObservables;
    }

    /**
     * Computes and stores the probability distribution on observables if it is
     * not generated. Returns the stored distribution otherwise.
     */
    public SingleDistribution getDistributionOnObservables() {
        if (distObservables == null) {
            distObservables = computeDistributionOnObservables();
        }
        return distObservables;
    }

    /////////////////////////////////////////////////////////////
    // Methods for computing the sub-priors
    private SingleDistribution[] computeSubPriors() { 
		// compute listSubPriors from listSubJoints
		listSubPriors = new SingleDistribution[listSubJoints.length];
		for(int i = 0; i < listSubJoints.length; i++) {
			listSubPriors[i] = listSubJoints[i].getProjectionSecrets();
			// print size properties
            logger.changeVerbosity("PRINT_SAMPLE_SIZE");
            if (logger.getVerbosity() >= 2) {
    			listSubJoints[i].getSizeProperties();
            }
	        logger.restoreVerbosity();
		}
		return listSubPriors;
    }
    
    /////////////////////////////////////////////////////////////
    // Methods for computing the biases and confidence intervals
    // of mutual information estimates after obtaining the joint
    // total distribution and empirical estimates.

    /*
     * Checks whether all information necessary to compute bias
     * and confidence intervals are prepared.
     */
    private void checkAllInfoPrepared() {
        log("INFO(JPDM716): checkAllInfoPrepared is launched.", 3);
        // Check the list of joint sub-distributions
        if (listSubJoints == null || usedSampleSizes == null || listExecProbs == null) {
            setupLists();
        }
        // Check the joint (total-)distribution
        totalJointDist = getTotalJointDistribution();
        //totalJointDist.print("totalJointDist");
        // Check the distribution on secrets
        distSecrets = getDistributionOnSecrets();
        //distSecrets.print("distSecrets");
        // Check the distribution on observables
        distObservables = getDistributionOnObservables();
        //distObservables.print("distObservables");
    }
    

    /**
     * Computes the bias of estimated mutual information under a know prior.
     *
     * @return the bias of estimated mutual information
     */
    public BigDecimal computeBiasOfMIPointEstimate() {
        // Check whether all necessary information is prepared
        checkAllInfoPrepared();
		// Compute listSubPriors[] from listSubJoints[]
    	listSubPriors = computeSubPriors();
        // Compute the bias of the MI estimate
        MIPointEstimateBias = EstimateMI.getBiasOfEstimatedMI(totalJointDist, distSecrets, distObservables, usedSampleSizes, listSubJoints, listExecProbs, listSubPriors, priorIsKnown);
        return MIPointEstimateBias;
    }

    /**
     * Computes the estimated mutual information under a know prior. after the
     * correction of bias.
     *
     * @return the estimated mutual information after the correction of bias
     */
    public BigDecimal computeCorrectedMIPointEstimate() {
        // Compute the MI point estimate
        if (MIPointEstimate == null) {
            MIPointEstimate = computeMIPointEstimate();
        }
        // Compute the bias of the MI estimate
        if (MIPointEstimateBias == null) {
            MIPointEstimateBias = computeBiasOfMIPointEstimate();
        }
        // Compute the MI estimate after correction of bias
        MICorrectedPointEstimate = MIPointEstimate.subtract(MIPointEstimateBias).max(BigDecimal.ZERO);
        return MICorrectedPointEstimate;
    }

    /**
     * Computes the confidence interval of the estimated mutual information
     * under a know prior.
     *
     * @return the confidence interval of the estimated mutual information
     */
    public BigDecimal[] computeConfidenceIntervalOfMI() {
        // Check whether all necessary information is prepared
        if (MICorrectedPointEstimate == null) {
            MICorrectedPointEstimate = computeCorrectedMIPointEstimate();
        }
        confidenceInterval = EstimateMI.getCorrectedMIConfidenceInterval(MICorrectedPointEstimate, totalJointDist, distSecrets, distObservables, usedSampleSizes, listSubJoints, listExecProbs, listSubPriors, priorIsKnown);
        return confidenceInterval;
    }

    /////////////////////////////////////////////////////////////
    /**
     * Computes the sum of all probabilities in the current joint distribution.
     *
     * @return the sum of all probabilities in the current joint distribution
     */
    BigDecimal getSumOfProbabilities() {
        BigDecimal sum = BigDecimal.ZERO;
        for (Signature componentID : mapTotalProbs.keySet()) {
            sum = sum.add(mapTotalProbs.get(componentID));
        }
        return sum;
    }

    private static void log(Object aObject, int level) {
        if (logger.getVerbosity() >= level) {
            Logger loggy = Logger.getInstance();
            loggy.log(String.valueOf(aObject), level);
        }
    }
}
