/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/
    
package hyleak;

import java.util.Iterator;
import java.math.BigDecimal;
import java.math.BigInteger;
//import java.util.ArrayList;
//import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.Map;

import distributions.JointDistribution;
import distributions.JointFrequencyDistribution;
import distributions.MapDistribution;
import support.Contexter;
import support.Logger;

/**
 *
 * @author  
 * @author  
 */
public class MarkovChainState {
    private static final Contexter contexter = Contexter.getInstance();
    private static final Logger logger = Logger.getInstance();
    
    private String name = "";
    private BigDecimal initialProbability = contexter.getDecimalZero();
//    private BigDecimal localEntropy = new BigDecimal(-1.0,contexter.getContext());
    //private BigDecimal reward;
    private BigDecimal probability;
    //private boolean recurrent = false;
    private VariableList variables;
//    private final MarkovStateSet predecessors;
//    private final MarkovStateSet successors;
//    private final HashSet<Transition> transitionsFrom;
    
    

    MarkovChainState(String name, BigDecimal initprob) {
        //this.reward = contexter.getDecimalZero();
        this.probability = contexter.getDecimalOne();
        this.name = name;
        this.initialProbability = initprob;
        variables = new VariableList();
//        predecessors = new MarkovStateSet();
//        successors = new MarkovStateSet();
//        transitionsFrom = new HashSet<>();
    }
    
    public BigDecimal getProbability() {
        return probability;
    }

    public void setProbability(BigDecimal probability) {
        this.probability = probability;
    }
    
    public void multiplyProbabilityBy(BigDecimal tomult){
        this.probability = this.probability.multiply(tomult, contexter.getContext());
    }
    
    public void addVariable(Variable var){
        this.variables.add(var);
        
    }
    
    public boolean setPublicVariable(String varname, BigInteger newval){
        return this.variables.setPublicVariable(varname, newval);
    }
    
    public BigInteger getPublicVariableValue(String varname){
        return this.variables.getPublicVariableValue(varname);
    }
    
    public Interval getPrivateVariableInterval(String varname){
        return this.variables.getPrivateVariableInterval(varname);
    }
    
    public boolean deleteVariable(Variable var){
        return this.variables.deleteVariable(var);
    }

    public VariableList getVariables() {
        return variables;
    }

    public void setVariables(VariableList variables) {
        this.variables = variables;
    }
    
    public boolean deleteVariableByName(String name){
        return this.variables.deleteVariableByName(name);
    }
    
    public Variable getVariable(String name){
        return this.variables.getVariable(name);
    }

    public BigDecimal getInitialProbability() {
        return initialProbability;
    }

    public void setInitialProbability(BigDecimal initialProbability) {
        this.initialProbability = initialProbability;
    }

//    public BigDecimal getLocalEntropy() {
//        return localEntropy;
//    }
//
//    public void setLocalEntropy(BigDecimal localEntropy) {
//        this.localEntropy = localEntropy;
//    }

//    public BigDecimal getReward() {
//        return reward;
//    }
//  
//    public void setReward(BigDecimal rew){
//        this.reward = rew;
//    }
    
    public String getName() {
        return name;
    }

    public void setName(String name) {
        this.name = name;
    }

//    public boolean isRecurrent() {
//        return recurrent;
//    }
//
//    public void setRecurrent(boolean recurrent) {
//        this.recurrent = recurrent;
//    }

    

//    public MarkovStateSet getPredecessors() {
//        return predecessors;
//    }
//
//    public void addPredecessor(MarkovChainState predecessor) {
//        this.predecessors.add(predecessor);
//    }
//
//    public void removePredecessor(MarkovChainState predecessor) {
//        this.predecessors.remove(predecessor);
//    }
//
//    public MarkovStateSet getSuccessors() {
//        return successors;
//    }
//
//    public void addSuccessor(MarkovChainState successor) {
//        this.successors.add(successor);
//    }
//
//    public void removeSuccessor(MarkovChainState successor) {
//        this.successors.remove(successor);
//    }
//    
//    public void addTransition(Transition trans){
//        this.transitionsFrom.add(trans);
//    }
//
//    public void removeTransition(Transition trans){
//        this.transitionsFrom.remove(trans);
//    }
    
    

    private static void log(Object aObject, int level){
        logger.log(String.valueOf(aObject),level);
    }
  
    @Override
    public String toString(){
        return this.name;
    }
   
    public String printNice(){
        String s = this.getName();
        if (this.variables.isEmpty()==false){
            s += "{";
            Iterator<Variable> varit = variables.iterator();
            while (varit.hasNext()){
                Variable var =varit.next();
                s += var.toString();
                
                if (varit.hasNext()){
                    s += ",";
                }
            }
            s += "}";
        }
        return s;
    }

//    public HashSet<Transition> getTransitionsFrom() {
//        return this.transitionsFrom;
//    }
    
//    public void printToGraph(PrintWriter out) {
//        String varstring = "";
//        for (MarkovStateVariable var : variables) {
//            varstring = varstring + var.toString() + "\\n";
//        }
//
//        out.println("\t\"" + this.getName() + "\" [" + "label = \"<f0> " + this.getName() + " | <f1> " + varstring + "\" shape = \"record\"];");
//        for (Transition t : transitionsFrom) {
//            if (!t.getSource().equals(t.getDest())) {
//                out.println("\t\"" + this.getName() + "\":f0 -> \"" + t.getDest() + "\":f0 [label = \"" + t.getValue() + "\"];");
//            }
//        }
//    }

    boolean setPrivateVariable(String varname, String interval) {
        return this.variables.setPrivateVariable(varname, interval);
    }
    
 
    @Override
    public MarkovChainState clone() throws CloneNotSupportedException{
        MarkovChainState copy = new MarkovChainState(this.name,this.initialProbability);
        copy.setVariables(variables.clone());
        copy.setProbability(probability);
        return copy;
    }

//    Transition getTransitionToState(MarkovChainState dest) {
//        for (Transition trans : this.transitionsFrom) {
//            if (trans.getDest().equals(dest.getName())) {
//                return trans;
//            }
//        }
//        return new Transition(this.name,dest.getName(),contexter.getDecimalZero());
//    }

//    void addToReward(BigDecimal toadd) {
//        this.reward = this.reward.add(toadd, contexter.getContext());
//    }
        
    public VariableList getCloneOfVariables(){
        return this.variables.clone();
    }

    void setDistributionVariable(String varname, BigInteger value) {
        this.variables.getDistributionVariable(varname).setDistribution(new MapDistribution(value));
    }
    
    void setDistributionVariable(String varname, MapDistribution distrib) {
        this.variables.getDistributionVariable(varname).setDistribution(distrib);
    }
    
    void setDistributionVariable(String varname, BigInteger lowbound, BigInteger uppbound) {
        this.variables.getDistributionVariable(varname).setDistribution(new MapDistribution(lowbound,uppbound));
    }

    void setDistributionVariable(String varname, BigInteger[] values, BigDecimal[] probs) {
        this.variables.getDistributionVariable(varname).setDistribution(new MapDistribution(values,probs));
    }
    
//    void setCompressedDistributionVariable(String varname, BigInteger value) {
//        this.variables.getDistributionVariable(varname).setDistribution(new CompressedDistribution(value));
//    }
//    
//    void setDistributionVariable(String varname, HybridDistribution distrib) {
//        this.variables.getDistributionVariable(varname).setDistribution(distrib);
//    }
//    
//    void setDistributionVariable(String varname, BigInteger lowbound, BigInteger uppbound) {
//        this.variables.getDistributionVariable(varname).setDistribution(new HybridDistribution(lowbound,uppbound));
//    }
//
//    void setDistributionVariable(String varname, BigInteger[] values, BigDecimal[] probs) {
//        this.variables.getDistributionVariable(varname).setDistribution(new HybridDistribution(values,probs));
//    }

//    void setCompressedDistributionVariable(String varname, CompressedDistribution varvalue) {
//        throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
//    }

    String getSecretIdentifier(){
        VariableList secrets = new VariableList();
        for (Variable var : variables){
            if (var.isPrivate()){
                
                PrivateVariable privvar =(PrivateVariable)var;
                if (privvar.isSecret()){
                    secrets.add(privvar);    
                }
            }
        }
        return secrets.getIdentifier();
    }
    
    String getSecretSingleIdentifier(){
        VariableList secrets = new VariableList();
        for (Variable var : variables){
            if (var.isPrivate()){
                
                PrivateVariable privvar =(PrivateVariable)var;
                if (privvar.isSecret()){
                    secrets.add(privvar);    
                }
            }
        }
        if (logger.getVerbosity()>=4){
            log("Returning single secret identifier "+secrets.getSingleIdentifier(),4);
        }
        
        return secrets.getSingleIdentifier();
    }
    
    String getObservableIdentifier(){
        VariableList observables = new VariableList();
        for (Variable var : variables){
            if (!var.isPrivate()){
                PublicVariable pubvar =(PublicVariable)var;
                if (pubvar.isObservable()){
                    observables.add(pubvar);            
                }
            }
        }
        if (logger.getVerbosity()>=4){
            log("Returning observable identifier "+observables.getIdentifier(),4);
        }
        return observables.getIdentifier();
    }
    
    JointFrequencyDistribution getEmptyJointFrequencyDistribution(){
        VariableList secrets = new VariableList();
        VariableList observables = new VariableList();
        
        int secretsize = 1;
        HashMap<String,Interval> secretinters = new HashMap();
        
        for (Variable var : variables){
            if (var.isPrivate()){
                
                PrivateVariable privvar =(PrivateVariable)var;
                if (privvar.isSecret()){
                    secrets.add(privvar);         
                    secretsize *= privvar.getIntervalSize().intValue();
                    secretinters.put(privvar.getName(), privvar.getInterval());
                }
                
                //newsecretsize = newsecretsize.add(asecret.getSize());
            }else{
                PublicVariable pubvar =(PublicVariable)var;
                if (pubvar.isObservable()){
                    observables.add(pubvar);            
                }
            }
        }
        
        log("",3);
        log("Observation: "+observables.getIdentifier()+"\nSecret: "+secrets.getIdentifier(),3);
        String[] obslabels = new String[1];
        obslabels[0] = observables.getIdentifier();
        
        String seclabels[] = new String[secretsize];
        IntervalAssignments secAssign = new IntervalAssignments(secretinters);
        seclabels = secAssign.getAllValues();
        
        log("possible secret values: "+secretsize,3);
        if (logger.getVerbosity()>=4){
            log("secret labels:",4);
            for (String seclabel : seclabels) {
                log("  -" + seclabel, 4);
            }
            log("observable labels:",4);
            for (String obslabel : obslabels) {
                log("  -" + obslabel, 4);
            }
        }
        Map<String,Map<String,Integer>> jointProbMap = new LinkedHashMap();
        
        for(int i = 0; i < secretsize; i++){
            Map<String,Integer> innermap = new LinkedHashMap();
            innermap.put(obslabels[0], 0);
            jointProbMap.put(seclabels[i], innermap);
        }
        
        JointFrequencyDistribution normalizedsubjointdist = new JointFrequencyDistribution(seclabels,obslabels,jointProbMap);
        return normalizedsubjointdist;
    }
    
    JointDistribution getJointSecretObservablesDistribution() {
        VariableList secrets = new VariableList();
        VariableList observables = new VariableList();
        
        int secretsize = 1;
        HashMap<String,Interval> secretinters = new HashMap();
        
        for (Variable var : variables){
            if (var.isPrivate()){
                
                PrivateVariable privvar =(PrivateVariable)var;
                if (privvar.isSecret()){
                    secrets.add(privvar);         
                    secretsize *= privvar.getIntervalSize().intValue();
                    secretinters.put(privvar.getName(), privvar.getInterval());
                }
                
                //newsecretsize = newsecretsize.add(asecret.getSize());
            }else{
                PublicVariable pubvar =(PublicVariable)var;
                if (pubvar.isObservable()){
                    observables.add(pubvar);            
                }
            }
        }
        
        log("",3);
        log("Observation: "+observables.getIdentifier()+"\nSecret: "+secrets.getIdentifier(),3);
        String[] obslabels = new String[1];
        obslabels[0] = observables.getIdentifier();
        
        String seclabels[] = new String[secretsize];
        IntervalAssignments secAssign = new IntervalAssignments(secretinters);
        seclabels = secAssign.getAllValues();
        
        log("possible secret values: "+secretsize,3);
        if (logger.getVerbosity()>=4){
            log("secret labels:",4);
            for (String seclabel : seclabels) {
                log("  -" + seclabel, 4);
            }
            log("observable labels:",4);
            for (String obslabel : obslabels) {
                log("  -" + obslabel, 4);
            }
        }
        Map<String,Map<String,BigDecimal>> jointProbMap = new LinkedHashMap();
        
        BigDecimal eachprob = (contexter.getDecimalOne().divide(new BigDecimal(secretsize), contexter.getContext())).multiply(probability, contexter.getContext());
        for(int i = 0; i < secretsize; i++){
            Map<String,BigDecimal> innermap = new LinkedHashMap();
            innermap.put(obslabels[0], eachprob);
            jointProbMap.put(seclabels[i], innermap);
        }
        
        JointDistribution weightedsubjointdist = new JointDistribution(seclabels,obslabels,jointProbMap);
        return weightedsubjointdist;
    }

    SimulationSignature getSimulationSignature(boolean isAbstract) {

        VariableList privList = new VariableList();
        VariableList obsList = new VariableList();
        
        
        for (Variable var : variables){
            if (var.isPrivate()){
                privList.add(var);
                
                
                //newsecretsize = newsecretsize.add(asecret.getSize());
            }else{
                PublicVariable pubvar =(PublicVariable)var;
                if (pubvar.isObservable()){
                    obsList.add(pubvar);            
                }
            }
        }
        return new HyleakSimulationSignature(this,obsList,privList,isAbstract);
    }
   
}
