/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package hyleak;

import bham.leakiest.ProbDist;
import distributions.JointDistribution;
import distributions.JointFrequencyDistribution;
import distributions.SingleDistribution;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import support.Contexter;
import support.Logger;

/**
 *
 * @author  
 */
class MatrixComputator {

    private static final Contexter contexter = Contexter.getInstance();
    private static final Logger logger = Logger.getInstance();

    private final int totSims;
    private final int rounds = 10;

    MatrixComputator(int iterations) {
        totSims = iterations;
    }

    void computeMatrices(String filename) {
        /*
         The filename points to the master file of the matrices. The master 
         file contains on the first line the file name of the prior, and on 
         each subsequent line the filename of the prior and the filename
         of the joint distribution, separated by a space.
         */

        MasterMatrixReader mmr = new MasterMatrixReader(filename);
        SingleDistribution prior = mmr.getInitialPrior();
        DistributionPriorDistributionManager priorDistMan = new DistributionPriorDistributionManager("h",prior);
        JointPosteriorDistributionManager postDistMan = new JointPosteriorDistributionManager(priorDistMan, BigInteger.valueOf(totSims));
        HashSet<MatrixSignature> signatures = new HashSet();
        HashMap<Signature, BigDecimal> mapExecProbs = new HashMap();
        for (Iterator<MatrixSignature> iter = mmr.iterator(); iter.hasNext();) {
            MatrixSignature sig = iter.next();
            signatures.add(sig);
            mapExecProbs.put(sig, sig.getProbability());
        }
        Map<Signature, BigInteger> simNumber = postDistMan.computeSampleSizesInitially(mapExecProbs, rounds);
        for (int i = 0; i < rounds; i++) {
            log("\n-------------------------------------------", 3);
            log("Starting simulation round " + (i + 1) + " of " + rounds, 3);
            for (MatrixSignature sig : signatures) {
                JointFrequencyDistribution freqDist = support.SimulatorOfJointDistribution.simulateJointDistribution(simNumber.get(sig).intValue(), sig.getChannel());
                postDistMan.storeFreqDistFromMatrixToManager(sig, freqDist, simNumber.get(sig).intValue(), sig.getWeight());
            }
            if (i < rounds - 1) {
                int roundsLeft = rounds - i - 1;
                simNumber = postDistMan.computeSampleSizesAdaptively(roundsLeft);
            }
        }

        // Computes the joint (total) distribution
        JointDistribution jointTotalDist = postDistMan.getTotalJointDistribution();
        if (logger.getVerbosity()>3){
            // TODO: print the jointTotalDist?
        }
        
        BigDecimal priorentropy = priorDistMan.computePriorShannonEntropy();
        
        // Computes the posterial Shannon entropy
        BigDecimal postentropy = postDistMan.computePosteriorShannonEntropyWithUniformPrior();
        log("\nThe prior Shannon entropy of the secret is "+priorentropy,2);
        log("The posterior Shannon entropy is "+postentropy,2);
        //postdistman.computeConfidenceIntervalOfMI();
        
        // Compute the mutual information
        BigDecimal MI = priorentropy.subtract(postentropy,contexter.getContext());
        if (logger.getVerbosity()>1){
            log("The Shannon leakage (with possible bias) is "+MI+" bits",1);
        }else{
            log(MI,0);
        }
        
        // Compute the corrected mutual information
        BigDecimal correctedMI = postDistMan.computeCorrectedMIPointEstimate();
        if (logger.getVerbosity()>0){
            log("The Shannon leakage (after bias correction) is "+correctedMI+" bits",1);
        }else{
            log(correctedMI,0);
        }

        // Compute the confidence interval of the estimated mutual information
        BigDecimal[] confidenceInterval = postDistMan.computeConfidenceIntervalOfMI();
        if (logger.getVerbosity()>0){
            log("The confidence interval is ["+confidenceInterval[0]+", "+confidenceInterval[1]+"]",1);
        }else{
            log("["+confidenceInterval[0]+", "+confidenceInterval[1]+"]",0);
        }

    }

    private static void log(Object aObject, int level) {
        logger.log(String.valueOf(aObject), level);
    }
}
