/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package hyleak;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.HashMap;
import support.Contexter;
import support.Logger;
import theory.BigFunctions;

/**
 *
 * @author  
 */
public abstract class PriorDistributionManager {
    protected static final Contexter contexter = Contexter.getInstance();
    protected static final Logger logger = Logger.getInstance();
    
    //private ArrayList<String> codelines = null;
    protected final HashMap<String,Interval> initialsecretintervals = new HashMap();
    protected BigInteger secretsize = new BigInteger("-1");
    protected String[] initialSecretValues = null;


    protected PriorDistributionManager() {

    }
    
    
    public BigInteger getSecretSize(){
        if (secretsize.compareTo(BigInteger.ZERO)<0){
            this.analyze();
        }
        return secretsize;
    }
    
    protected abstract void analyze();
    
    public BigDecimal computePriorShannonEntropy() {
       
        if (secretsize.compareTo(BigInteger.ZERO)<0){
            this.analyze();
        }
        BigDecimal decsecretsize = new BigDecimal(secretsize);
        BigDecimal priorentropy = log2(decsecretsize);
        log("Secret size as decimal: value "+decsecretsize.toPlainString()+", scale: "+decsecretsize.scale()+", base 2 logarithm "+priorentropy.toPlainString(),4);
         
        return priorentropy;
    }

    /**
     * Computes the hash map that stores the initial secret interval of each variable.
     * @return the hash map that stores the initial secret interval of each variable
     */
    HashMap<String, Interval> getInitialSecretIntervals() {
        if (initialsecretintervals == null){
            this.analyze();
        }
        return this.initialsecretintervals;
    }
    
    /**
     * Computes the string array of all secrets that have non-zero probabilities.
     * @return the string array of all secrets that have non-zero probabilities
     */
    String[] getInitialSecretValues(){
    	if(initialSecretValues == null) {
            if (initialsecretintervals == null){
                this.analyze();
            }
            IntervalAssignments interAss = new IntervalAssignments(this.initialsecretintervals);
            initialSecretValues = interAss.getAllValues();
    	}
        return initialSecretValues;
    }

    
    private BigDecimal log2(BigDecimal value) {
        if (value.compareTo(contexter.getDecimalOne())==0){
            return contexter.getDecimalZero();
        }
        BigDecimal toreturn = BigFunctions.ln(value,contexter.getContext().getPrecision());
        toreturn = toreturn.divide(BigFunctions.ln(new BigDecimal("2",contexter.getContext()),contexter.getContext().getPrecision()),contexter.getContext());
        return toreturn;
    }


    BigDecimal computePriorMinEntropy() {
         if (secretsize.compareTo(BigInteger.ZERO)<0){
            this.analyze();
        }
        BigDecimal decsecretsize = new BigDecimal(secretsize);
        BigDecimal highestprob = contexter.getDecimalOne().divide(decsecretsize, contexter.getContext());
        BigDecimal priorminentropy = contexter.getDecimalZero().subtract(log2(highestprob),contexter.getContext());
         
        return priorminentropy;
    }
    
    protected static void log(Object aObject, int level){
        if (logger.getVerbosity() >= level) {
            Logger loggy = Logger.getInstance();
            loggy.log(String.valueOf(aObject),level);
        }
    }
}
