/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package hyleak;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.logging.Level;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import distributions.MapDistribution;
import support.Contexter;
import support.Logger;

/**
 *
 * @author  
 */
class RunnableSingleProbabilisticExplorer implements Runnable {
    private static final Contexter contexter = Contexter.getInstance();
    private static final Logger logger = Logger.getInstance();
    

    private ArrayList<String> codelines = null;
    private final ConcurrentSimplePosteriorDistributionManager postdistman;
    private final ThreadPoolExecutor tpoolexplorer;
    private final AtomicInteger count;
    private final MarkovChainState currstate;
    private final ExpressionEvaluator eval = new ExpressionEvaluator();
    private boolean terminated;
    private boolean parsingerror;

    public RunnableSingleProbabilisticExplorer(MarkovChainState firststate, ArrayList<String> codelines, ConcurrentSimplePosteriorDistributionManager postdistman, ThreadPoolExecutor tpoolexplorer, AtomicInteger count) {
        currstate = firststate;
        this.codelines = codelines;
        this.postdistman = postdistman;
        this.tpoolexplorer = tpoolexplorer;
        this.count = count;
    }
    
    @Override
    public void run(){
        //initializing single exploration
        
        
     
        log("Starting exploration in state "+currstate.printNice()+" having probability "+currstate.getProbability().toEngineeringString(),3);
        /*if (initialstate == null){
            currstate = new MarkovChainState("current",contexter.getDecimalOne());
            MarkovStatePublicVariable programcounter = new MarkovStatePublicVariable("pc",BigInteger.ONE,Integer.valueOf(32));
            currstate.addVariable(programcounter);        
        }else{
            try {
                currstate =initialstate.clone();
            } catch (CloneNotSupportedException ex) {
                java.util.logging.Logger.getLogger(Simulator.class.getName()).log(Level.SEVERE, "Unable to clone state"+currstate.getName(), ex);
                System.exit(1);
            }
        }*/
        
        terminated = false;
        parsingerror = false;
        while (!terminated){
          BigInteger pc = currstate.getPublicVariableValue("pc");
          if (pc.compareTo(BigInteger.valueOf(-1)) ==0){
              log("ERROR: found state with no program counter information; exiting explorer",1);
              break;
          }
          if (currstate.getProbability().compareTo(contexter.getDecimalZero())==0){
              log("ERROR: current state has probability zero",1);
              System.exit(1);
          }
          String line = codelines.get(pc.intValue()); 
          log("",4);
          if (logger.getVerbosity()>=4){
              log("Exploring line "+pc+": "+line+" in state "+currstate.printNice(),4);
          }
          
          if (line.startsWith("private")){
              explorePrivateLine(line,pc,false);
          }else if (line.startsWith("public")){
              explorePublicLine(line,pc,false);
          }else if (line.startsWith("secret")){
              
              explorePrivateLine(line,pc,true);
          }else if (line.startsWith("observable")){
              explorePublicLine(line,pc,true);
          }else if (line.startsWith("assign")){
              exploreAssignLine(line,pc);
          }else if (line.startsWith("skip")){
              exploreSkipLine(line,pc);
          }else if (line.startsWith("random")){
              exploreRandomLine(line,pc);
          }else if (line.startsWith("goto")){
              exploreGotoLine(line,pc);
          }else if (line.startsWith("free")){
              exploreFreeLine(line,pc);
          }else if (line.startsWith("return")){
              exploreReturnLine(line,pc);
          }else if (line.startsWith("if")){
              exploreIfLine(line,codelines.get(pc.add(BigInteger.ONE).intValue()),codelines.get(pc.add(new BigInteger("2")).intValue()),pc);
          }else {
              log("ERROR: line "+pc.toString()+" unrecognized, this should never happen; exiting parser",1);
              parsingerror = true;
              break;
          }
          if (parsingerror){
              log("ERROR: parsing error; exiting parser",1);
              System.exit(1);
          }  
        }
        log("Found final state "+currstate.printNice(),3);
       
   
        
        int c = count.decrementAndGet();
        log("Decreasing count to "+c,3);
        if (c== 0){
            tpoolexplorer.shutdown();
            log("I am the last thread, terminating the pool. Active count = "+tpoolexplorer.getActiveCount(),2);
        } 
        postdistman.inferInformation(currstate);
    }
    
    private void explorePrivateLine(String aLine, BigInteger pc, boolean isSecret) {
        String varname;
        Integer varlength;
        Interval varinterval;
    
    
    /*
     * String input = "1 fish 2 fish red fish blue fish";
     Scanner s = new Scanner(input);
     s.findInLine("(\\d+) fish (\\d+) fish (\\w+) fish (\\w+)");
     MatchResult result = s.match();
     for (int i=1; i<=result.groupCount(); i++)
         System.out.println(result.group(i);
     s.close(); 
     */
    
      Pattern pattern;
       if (isSecret){
          pattern = HyleakPatterns.pattern_secret;
       }else{
          pattern = HyleakPatterns.pattern_private;
      }
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+pattern.pattern(),4);
    }
    
    Matcher matcher = pattern.matcher(aLine );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("Variable length: " + matcher.group(1),4);
      log("Variable name: " + matcher.group(2),4);
      varlength = Integer.valueOf(matcher.group(1));
      varname = matcher.group(2);
      BigInteger interend = BigInteger.valueOf(2).pow(varlength.intValue()).subtract(BigInteger.ONE);
      varinterval= new Interval (BigInteger.ZERO,interend);
    }else{
        if (isSecret){
            pattern = HyleakPatterns.pattern_secret_with_assignment ;
        }else{
            pattern = HyleakPatterns.pattern_private_with_assignment ;
        }
        if (logger.getVerbosity()>=4){
            log("Testing if it matches regular expression "+pattern.pattern(),4);
        }
    
        matcher = pattern.matcher(aLine );
        if ( matcher.find() ) {
            log("Num groups: " + matcher.groupCount(),4);
            log("Variable length: " + matcher.group(1),4);
            log("Variable name: " + matcher.group(2),4);
            log("Assigned interval: "+matcher.group(3),4);
            varlength = Integer.valueOf(matcher.group(1));
            varname = matcher.group(2);
            varinterval=new Interval(matcher.group(3));
        }else{
            log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
            parsingerror = true;
            return;
        }
        
    }
    
    
    
    MarkovStatePrivateDistributionVariable var = new MarkovStatePrivateDistributionVariable(varname, varinterval, varlength);
    //MarkovStatePrivateVariable var = new MarkovStatePrivateVariable(varname,varinterval,varlength);
    var.setSecret(isSecret);
    currstate.addVariable(var);

    if (logger.getVerbosity()>=3){
        if (isSecret){
            log("Added secret variable "+var.toString()+" with "+var.getSize()+" possible values to state "+currstate.getName(),3);        
        }else{
            log("Added private variable "+var.toString()+" with "+var.getSize()+" possible values to state "+currstate.getName(),3);
        }
    }
    
    
      
    //
       
    BigInteger newpc = pc.add(BigInteger.ONE);
    currstate.setPublicVariable("pc", newpc);
                
    }

    private void explorePublicLine(String aLine, BigInteger pc, boolean isObservable) {
        if (isObservable){
        log("Parsing observable line: "+aLine,4);
    }else{
        log("Parsing public line: "+aLine,4);
    }
       
       
    //Scanner scanner = new Scanner(aLine);
    String varname;
    Integer varlength;
    BigInteger varvalue;
    
    
    /*
     * String input = "1 fish 2 fish red fish blue fish";
     Scanner s = new Scanner(input);
     s.findInLine("(\\d+) fish (\\d+) fish (\\w+) fish (\\w+)");
     MatchResult result = s.match();
     for (int i=1; i<=result.groupCount(); i++)
         System.out.println(result.group(i);
     s.close(); 
     */
    /*
    scanner.findInLine("public int[\d+] \w+;");
    MatchResult result = scanner.match();*/
    Pattern pattern;
    if (isObservable){
        pattern = HyleakPatterns.pattern_observable ;
    }else{
        pattern = HyleakPatterns.pattern_public ;
    }
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+pattern.pattern(),4);
    }
    
    Matcher matcher = pattern.matcher(aLine );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("Variable length: " + matcher.group(1),4);
      log("Variable name: " + matcher.group(2),4);
      varlength = Integer.valueOf(matcher.group(1));
      varname = matcher.group(2);
      varvalue = BigInteger.ZERO;
    
    }else{
        if (isObservable){
            pattern = HyleakPatterns.pattern_observable_with_assignment ;
        }else{
            pattern = HyleakPatterns.pattern_public_with_assignment ;
        }
        if (logger.getVerbosity()>=4){
            log("Testing if it matches regular expression "+pattern.pattern(),4);
        }
        
        matcher = pattern.matcher(aLine );
        if ( matcher.find() ) {
          log("Num groups: " + matcher.groupCount(),4);
          log("Variable length: " + matcher.group(1),4);
          log("Variable name: " + matcher.group(2),4);
          log("Assigned value: "+matcher.group(3),4);
          varlength = Integer.valueOf(matcher.group(1));
          varname = matcher.group(2);
          try{
          varvalue = eval.evaluateExpressionToDirac(matcher.group(3),currstate.getVariables()).toBigInteger();
          } catch (ParseException e) {
              log("ERROR: unable to evaluate expression "+matcher.group(3)+" in state "+currstate.printNice(),1);
            
              parsingerror = true;
              return;
          }
    
        }else{
            log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
     
            parsingerror = true;
            return;
        }
    }
    

    MarkovStatePublicDistributionVariable var = new MarkovStatePublicDistributionVariable(varname,varvalue,varlength);
    var.setObservable(isObservable);
    currstate.addVariable(var);

    if (logger.getVerbosity()>=3){
        if (isObservable){
            log("Added observable variable "+var.toString()+" to state "+currstate.getName(),3);
        }else{
            log("Added public variable "+var.toString()+" to state "+currstate.getName(),3);
        }
    }
    
    
  
    //
   
    
   
    BigInteger newpc = pc.add(BigInteger.ONE);
    currstate.setPublicVariable("pc", newpc);
  
  
   
    }
    
    

    private void exploreRandomLine(String line, BigInteger pc){
        String varname;
        BigInteger newpc = pc.add(BigInteger.ONE);
        if (logger.getVerbosity()>=4){
            log("Testing if it matches regular expression "+HyleakPatterns.pattern_randombit.pattern(),4);
        }
        
        Matcher matcher = HyleakPatterns.pattern_randombit.matcher(line );
        if ( matcher.find() ) {
          log("Num groups: " + matcher.groupCount(),4);
          log("Variable name: " + matcher.group(1),4);
          log("Probability of being 0: " + matcher.group(2),4);
          varname = this.getVariableName(matcher.group(1),currstate.getVariables());
          BigDecimal probofzero = new BigDecimal(matcher.group(2),contexter.getContext());
          
          if (probofzero.compareTo(contexter.getDecimalOne())>0){
                log("Found a probability greater than 1; using 1 instead",4);
                probofzero = contexter.getDecimalOne();
            } else if (probofzero.compareTo(contexter.getDecimalZero())<0){
                log("Found a probability lesser than 0; using 0 instead",4);
                probofzero = contexter.getDecimalZero();
            }


         
           
          log("Assiging to public variable "+varname+" value 0 with probability "+probofzero+" in state "+currstate.getName(),3);
          if (probofzero.compareTo(contexter.getDecimalOne())==0){
              currstate.setDistributionVariable(varname,BigInteger.ZERO);
              log("The value of "+varname+" is 0 with probability 1",3);
          }else if (probofzero.compareTo(contexter.getDecimalZero())==0){
              currstate.setDistributionVariable(varname,BigInteger.ONE);
              log("The value of "+varname+" is 1 with probability 1",3);
          }else{
               log("Assigning to "+varname+" binary distribution with probability of zero "+probofzero,3);
              
               BigInteger[] values = new BigInteger[2];
               values[0] = BigInteger.ZERO;
               values[1] = BigInteger.ONE;
               
               BigDecimal[] probs = new BigDecimal[2];
               probs[0] = probofzero;
               probs[1] = contexter.getDecimalOne().subtract(probofzero, contexter.getContext());
               currstate.setDistributionVariable(varname,values,probs);
               
               
          }

      
           currstate.setPublicVariable("pc", newpc);
          
          
        }else{
            if (logger.getVerbosity()>=4){
            log("Testing if it matches regular expression "+HyleakPatterns.pattern_random.pattern(),4);
            }

            matcher = HyleakPatterns.pattern_random.matcher(line );
            if ( matcher.find() ) {
                log("Num groups: " + matcher.groupCount(),4);
          log("Variable name: " + matcher.group(1),4);
          log("Lower bound: " + matcher.group(2),4);
          log("Upper bound: " + matcher.group(3),4);
          varname = this.getVariableName(matcher.group(1),currstate.getVariables());

        BigInteger lowbound = new BigInteger("0");
        try{
            lowbound = eval.evaluateExpressionToDirac(matcher.group(2),currstate.getVariables()).toBigInteger();
        }catch (ParseException e){
            log(e.getMessage(),1);
        }

 
        BigInteger uppbound = new BigInteger("0");
        try{
            uppbound = eval.evaluateExpressionToDirac(matcher.group(3),currstate.getVariables()).toBigInteger();
        }catch (ParseException e){
            log(e.getMessage(),1);
        }
           
          if (lowbound.compareTo(uppbound)>0){
              log("ERROR: lower bound greater than upper bound?",2);
          
              parsingerror = true;
              return;
          }
          
          BigInteger diff = uppbound.subtract(lowbound).add(BigInteger.ONE);
          if (diff.compareTo(BigInteger.ONE)==0){
            //VariableList vars =currstate.getVariables().clone();
            currstate.setDistributionVariable(varname, uppbound);
            log("Assigned to public variable "+varname+" value "+uppbound.toString()+" with probability 1 in state "+currstate.getName(),3);
           


            currstate.setPublicVariable("pc", newpc);
           
            
            
          }else{
                           
              currstate.setDistributionVariable(varname, lowbound,uppbound);
              
              
              log("Assigned to variable "+varname+" a uniform distribution from "+lowbound+" to "+uppbound,3);

      

    
          }
        
            }else{
                log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);

             
                parsingerror = true;
            }
                
        }
        
    }
    private String getVariableName(String varname, VariableList variables) {
        String toreturn =varname;
        if (varname.contains("[")){
            String temp = varname.substring(varname.indexOf('['));
            while (temp.contains("[")){
                int closedpar = temp.indexOf(']');
                String index = temp.substring(1, closedpar);
                log("Evaluating index "+index,4);
                index = this.getVariableName(index, variables);
                BigInteger val = new BigInteger("0");
                try {
                    val = eval.evaluateInnerExpressionToDirac(index, variables).toBigInteger();
                } catch (ParseException e){
                    log(e.getMessage(),1);
                }
                toreturn = toreturn.replace("["+index+"]", "["+val.toString()+"]");
                temp = temp.substring(closedpar+1);
            }
        }
        return toreturn;
    }

    private void exploreAssignLine(String aLine, BigInteger pc) {
        String varname;
    String varexpression;
    MapDistribution varvalue;
    
    
    
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+HyleakPatterns.pattern_assign.pattern(),4);
    }
    
    Matcher matcher = HyleakPatterns.pattern_assign.matcher(aLine );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("Variable name: " + matcher.group(1),4);
      log("Variable expression: " + matcher.group(2),4);
      varname = this.getVariableName(matcher.group(1),currstate.getVariables());
      varexpression = matcher.group(2);
    }else{
        log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
        
        parsingerror = true;
        return;
    }
    VariableList vars =currstate.getVariables().clone();
    try{
        varvalue = eval.evaluateExpressionToDistribution(varexpression,vars);
    }catch (ParseException e){
        log("ERROR: unable to evaluate expression "+varexpression+" in state "+currstate.printNice(),1);
     
        parsingerror = true;
        return;
    }
    
    log("Assigned to public or observable variable "+varname+" distribution "+varvalue+" in state "+currstate.getName(),3);
    
    BigInteger newpc = pc.add(BigInteger.ONE);
    currstate.setPublicVariable("pc", newpc);
    currstate.setDistributionVariable(varname, varvalue);
    
    }

    private void exploreIfLine(String ifline, String thenline, String elseline, BigInteger pc) {

    
    String guard;
    BigInteger gotothen,gotoelse;
    
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+HyleakPatterns.pattern_if.pattern(),4);
    }
    
    Matcher matcher = HyleakPatterns.pattern_if.matcher(ifline );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("If guard: " + matcher.group(1),4);
      guard = matcher.group(1);
    }else{
        log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
       
        parsingerror = true;
        return;
    }
    
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+HyleakPatterns.pattern_then.pattern(),4);
    }
    
    matcher = HyleakPatterns.pattern_then.matcher(thenline );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("Then destination: " + matcher.group(1),4);
      gotothen = new BigInteger(matcher.group(1));
    }else{
        log("ERROR: line "+currstate.getPublicVariableValue("pc").add(BigInteger.ONE) +" does not match regular expression.",1);
      
        parsingerror = true;
        return;
    }
    
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+HyleakPatterns.pattern_else.pattern(),4);
    }
    
    matcher = HyleakPatterns.pattern_else.matcher(elseline );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("Else destination: " + matcher.group(1),4);
      gotoelse = new BigInteger(matcher.group(1));
    }else{
        log("ERROR: line "+currstate.getPublicVariableValue("pc").add(BigInteger.ONE).add(BigInteger.ONE) +" does not match regular expression.",1);
       
        parsingerror = true;
        return;
    }
    
    log("Exploring if statement with guard "+guard+", then line "+gotothen+" and else line "+gotoelse+" to state "+currstate.printNice(),3);
    
    
    String varname;
    String operator;
    String expression;
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+HyleakPatterns.pattern_guard.pattern(),4);
    }
    
    matcher = HyleakPatterns.pattern_guard.matcher(guard );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("Variable name: " + matcher.group(1),4);
      log("Operator: " + matcher.group(2),4);
      log("Guard expression: "+matcher.group(3),4);
      varname = this.getVariableName(matcher.group(1),currstate.getVariables());
      operator = matcher.group(2);
      expression = matcher.group(3);
    }else{
        log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
        
        parsingerror = true;
        return;
    }
   
    // evaluate the guard expression
        MapDistribution expval = new MapDistribution(BigInteger.ZERO);
        try{
            expval = eval.evaluateExpressionToDistribution(expression, currstate.getVariables());
        }catch (ParseException e){
            log(e.getMessage(),1);
            System.exit(1);
        }
        log("The expression evaluates to the distribution "+expval,4);
    
   //boolean secretdependent = true;
   MapDistribution argdist = new MapDistribution(BigInteger.ZERO);
   MarkovStateDistributionVariable variable = null;
   log("The regular expression for a constant is "+HyleakPatterns.pattern_expression_constant.pattern(),4);
   matcher = HyleakPatterns.pattern_expression_constant.matcher(varname);
   if ( matcher.matches() ) {
       log("The argument is the constant " + matcher.group(1),4);
       argdist = new MapDistribution(new BigInteger(matcher.group(1)));
      // secretdependent = false;
   }else{
    
        // obtain the variable
        Variable var = currstate.getVariables().getVariable(varname);
            
        // if the variable is private, the if is secret-dependent; if not take the value
        //if (var.isPrivate()==false){
            variable = (MarkovStateDistributionVariable)var;
            argdist= variable.getDistribution();
            log("The argument variable "+variable.getName()+" has distribution "+argdist,4);
           // secretdependent = false;
        //}
        
    }
        

       VariableList thenvars =currstate.getVariables().clone();
       thenvars.setPublicVariable("pc", gotothen);
       VariableList elsevars =currstate.getVariables().clone();
       elsevars.setPublicVariable("pc", gotoelse);

       
       log("Variables cloned for then and else states",4);
       
       ProbabilisticGuard probguard = new ProbabilisticGuard(variable,thenvars,argdist,operator,expval);
       
       
       log("I built the probabilistic guard",4);
       
    

       BigDecimal thenprob = probguard.getProbOfTrue();
    
       log("The guard is true with probability "+thenprob.toString(),3);
       if (thenprob.compareTo(contexter.getDecimalOne())==0){
           // the guard is certainly true
           
           thenvars = probguard.getVarsIfTrue();
           currstate.setVariables(thenvars);
           currstate.setPublicVariable("pc",gotothen);
           log("The guard is always true; reducing distribution on variable "+varname+" to "+thenvars.getDistributionVariable(varname),3);
       }else if (thenprob.compareTo(contexter.getDecimalZero())==0){
           // the guard is certainly false
           elsevars = probguard.getVarsIfFalse();
           currstate.setVariables(elsevars);
           currstate.setPublicVariable("pc",gotoelse);
           
           log("The guard is always false; reducing distribution on variable "+varname+" to "+elsevars.getDistributionVariable(varname),3);
       }else{
           // the guard is true with probability thenprob
           //log("Exploring the case in which the guard is true and variable "+varname+" is "+hthen.getInterval()+"; putting the case in which the guard is false and variable "+varname+" is "+helse.getInterval()+" on the stack",3);
           
           MarkovChainState otherstate = null;
              try {
                    otherstate = currstate.clone();
              } catch (CloneNotSupportedException ex) {
                  java.util.logging.Logger.getLogger(RunnableSingleExplorer.class.getName()).log(Level.SEVERE, "Unable to clone state "+currstate.getName(), ex);
                  System.exit(1);
           }
           thenvars = probguard.getVarsIfTrue();
           currstate.setVariables(thenvars);
           currstate.setPublicVariable("pc",gotothen);
           currstate.multiplyProbabilityBy(thenprob);
           log("Next state explored will be "+currstate.printNice()+" with probability "+currstate.getProbability().toEngineeringString(),3);
           
           BigDecimal elseprob = contexter.getDecimalOne().subtract(thenprob,contexter.getContext());
           elsevars = probguard.getVarsIfFalse();
           otherstate.setVariables(elsevars);
           otherstate.setPublicVariable("pc",gotoelse);
           otherstate.multiplyProbabilityBy(elseprob);
           
           //log("thenprob is "+thenprob.toEngineeringString()+", elseprob is "+elseprob.toEngineeringString(),3);
           log("Put on the stack state "+otherstate.printNice()+" with probability "+otherstate.getProbability().toEngineeringString(),3);
           //toexplore.add(otherstate);
           int c = count.incrementAndGet();
           log("Increasing count to "+c,3);
           tpoolexplorer.submit(new RunnableSingleProbabilisticExplorer(otherstate,codelines,postdistman,tpoolexplorer,count));
       }
       
       
  /* }else{
       // the if does not depend on any private variables, and the value of the argument is known
       // since ifs don't have side effects, no variables except pc will be changed; we just need
       // to know if the next state will have the pc value of the then or of the else
       boolean guardistrue = false;
            switch (operator) {
                case "==":
                    if (argdist.compareTo(expval)==0) {
                        guardistrue = true;
                    }
                    break;
                case "!=":
                    if (argdist.compareTo(expval)!=0) {
                        guardistrue = true;
                    }
                    break;
                case ">=":
                    if (argdist.compareTo(expval)>=0) {
                        guardistrue = true;
                    }
                    break;
                case "<=":
                    if (argdist.compareTo(expval)<=0) {
                        guardistrue = true;
                    }
                    break;
                case "<":
                    if (argdist.compareTo(expval)<0) {
                        guardistrue = true;
                    }
                    break;
                case ">":
                    if (argdist.compareTo(expval)>0){
                        guardistrue = true;
                    }
                
                    break;
            }
       
       if (guardistrue) {
           currstate.setPublicVariable("pc", gotothen);
       }else{
           currstate.setPublicVariable("pc", gotoelse);
       }

       
   }*/
    
    }
    
    private VariableList processGuard(VariableList vars, String varname, String operator, BigInteger expval, boolean b) {
        if (logger.getVerbosity()>=4){
            log("Processing guard "+varname+operator+expval+" with variables "+vars.toString(),4);
        }

        // invert the operator if needed
        if (b==false){
            log("Inverting the operator as requested; original operator: "+operator,4);
            switch (operator) {
                case "==":
                    operator = "!=";
                    break;
                case "!=":
                    operator = "==";
                    break;
                case ">=":
                    operator = "<";
                    break;
                case "<=":
                    operator = ">";
                    break;
                case "<":
                    operator = ">=";
                    break;
                case ">":
                    operator = "<=";
                    break;
            }
            log ("New operator: "+operator,4);
        }
        
        // the guard is reduced to private variable -- operator -- constant; now we reduce the allowed intervals of the private variable accordingly
       PrivateVariable privvar = (PrivateVariable)vars.getVariable(varname);
       String intervals = privvar.getInterval().toString();
       String newintervals = "";
       
       while (intervals.compareTo("")!=0){
           // look for a closed parenthesis
           int clospos =-1;
           for (int i=0;i<intervals.length();i++){
               if (intervals.charAt(i)==']'){
                   clospos = i;
                   break;
               }
           }
       
           // if there are no closed parenthesis we are done; return
           if (clospos==-1){
               break;
           
           }else{
               //else find the last opened parenthesis before the closed one
               String oneinterval = intervals.substring(0, clospos+1);
               log ("Found closed parenthesis: "+oneinterval,4);
               int openpos = oneinterval.lastIndexOf('[');
               oneinterval = oneinterval.substring(openpos);
               log("Found single interval "+oneinterval,4);
               // found an inner expression; evaluate it (without the parenthesis)
               int commapos = oneinterval.lastIndexOf(',');
               String bottomstring = oneinterval.substring(1, commapos);
               log("Bottom string: "+bottomstring,4);
               BigInteger bottom = new BigInteger(bottomstring);
               String topstring = oneinterval.substring(commapos+1, oneinterval.length()-1);
               log("Top string: "+topstring,4);
               BigInteger top = new BigInteger(topstring);
               
               String onenewinterval = this.reduceInterval(bottom,top,operator,expval);
               newintervals += onenewinterval;
               
               if (intervals.lastIndexOf('[', oneinterval.length())==-1){
                   intervals = "";
               }else{
                   intervals = intervals.substring(oneinterval.length());
               }
               
           }
       }
       privvar.setInterval(newintervals);
       if (logger.getVerbosity()>=4){
           log("Setting intervals of variable "+privvar.getName()+" to "+privvar.getInterval(),4);
       }
        
        return vars;
    }
    
    private String reduceInterval(BigInteger bottom, BigInteger top, String operator, BigInteger expval) {
        log("Reducing inverval ["+bottom+","+top+"] assuming that "+operator+expval+" holds",4);
        String toreturn = "";
        switch (operator) {
            case "==":
                // processing equality
                // if the value is outside the range return empty; else return singleton interval
                if ((expval.compareTo(bottom)>=0)&&(expval.compareTo(top)<=0)){
                    toreturn = "["+expval+","+expval+"]";
                }else{
                    toreturn = "";
                }
                break;
            case "!=":
                // processing inequality
                // if the value is less than bottom or more than top, nothing to do
                if ((expval.compareTo(bottom)<0)||(expval.compareTo(top)>0)){
                    toreturn = "["+bottom+","+top+"]";
                }else if ((expval.compareTo(bottom)==0)&&(expval.compareTo(top)==0)){
                    // return empty string
                    toreturn = "";
                }else if (expval.compareTo(bottom)==0){
                    // cut the bottom value
                    toreturn = "["+(bottom.add(BigInteger.ONE))+","+top+"]";
                }else if (expval.compareTo(top)==0){
                    // cut the top value
                    toreturn = "["+bottom+","+(top.subtract(BigInteger.ONE))+"]";
                }else{
                    // the value is inside; we need to split the interval in two
                    toreturn = "["+bottom+","+(expval.subtract(BigInteger.ONE))+"]["+(expval.add(BigInteger.ONE))+","+top+"]";
                }
                break;
            case ">=":
                // processing greater or equal
                // if the value is lesser or equal to bottom, nothing to do
                if (expval.compareTo(bottom)<=0){
                    toreturn = "["+bottom+","+top+"]";
                }else if (expval.compareTo(top)>0){
                    // if the value is greater that top, return empty string
                    toreturn = "";
                }else {
                    // else move the bottom to expval
                    toreturn = "["+expval+","+top+"]";
                }
                break;
            case "<=":
                // processing lesser or equal
                // if the value is greater or equal to top, nothing to do
                if (expval.compareTo(top)>=0){
                    toreturn = "["+bottom+","+top+"]";
                }else if (expval.compareTo(bottom)<0){
                    // if the value is lesser that bottom, return empty string
                    toreturn = "";
                }else {
                    // else move the top to expval
                    toreturn = "["+bottom+","+expval+"]";
                }
                break;
            case "<":
                // processing lesser
                // if the value is greater than top, nothing to do
                if (expval.compareTo(top)>0){
                    toreturn = "["+bottom+","+top+"]";
                }else if (expval.compareTo(bottom)<=0){
                    // if the value is lesser or equal that bottom, return empty string
                    toreturn = "";
                }else {
                    // else move the top to expval-1
                    toreturn = "["+bottom+","+(expval.subtract(BigInteger.ONE))+"]";
                }
                break;
            case ">":
                // processing greater
                // if the value is lesser than bottom, nothing to do
                if (expval.compareTo(bottom)<0){
                    toreturn = "["+bottom+","+top+"]";
                }else if (expval.compareTo(top)>=0){
                    // if the value is greater or equal than top, return empty string
                    toreturn = "";
                }else {
                    // else move the bottom to expval+1
                    toreturn = "["+(expval.add(BigInteger.ONE))+","+top+"]";
                }
                break;
        }
        log("Reduced to "+toreturn,4);
        return toreturn;
    }

    private void exploreGotoLine(String aLine, BigInteger pc) {
        BigInteger gotoline;
    
    
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+HyleakPatterns.pattern_goto.pattern(),4);
    }
    
    Matcher matcher = HyleakPatterns.pattern_goto.matcher(aLine );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("Goto line destination: " + matcher.group(1),4);
      gotoline = new BigInteger(matcher.group(1));
    }else{
        log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
    
        parsingerror = true;
        return;
    }
   
    log("Adding goto transition from state "+currstate.getName()+" to line "+gotoline,3);
    
    
    
      
    BigInteger newpc = gotoline;
   
    currstate.setPublicVariable("pc", newpc);
 
 
   
    }
    
    private void exploreSkipLine(String aLine, BigInteger pc) {
   
    
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+HyleakPatterns.pattern_skip.pattern(),4);
    }
    
    Matcher matcher = HyleakPatterns.pattern_skip.matcher(aLine );
    if ( matcher.find()==false ) {
        log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
        parsingerror = true;
        return;
    }
    log("Adding skip transition in state "+currstate.getName(),3);
       
        
    BigInteger newpc = pc.add(BigInteger.ONE);
    currstate.setPublicVariable("pc", newpc);
                
   }
    

    private void exploreReturnLine(String aLine, BigInteger pc) {
        if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+HyleakPatterns.pattern_return.pattern(),4);
    }
    
    Matcher matcher = HyleakPatterns.pattern_return.matcher(aLine );
    if ( matcher.find()==false ) {
        log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
       
        parsingerror = true;
        return;
    }
      log("Found return statement. Terminating.",3);
      this.terminated = true;
   
    }
    
    private void exploreFreeLine(String aLine, BigInteger pc) {
        
       
    String varname;
        
    
    
    if (logger.getVerbosity()>=4){
        log("Testing if it matches regular expression "+HyleakPatterns.pattern_free.pattern(),4);
    }
    
    Matcher matcher = HyleakPatterns.pattern_free.matcher(aLine );
    if ( matcher.find() ) {
      log("Num groups: " + matcher.groupCount(),4);
      log("Variable name: " + matcher.group(1),4);
      varname = this.getVariableName(matcher.group(1),currstate.getVariables());
      
    }else{
        log("ERROR: line "+currstate.getPublicVariableValue("pc") +" does not match regular expression.",1);
       
        parsingerror = true;
        return;
    }
  
    
    
    log("Freed variable "+varname+" in state "+currstate.getName(),3);
    
    BigInteger newpc = pc.add(BigInteger.ONE);
    currstate.setPublicVariable("pc", newpc);
    currstate.deleteVariableByName(varname);
    
  
    }
    
    private static void log(Object aObject, int level){
        Logger loggy = Logger.getInstance();
        loggy.log(String.valueOf(aObject),level);
    }

    
    
}
