/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package hyleak;

import support.Logger;
import distributions.JointFrequencyDistribution;

/**
 *
 * @author  
 */
class SimulationStorage {

    private static final Logger logger = Logger.getInstance();
    private final SimulationSignature signature;
    private JointFrequencyDistribution dist;
    private int numOfSims;
    private String secretid;
    private boolean abstractable;

    SimulationStorage(SimulationSignature sig) {
        signature = sig;
        dist = new JointFrequencyDistribution();
        numOfSims = 0;
        secretid = "";
        abstractable = false;
    }

    synchronized void storeResult(MarkovChainState currstate) {
        numOfSims++;
        String secretId = currstate.getSecretSingleIdentifier();
        String observId = currstate.getObservableIdentifier();
        //log("Stored result "+secretId+" , "+observId,3);
        dist.increaseFrequency(secretId, observId);
    }

    synchronized void storeSecretIdentifierForAbstraction(String secid) {
        if (!signature.isAbstract()) {
            abstractable = false;
            //return false;
        } else {
            abstractable = true;
            log("Secret identifier to store for abstraction: "+secid,3);
            this.secretid = secid;
            //return true;
        }
    }

    void sendResults(JointPosteriorDistributionManager postDistMan) {
        //void sendResults(JointPosteriorDistributionManagerWithKnownPrior postDistMan) {
        if (signature.isAbstract()) {
            // set the distribution of each secret to the distribution of the secret that you have actually tested
            if (secretid.isEmpty()) {
                log("ERROR: abstract storage has not been initialized for abstraction", 1);
                System.exit(1);
            } else {
                dist = dist.replicateSecretDistribution(secretid, signature);
                

                postDistMan.storeFreqSubDistToManager(signature, dist, numOfSims, signature.getProbability(), signature.isAbstract());
                //postDistMan.storeFreqSubDistToManager(signature, dist, numOfSims,signature.getProbability(), signature.getSubPrior());
                //postDistMan.storeAbstractableToManager(signature, abstractable);
            }
        } else {
            // you can send the distribution as it is
            
            postDistMan.storeFreqSubDistToManager(signature, dist, numOfSims, signature.getProbability(), signature.isAbstract());
            //postDistMan.storeFreqSubDistToManager(signature, dist, numOfSims,signature.getProbability(), signature.getSubPrior());
        }
    }

    void printFrequencyDistribution() {
        dist.print();
    }

    private static void log(Object aObject, int level) {
        if (logger.getVerbosity() >= level) {
            logger.log(String.valueOf(aObject), level);
        }
    }
}
