/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package hyleak;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.logging.Level;
import java.util.logging.Logger;
import theory.BigIntegerCartesianIterator;

/**
 *
 * @author  
 * @author  
 */
public class Simulator {
    private static final support.Logger logger = support.Logger.getInstance();
    private final int waitTimeDefault = 1;
    private final boolean priorIsKnown = JointPosteriorDistributionManager.priorIsKnown;

    private int numSims = 0;
    private int numSimsDone = 0;
    private final int numThreads;
    private final long keepAliveTime;
    private final ArrayList<String> codelines;
    private final int rounds = 1;
    private final ConcurrentLinkedQueue<SimulationSignature> simQueue;
    private final JointPosteriorDistributionManager postDistMan;
    private Map<Signature, BigInteger> simNumber = new HashMap<>();  //to store the number of simulations per signature
    //private final HashMap<SimulationSignature,SimulationStorage> simStorage = new HashMap();
    private int submitted;
    private final String mode;

    public Simulator(int numSimulations, JointPosteriorDistributionManager postdistman, int nThreads, long keepAlive, ArrayList<String> code) {
    	if(priorIsKnown) {
            this.mode = "ALLSECRETS";
    	} else {
            this.mode = "RNDSECRETS";
    	}
        numSims = numSimulations;
        numSimsDone = 0;
        simQueue = new ConcurrentLinkedQueue();
        postDistMan = postdistman;
        numThreads = nThreads;
        keepAliveTime = keepAlive;
        codelines = code;
    }

    public int getNumSims() {
        return numSims;
    }

    
    
    public void queueSimulation(String filename, MarkovChainState currstate, VariableList obsvar, VariableList privvar, boolean abstr) {
        simQueue.add(new JavaSimulationSignature(filename, currstate, obsvar, privvar,abstr));
    }

    public void queueSimulation(MarkovChainState currstate, VariableList obsvar, VariableList privvar, boolean abstr) {
        simQueue.add(new HyleakSimulationSignature(currstate, obsvar, privvar, abstr));
    }

    public void queueSimulation(SimulationSignature simSig) {
        simQueue.add(simSig);
    }

    public int getQueueSize() {
        return simQueue.size();
    }

    void simulate() {
        HashMap<Signature, BigDecimal> mapExecProbs = new HashMap<>();
        HashMap<Signature, Integer> mapEsitmateNumOfObservables = new HashMap<>();
        HashMap<Signature, Integer> mapSizesOfSupports = new HashMap<>();

        for (SimulationSignature sig : simQueue) {
            //simStorage.put(sig, new SimulationStorage(sig));
        	// Store the probability of executing the component with the signature
            mapExecProbs.put(sig, sig.getProbability());
            // Store the estimated size of the observable domain for each signature
            mapEsitmateNumOfObservables.put(sig, sig.getSizeOfEstimatedSizesOfObservables());
         // Store the number of division of each signature
            if(sig.isAbstract) {
                mapSizesOfSupports.put(sig, sig.getSizeOfSupport());
            } else {
                mapSizesOfSupports.put(sig, 1);
            }
        }
        log("mapExecProbs has " + mapExecProbs.size() + " elements", 4);
        simNumber = postDistMan.computeSampleSizesInitially(mapExecProbs, mapEsitmateNumOfObservables, mapSizesOfSupports, rounds);
        BigInteger expectedSimNumber = BigInteger.ZERO;
        for (BigInteger bi : simNumber.values()) {
            expectedSimNumber = expectedSimNumber.add(bi);
        }
        log("Total simulations expected: " + expectedSimNumber.toString(), 2);
        for (int i = 0; i < rounds; i++) {
            log("\n-------------------------------------------", 3);
            log("Starting simulation round " + (i + 1) + " of " + rounds, 2);

            for (SimulationSignature sig : simQueue) {
                int numOfSims = Math.max(1,simNumber.get(sig).intValue());
                log("Running " + numOfSims + " simulations for a component with probability " + sig.getProbability().toString(), 3);
                if (numOfSims > 0) {
                    ThreadPoolExecutor tpoolsimulator = new ThreadPoolExecutor(numThreads, numThreads, keepAliveTime, TimeUnit.MILLISECONDS, new LinkedBlockingQueue<Runnable>());
                    submitted = 0;
                    SimulationStorage storage = new SimulationStorage(sig);
                    if(sig.isAbstract) {
                        runAbstractSimulation(sig, numOfSims, storage, tpoolsimulator);
                    } else if ("ALLSECRETS".equals(mode)) {
                        runSimulationsAllSecret(sig, numOfSims, storage, tpoolsimulator);
                    } else {
                        runSimulationRandomSecrets(sig, numOfSims, storage, tpoolsimulator);
                    }

                    while (!tpoolsimulator.isTerminated()) {

                        log("Completed " + tpoolsimulator.getCompletedTaskCount() + "/" + submitted + " simulations...", 3);
                        if (tpoolsimulator.getCompletedTaskCount() >= submitted) {
                            log("Shutting down simulator", 3);
                            tpoolsimulator.shutdown();
                            try {
                                if (!tpoolsimulator.awaitTermination(waitTimeDefault, TimeUnit.MILLISECONDS)) { // 100
                                    tpoolsimulator.shutdownNow();
                                }
                            } catch (InterruptedException ex) {
                                Logger.getLogger(Simulator.class.getName()).log(Level.SEVERE, null, ex);
                            }
                        } else {
                            try {
                                Thread.sleep(waitTimeDefault); // 500
                            } catch (InterruptedException ex) {
                                Logger.getLogger(Simulator.class.getName()).log(Level.SEVERE, null, ex);
                            }
                        }

                    }

                    storage.sendResults(postDistMan);
                }
            }

            if (i < rounds - 1) {
                int roundsLeft = rounds - i - 1;
                simNumber = postDistMan.computeSampleSizesAdaptively(roundsLeft);
            }
        }
        log("Total simulations ran: " + numSimsDone, 2);
    }

    private void runSimulationsAllSecret(SimulationSignature sig, int num, SimulationStorage storage, ThreadPoolExecutor tpoolsim) {
        VariableList privVars = sig.getPrivVars();
        Iterable[] iters = new Iterable[privVars.size()];
        BigInteger totalsize = BigInteger.ONE;
        for (int i = 0; i < privVars.size(); i++) {
            PrivateVariable pvar = (PrivateVariable) privVars.get(i);
            iters[i] = pvar.getInterval();
            totalsize = totalsize.multiply(pvar.getIntervalSize());
        }
        log("Private variables have " + totalsize.toString() + " possible value combinations for simulations", 4);
        int simPerSecretValue = num / (totalsize.intValue());
        if (simPerSecretValue < 1) {
            simPerSecretValue = 1;
        }
        log("Running " + simPerSecretValue + " simulations per private variable value combination.", 4);
        for (BigIntegerCartesianIterator cartiter = new BigIntegerCartesianIterator(iters); cartiter.hasNext();) {
            // creating initial state
            MarkovChainState initState = new MarkovChainState("temp", BigDecimal.ONE);
            try {
                initState = sig.getInitialState().clone();
            } catch (CloneNotSupportedException ex) {
                Logger.getLogger(Simulator.class.getName()).log(Level.SEVERE, null, ex);
                System.exit(1);
            }
            log("Creating secret value:", 4);
            BigInteger values[] = cartiter.next();
            // giving a single value to each private variable
            for (int i = 0; i < privVars.size(); i++) {

                BigInteger value = values[i];
                String valueStr = value.toString();
                log(privVars.get(i).getName() + ":=" + valueStr, 4);
                initState.setPrivateVariable(privVars.get(i).getName(), '[' + valueStr + ',' + valueStr + ']');
            }

            // submitting simulations
            for (int j = 0; j < simPerSecretValue; j++) {
                log("Submitting simulation number " + j + " for secret " + initState.getSecretSingleIdentifier(), 3);
                MarkovChainState tempState = new MarkovChainState("temp", BigDecimal.ONE);
                try {
                    tempState = initState.clone();
                } catch (CloneNotSupportedException ex) {
                    Logger.getLogger(Simulator.class.getName()).log(Level.SEVERE, null, ex);
                    System.exit(1);
                }
                submitted++;
                numSimsDone++;
                tpoolsim.submit(new RunnableSingleSimulator(tempState, codelines, storage));
            }

        }

    }

    private void runSimulationRandomSecrets(SimulationSignature sig, int num, SimulationStorage storage, ThreadPoolExecutor tpoolsimulator) {
        log("Running " + num + " simulations using random private variable value combinations.", 2);
        log("The signature is " + sig.getInitialState().printNice(), 2);
        for (int i = 0; i < num; i++) {
            // creating initial state
            MarkovChainState initState = new MarkovChainState("temp", BigDecimal.ONE);
            try {
                initState = sig.getInitialState().clone();
            } catch (CloneNotSupportedException ex) {
                Logger.getLogger(Simulator.class.getName()).log(Level.SEVERE, null, ex);
                System.exit(1);
            }
            log("Creating secret value:", 4);
            VariableList values = initState.getVariables().getRandomAssignment();
            initState.setVariables(values);
            submitted++;
            numSimsDone++;
            log("Submitting simulation number " + i + " with variables " + initState.printNice(), 3);

            tpoolsimulator.submit(new RunnableSingleSimulator(initState, codelines, storage));
        }
    }

    private void runAbstractSimulation(SimulationSignature sig, int num, SimulationStorage storage, ThreadPoolExecutor tpoolsimulator) {
        log("Running " + num + " simulations using abstract simulation.", 2);
        log("The signature is " + sig.getInitialState().printNice(), 2);
        MarkovChainState fixedInitState = new MarkovChainState("temp", BigDecimal.ONE);
        try {
            fixedInitState = sig.getInitialState().clone();
        } catch (CloneNotSupportedException ex) {
            Logger.getLogger(Simulator.class.getName()).log(Level.SEVERE, null, ex);
            System.exit(1);
        }
        log("Creating secret value:", 4);
        VariableList values = fixedInitState.getVariables().getRandomAssignment();
        fixedInitState.setVariables(values);
        storage.storeSecretIdentifierForAbstraction(fixedInitState.getSecretSingleIdentifier());
        for (int i = 0; i < num; i++) {
            // creating initial state
            MarkovChainState initState = new MarkovChainState("temp", BigDecimal.ONE);
            try {
                initState = fixedInitState.clone();
            } catch (CloneNotSupportedException ex) {
                Logger.getLogger(Simulator.class.getName()).log(Level.SEVERE, null, ex);
                System.exit(1);
            }

            submitted++;
            numSimsDone++;
            log("Submitting simulation number " + i + " with variables " + initState.printNice(), 3);

            tpoolsimulator.submit(new RunnableSingleSimulator(initState, codelines, storage));
        }
    }

    private static void log(Object aObject, int level) {
        if (logger.getVerbosity() >= level) {
            logger.log(String.valueOf(aObject), level);
        }
    }
}
