/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/
    
package hyleak;

import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collections;
import support.Logger;


/**
 *
 * @author  
 */
public class VariableList extends ArrayList<Variable>{
    private static final Logger logger = Logger.getInstance();

    public VariableList() {
    }
    
    
 
    
    public Variable getVariable(String name){
        Variable toreturn = null;
        for (Variable temp : this) {
            if (temp.getName().equals(name)){
                toreturn = temp;
                break;
            }
        }
        //if (toreturn==null){
        //    log("WARNING: asked to get non-existent variable "+name+" from state "+this.getName()+", returning null instead");
        //}
        return toreturn;
    } 
    
    public MarkovStateDistributionVariable getDistributionVariable(String name){
        MarkovStateDistributionVariable toreturn = null;
        for (Variable temp : this) {
            if (temp.getName().equals(name)){
                toreturn =(MarkovStateDistributionVariable) temp;
                break;
            }
        }
        
        return toreturn;
    }
    
    public boolean deleteVariableByName(String name){
        Variable var = this.getVariable(name);
        if (var==null) {            
            log("WARNING: attempted to delete non-existent variable "+name,1);
            return false;
        }else{
            
            return this.deleteVariable(var);
        }
            
    }
    public boolean deleteVariable(Variable var){
        if (this.contains(var)){
            this.remove(var);
            return true;
        }else{
            log("WARNING: attempted to delete non-existent variable "+var.getName(),1);
            return false;
        }
    }
    
    public boolean setPublicVariable(String varname, BigInteger newval) {
        Variable var = this.getVariable(varname);
        if (var==null){
            log("ERROR: attempt to assign value "+newval+" to variable "+varname+", but the variable does not exist",1);
            return false;
        }else if (var.isPrivate()){
            log("ERROR: attempt to assign value "+newval+" to variable "+varname+", but the variable is private",1);
            return false;
        }else{
            PublicVariable pubvar =(PublicVariable) var;
            pubvar.setVal(newval);
            return true;
        }
    }
    
    public boolean setPublicVariable(String varname, long newval){
       return this.setPublicVariable(varname, BigInteger.valueOf(newval));
    }
    
    public BigInteger getPublicVariableValue(String varname){
        Variable var = this.getVariable(varname);
        if (var==null){
            log("ERROR: attempt to get value of public variable "+varname+", but the variable does not exist",1);
            return BigInteger.valueOf(-1);
        }else if (var.isPrivate()){
            log("ERROR: attempt to get value of public variable "+varname+", but the variable is private",1);
            return BigInteger.valueOf(-1);
        }else{
            PublicVariable pubvar =(PublicVariable) var;
            return pubvar.getVal();
        }
    }
    
    public Interval getPrivateVariableInterval(String varname){
        Variable var = this.getVariable(varname);
        if (var==null){
            log("ERROR: attempt to get interval of private variable "+varname+", but the variable does not exist",1);
            return null;
        }else if (var.isPrivate()==false){
            log("ERROR: attempt to get interval of private variable "+varname+", but the variable is public",1);
            return null;
        }else{
            PrivateVariable privvar =(PrivateVariable) var;
            return privvar.getInterval();
        }
    }
    
    public String getVariableName(int i){
        return this.get(i).getName();
    }
    
    public Integer getVariableLength(int i){
        return this.get(i).getLength();
    }
    
    public String getIdentifier(){
        this.sort();
        return this.toString();
    }    
    
    
    public VariableList clone(){
    
        VariableList toreturn =new VariableList();
        for (Variable myvar : this) {
          
            if (myvar instanceof PrivateVariable){
                PrivateVariable myprivvar = (PrivateVariable)myvar;
                PrivateVariable newprivvar = new PrivateVariable(myprivvar.getName(),myprivvar.getInterval().toString(),myprivvar.getLength());
                newprivvar.setSecret(myprivvar.isSecret());
                toreturn.add(newprivvar);
            }else if (myvar instanceof PublicVariable){
                PublicVariable mypubvar = (PublicVariable)myvar;
                PublicVariable newpubvar = new PublicVariable(mypubvar.getName(),mypubvar.getVal(),mypubvar.getLength());
                newpubvar.setObservable(mypubvar.isObservable());
                toreturn.add(newpubvar);
            }else if (myvar instanceof MarkovStatePublicDistributionVariable){
                MarkovStatePublicDistributionVariable mypubvar = (MarkovStatePublicDistributionVariable)myvar;
                MarkovStatePublicDistributionVariable newpubvar = new MarkovStatePublicDistributionVariable(mypubvar);
                toreturn.add(newpubvar);
            }else if (myvar instanceof MarkovStatePrivateDistributionVariable){
                MarkovStatePrivateDistributionVariable myprivvar = (MarkovStatePrivateDistributionVariable)myvar;
                MarkovStatePrivateDistributionVariable newprivvar = new MarkovStatePrivateDistributionVariable(myprivvar);
                toreturn.add(newprivvar);
            }
        }
        return toreturn;
        
    }
   
    
    @Override
    public String toString(){
        String s="";
        for (Variable var : this) {
            s+=var.toString()+";";
        }
        return s;
    }
    
    public String toNonIntervalString(){
        String s="";
        for (Variable var : this) {
            s+=var.toNonIntervalString()+";";
        }
        return s;
    }
    
    private static void log(Object aObject, int level){
        if (logger.getVerbosity() >= level) {
            logger.log(String.valueOf(aObject),level);
        }
    }

    

    public boolean setPrivateVariable(String varname, String interval) {
        Variable var = this.getVariable(varname);
        if (var==null){
            log("ERROR: attempt to assign interval "+interval+" to variable "+varname+", but the variable does not exist",1);
            return false;
        }else if (!var.isPrivate()){
            log("ERROR: attempt to assign interval "+interval+" to variable "+varname+", but the variable is public",1);
            return false;
        }else{
            PrivateVariable privvar =(PrivateVariable) var;
            privvar.setInterval(interval);
            return true;
        }
    }

    private void sort() {
        Collections.sort(this, new MarkovStateVariableComparator());
    }

    public String getSingleIdentifier() {
        this.sort();
        return this.toNonIntervalString();
    }

    VariableList getRandomAssignment() {
        VariableList toreturn = new VariableList();
        for (Variable v : this){
            if (!v.isPrivate()){
                PublicVariable pubvar = (PublicVariable) v;
                PublicVariable newpubvar = new PublicVariable(pubvar.getName(),pubvar.getVal(),pubvar.getLength());
                newpubvar.setObservable(pubvar.isObservable());
                toreturn.add(newpubvar);
                log("Added public variable "+newpubvar.getName()+" with value "+newpubvar.getVal(),4);
            }else{
                PrivateVariable privvar = (PrivateVariable) v;
                PrivateVariable newprivvar = new PrivateVariable(privvar.getName(),privvar.getInterval().RandomlyChoose(),privvar.getLength());
                newprivvar.setSecret(privvar.isSecret());
                toreturn.add(newprivvar);
                log("Added private variable "+newprivvar.getName()+" with interval "+newprivvar.getInterval(),4);
            }
        }
        return toreturn;
    }

    
    
}

    
    
    

