/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/
package preprocessor;

import java.math.BigInteger;

public class BigIntegerInterval {
      
  private BigInteger min;
  private BigInteger max;
  
  public BigIntegerInterval(BigInteger min, BigInteger max) {
    this.min = min;
    this.max = max;
  }
  
  public BigInteger getMin() {
    return min;
  }

  public BigInteger getMax() {
    return max;
  }
  
  public BigInteger getSize() {
    return max.subtract(min).add(BigInteger.ONE);
  }
  
  public boolean isZero() {
    if (min.signum()<0)
      System.err.println("WARNING: range estimation does not work with negative values.");
    return max.signum()==0; // only works if min and max are positive values
  }
  
  public boolean isStriclyPositive() {
    return min.signum() > 0;
  }
  
  /** 2 if this > i
   *  1 if this >= i
   *  0 if this and i overlapp
   *  -1 if i >= this
   *  -2 if i > this
   */
  public int compareTo(BigIntegerInterval i) {
    if (max.compareTo(i.getMin()) < 0)
      return -2;
    else if (max.compareTo(i.getMin()) == 0)
      return -1;
    else if (min.compareTo(i.getMax()) == 0)
      return 1;      
    else if (min.compareTo(i.getMax()) > 0)
      return 2; 
    else
      return 0;
  }
  
  public boolean intersect(BigIntegerInterval i) {
    return (compareTo(i)>=-1 && compareTo(i)<=1);
  }
    
  public BigInteger size() {
    return max.subtract(min);
  }
  
  public BigIntegerInterval add(BigIntegerInterval i) {
    return new BigIntegerInterval(getMin().add(i.getMin()), getMax().add(i.getMax()));
  }
  
  public BigIntegerInterval subtract(BigIntegerInterval i) {
    return new BigIntegerInterval(getMin().subtract(i.getMax()), getMax().subtract(i.getMin()));
  }
  
  public BigIntegerInterval multiply(BigIntegerInterval i) {
    BigInteger prod1 = getMin().multiply(i.getMin());
    BigInteger prod2 = getMin().multiply(i.getMax());
    BigInteger prod3 = getMax().multiply(i.getMin());
    BigInteger prod4 = getMax().multiply(i.getMax());
    BigInteger prod_min = prod1.min(prod2).min(prod3).min(prod4);
    BigInteger prod_max = prod1.max(prod2).max(prod3).max(prod4);
    return new BigIntegerInterval(prod_min,prod_max);
  }
  
  public BigIntegerInterval divide(BigIntegerInterval i) {
    if (i.isStriclyPositive()) {
      BigInteger prod1 = getMin().divide(i.getMin());
      BigInteger prod2 = getMin().divide(i.getMax());
      BigInteger prod3 = getMax().divide(i.getMin());
      BigInteger prod4 = getMax().divide(i.getMax());
      BigInteger prod_min = prod1.min(prod2).min(prod3).min(prod4);
      BigInteger prod_max = prod1.max(prod2).max(prod3).max(prod4);
      return new BigIntegerInterval(prod_min,prod_max);
    } else {
      System.err.println("WARNING: division with a variable whose range may contain zero");
      return new BigIntegerInterval(BigInteger.ZERO,BigInteger.ZERO);
    }
  }
  
  public BigIntegerInterval negate() {
    return new BigIntegerInterval(getMax().negate(), getMin().negate());
  }
  
  public String toString() {
      return new String("[" + min + "," + max + "]");
  }
  
  
}
