/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/

package preprocessor;

import java.util.HashSet;
import java.util.LinkedList;

public class IfStatement extends Statement {

    private String guard;
    
    private GotoStatement thenGoto;
    
    private GotoStatement elseGoto;
    
    private LinkedList<Statement> thenStatements;
    
    private int nbThenStatements;
    
    private GotoStatement exitGoto;
    
    private LinkedList<Statement> elseStatements;
    
    private int nbElseStatements;
    
    public IfStatement() {
    	usedVariables = new HashSet<>();
    	thenStatements = new LinkedList<>();
    	elseStatements = new LinkedList<>();
    	nbStatements = 4;
    	nbThenStatements = 0;
    	nbElseStatements = 0;
    }
    
    public IfStatement(String g) {
       	usedVariables = new HashSet<Variable>();
    	thenStatements = new LinkedList<Statement>();
    	elseStatements = new LinkedList<Statement>();
        guard = g;
        nbStatements = 4;
    	nbThenStatements = 0;
    	nbElseStatements = 0;
    }
    
    public void setGuard(String g) {
    	guard = g;
    }
    
    public void setThenGoto(GotoStatement g) {
    	thenGoto = g;
    }

    public void setElseGoto(GotoStatement g) {
    	elseGoto = g;
    }
    
    public void addThenStatement(Statement st) {
    	thenStatements.add(st);
    	nbStatements += st.getNbStatements();
    	nbThenStatements += st.getNbStatements();
    }
    
    public void setExitGoto(GotoStatement g) {
    	exitGoto = g;
    }
    
    public void addElseStatement(Statement st) {
    	elseStatements.add(st);
    	nbStatements += st.getNbStatements();
    	nbElseStatements += st.getNbStatements();
    }
    
    public LinkedList<Statement> getThenStatements() {
    	return thenStatements;
    }
    
    public LinkedList<Statement> getElseStatements() {
    	return elseStatements;
    }
    
    public int getNbThenStatements() {
    	return nbThenStatements;
    }
    
    public int getNbElseStatements() {
    	return nbElseStatements;
    }
    
   
    @Override
    public Statement replace(String oldv, String newv) {
    	IfStatement newIfSt = new IfStatement();
        String newGuard = guard.replace("("+oldv+")","("+newv+")");
        newGuard = newGuard.replace("("+oldv+"[","("+newv+"[");
        newIfSt.setGuard(newGuard);
        newIfSt.setThenGoto(thenGoto.replace(oldv,newv));
        newIfSt.setElseGoto(elseGoto.replace(oldv,newv));
        for(Statement st : thenStatements) {
        	newIfSt.addThenStatement(st.replace(oldv,newv));
        }
        newIfSt.setExitGoto(exitGoto.replace(oldv,newv));
        for(Statement st : elseStatements) {
        	newIfSt.addElseStatement(st.replace(oldv,newv));
        }
        newIfSt.addUsedVariable(getUsedVariables());
        for (Variable v : newIfSt.getUsedVariables()) {
        	v.setLastUsed(newIfSt);
        }
        return newIfSt;
    }
    
    @Override
    public String toString(int line) {
        String ret = "if (" + guard + ")\n";
        line++;
        ret += "then " + thenGoto.toString(line);
        line++;
        ret += "else " + elseGoto.toString(line);
        line++;
        for (Statement st : thenStatements) {
        	ret += st.toString(line);
        	line += st.getNbStatements();
        }
        ret += exitGoto.toString(line);
        line++;
        for (Statement st : elseStatements) {
        	ret += st.toString(line);
        	line += st.getNbStatements();
        }
        return ret;
    }
}
