/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/

package preprocessor;

import java.util.LinkedList;
import java.util.ListIterator;

public class P3ListStatement extends P3Statement {
    
    protected LinkedList<P3Statement> list_statement;
    
    public P3ListStatement(Ranges r) {
	super("", false, false, r);
	this.list_statement = new LinkedList<P3Statement>();
    }
    
    public void addFirst(P3Statement s) {
	list_statement.addFirst(s);
    }
        
    public void add(P3Statement s) {
	list_statement.add(s);
    }
    
    public void addAll(P3ListStatement l) {
	list_statement.addAll(l.list_statement);
    }
    
    public ListIterator<P3Statement> listIterator(int index) {
	return list_statement.listIterator(index);
    }
    
    public Ranges getGlobalRanges(Ranges global_ranges) {
	for(P3Statement st : list_statement) {
	     global_ranges = global_ranges.union_large(st.getGlobalRanges(global_ranges));
	}
	return global_ranges;   	
    }
    
    public void updateRanges(Ranges global_ranges) {
	this.ranges.updateVariableRange(global_ranges);
	for(P3Statement st : list_statement) {
	    st.updateRanges(global_ranges);
	}
    }
    
    public void updateSecDep(boolean previousSecDep) { // if true set all statement sec_dep to true
	ListIterator<P3Statement> it = list_statement.listIterator(list_statement.size());
	boolean lastSecDep = false;  // when true set all statement sec_dep to true until a conditional
	while(it.hasPrevious()) {
	    P3Statement st = it.previous();
	    st.updateSecDep(previousSecDep || lastSecDep);
	    if (st.isSecDep()) {
		sec_dep = true;
		if (!st.isConditional())
		    lastSecDep = true;
	    }
	    else if(lastSecDep) {
		if (st.isConditional())
		    lastSecDep  = false;
		else
		    st.setSecDep(true);
	    }
	    else if(previousSecDep) {
		st.setSecDep(true);
	    }
	}
    }
    
    public Pair<String, Boolean> generateCode(boolean is_simulated) {
	String prog = "";
	ListIterator<P3Statement> it = list_statement.listIterator(0);
	boolean new_is_simulated_branch = false;
	boolean new_is_simulated = true;
	boolean conditonal = false;
	boolean else_branch = false;
	while(it.hasNext()) {
	    P3Statement st = it.next();
	    Pair<String, Boolean> ret = st.generateCode(is_simulated || new_is_simulated_branch);
	    prog += ret.getLeft();
	    if (ret.getRight())
		new_is_simulated_branch = true;
	    if (st.isConditional()) {
		if (!conditonal) {	// first if or while
		    conditonal = true;
		}
		else if (st.isElse()) {
		    else_branch = true;
    		    new_is_simulated = new_is_simulated && new_is_simulated_branch;
		    new_is_simulated_branch = false;
		}
		else {			// else if or fi
		    new_is_simulated = new_is_simulated && new_is_simulated_branch;
		    new_is_simulated_branch = false;
		}
	    }
	}
	return new Pair<String, Boolean>(prog, !is_simulated && new_is_simulated && ((!conditional && new_is_simulated_branch) || else_branch));
	// not simulated before, with a new simulate added in all branches, either in a non if/while conditional, or in a if-else alternative  
    }
    
    public String print() {
	String ret = "";
	ListIterator<P3Statement> it = list_statement.listIterator(0);
	while(it.hasNext()) {
	    P3Statement st = it.next();
	    ret += st.print();
	}
	return ret;
    }

    
}
