/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/

package preprocessor;

import support.Logger;

public class P3Statement {
    private static final Logger logger = Logger.getInstance();

    protected String instruction;
    protected boolean sec_dep; 
    protected boolean conditional;
    protected boolean iselse;
    protected Ranges ranges; 
    
    public P3Statement(String ins, boolean cond, boolean e, Ranges r) {
	this.instruction = ins;
	this.sec_dep = false;
	this.conditional = cond;
	this.iselse = e;
	this.ranges = r;
    }
        
    public String getInstruction() {
	return instruction;
    }
    
    public boolean isSecDep() {
	return sec_dep;
    }
    
    public void setSecDep(boolean b) {
	this.sec_dep = b;
    }
    
    public void updateSecDep(boolean previousSecDep) { }
    
    public boolean isConditional() {
	return conditional;
    }
    
    public boolean isElse() {
	return iselse;
    }
    
    public Ranges getRanges() {
	return ranges;
    }
    
    public void setRanges(Ranges r) {
	this.ranges = r;
    }
    
    public Ranges getGlobalRanges(Ranges global_ranges) {
	return ranges;
    }
    
    public void updateRanges(Ranges global_ranges) {
	this.ranges.updateVariableRange(global_ranges);	
    }

    public Pair<String, Boolean> generateCode(boolean is_simulated) {
	String prog = "";
	logger.log("DEBUG: generating code of statement " + instruction.replace('\n', ' '), 4);
	logger.log("DEBUG: range is " + ranges + " ; is simulated: " + is_simulated + " ; is sec dep: " + sec_dep, 4);
	if (!conditional && !is_simulated && ranges.simulateOk()) {
	    logger.log("DEBUG: adding simulate statement", 4);
	    if (sec_dep)
		prog += "simulate;\n";
	    else
		prog += "simulate-abs;\n";
	    is_simulated = true;
	}
	prog += instruction;
	return new Pair<String, Boolean>(prog, is_simulated);
    }
    
    public String print() {
	if (sec_dep)
	    return instruction.replace('\n', '\t') + ranges + "; SEC DEP\n";
	else
	    return instruction.replace('\n', '\t') + ranges +"\n";
    }
}
