/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/

package preprocessor;

import java.io.PrintWriter;
import java.math.BigDecimal;
import java.util.HashMap;
import org.antlr.runtime.*;
import support.Logger;

/** - replace constants
 *  - replace expressions by numbers where they are required (initialization).
 *  - check variable existance
 *  - check authorized usage: exlusion of private variable, sums, differences, multiplication, division, etc, in left members of guards
 *
 */
public class Preprocessor1 {
    public static void preprocess(String filename, HashMap<String,BigDecimal> constants) throws Exception {
	Logger.getInstance().log("==== Preprocessor 1 ====",3);
    	Programme prog = new Programme(constants);
        Preprocessor1Lexer lexer = new Preprocessor1Lexer(new ANTLRFileStream(filename));
        CommonTokenStream tokens = new CommonTokenStream(lexer);
        Preprocessor1Parser parser = new Preprocessor1Parser(tokens);
        String result = parser.prog_grammar(prog);
	prog.checkConstants();
        
        try(  PrintWriter out = new PrintWriter(filename.concat(".1.pp")) ) {
	    out.println( result );
	}
    }
    
    public static void main(String[] args) throws Exception {
        preprocess(args[0],null);
    }
}
