/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/

grammar Preprocessor3;

@parser::header {
    package preprocessor;
    
    import java.math.BigInteger;
    import java.math.BigDecimal;
    import java.util.LinkedList;
    import java.util.List;
    import support.Logger;
}

@lexer::header {
    package preprocessor; 
}

@members {
    Programme prog;
    Ranges init_ranges;
    P3ListStatement statements;
    Logger logger = Logger.getInstance();

    class AssignExprReturn {
	public BigIntegerInterval i;
	public boolean sec_dependent;
    }
}

prog_grammar[Programme p] 		@init	{ prog=p;
						  init_ranges = new Ranges();
						}:
		(decl)*
		st_list[init_ranges,false]	{ statements = $st_list.p3stlist; }
		EOF
		;

decl:	public_decl | private_decl;
	  
public_decl:	public_type INT IDENTIFIER num_indices init_val ';'
		{   int sizeInt = Integer.parseInt($INT.text.substring(3)); 
		    PublicVariable v = new PublicVariable(sizeInt);
		    v.setPublicity($public_type.b);
		    v.setName($IDENTIFIER.text + $num_indices.text);
		    prog.addPublicVariable(v.getName(),v);
		    prog.addDeclaration(v.getName());
		    if (!$init_val.l.isEmpty()) {
			v.updateInitialValues($init_val.l);
			init_ranges.addVariableRange(v, new BigIntegerInterval($init_val.l.getFirst(), $init_val.l.getFirst()));
		    }
// 		    else {
			//init_ranges.addVariableRange(v, new BigIntegerInterval(BigInteger.ZERO, v.getMaxInt()));
// 		    }
		    
		    // add a variable for arrays
		    if (!$num_indices.l.isEmpty()) {
			v = prog.getPublicVariable($IDENTIFIER.text);
			if (v == null) { // create a new variable
			    v = new PublicVariable(sizeInt);
			    v.setPublicity($public_type.b);
			    v.setName($IDENTIFIER.text);
			    prog.addPublicVariable($IDENTIFIER.text,v);
			    for (Integer index : $num_indices.l) {
				v.addDimension(new ArrayDimension(index,index));
			    }
			}
			else {	// update an existing variable
			    int j = 0;
			    for (Integer index : $num_indices.l) {
				ArrayDimension dim = v.getArrayDimension(j);
				dim.setMinIndex(Math.min(dim.getMinIndex(),index));
				dim.setMaxIndex(Math.max(dim.getMaxIndex(),index));
				j++;
			    }
			}
		    } 		    
		}
	        ;

public_type returns [boolean b]:
		'public' 		{ $b = true; }
		| 'observable' 		{ $b = false; }
		;
	   
private_decl:	private_type INT IDENTIFIER num_indices init_intervals ';'
		{   int sizeInt = Integer.parseInt($INT.text.substring(3));
		    PrivateVariable v = new PrivateVariable(sizeInt);
	    	    v.setPublicity($private_type.b);
		    v.setName($IDENTIFIER.text + $num_indices.text);
	    	    prog.addPrivateVariable(v.getName(),v);
		    prog.addDeclaration(v.getName());
		    if (!$init_intervals.l.isEmpty()) {
			v.updateIntRange($init_intervals.l);
			init_ranges.addVariableRange(v, new BigIntegerInterval($init_intervals.l.getFirst(), $init_intervals.l.getLast()));
		    }
		    else {
			BigInteger max = new BigInteger("2");
			max = max.pow(sizeInt);
			max = max.subtract(BigInteger.ONE);
			init_ranges.addVariableRange(v, new BigIntegerInterval(BigInteger.ZERO, max));
		    }
		    
		    // add a variable for arrays
		    if (!$num_indices.l.isEmpty()) {
			v = prog.getPrivateVariable($IDENTIFIER.text);
			if (v == null) {  // create a new variable
			    v = new PrivateVariable(sizeInt);
			    v.setPublicity($private_type.b);
			    v.setName($IDENTIFIER.text);
			    prog.addPrivateVariable($IDENTIFIER.text,v);
			    for (Integer index : $num_indices.l) {
				v.addDimension(new ArrayDimension(index,index));
			    }
			}
			else { 	// update an existing variable
			    int j = 0;
			    for (Integer index : $num_indices.l) {
				ArrayDimension dim = v.getArrayDimension(j);
				dim.setMinIndex(Math.min(dim.getMinIndex(),index));
				dim.setMaxIndex(Math.max(dim.getMaxIndex(),index));
				j++;
			    }
			}
		    }
		}
		;
	      
private_type returns [boolean b]:
	    'private'		{ $b = true; }
	    | 	'secret' 	{ $b = false; }
	    ;
	    
num_indices returns [LinkedList<Integer> l] @init{ $l = new LinkedList<Integer>(); }:
    ('[' NUMBER ']' 		{ $l.add(new Integer(Integer.parseInt($NUMBER.text))); }
    )*
    ;
	    
init_val returns [LinkedList<BigInteger> l] @init{ l = new LinkedList<BigInteger>(); } :
								// empty
	| a=':=' val=NUMBER					{   BigInteger int_val = new BigInteger($val.text);
								    $l.add(int_val);
								}
	;

init_intervals returns [LinkedList<BigInteger> l] @init{ l = new LinkedList<BigInteger>(); } :
	// empty
	| a=':=' (interval		  			{ $l.addAll($interval.l); }
		 )+
	;
	
interval returns [LinkedList<BigInteger> l]:
	a='[' min=NUMBER ',' max=NUMBER ']'			{   $l = new LinkedList<BigInteger>();
								    $l.add(new BigInteger($min.text));
								    $l.add(new BigInteger($max.text));
								}
	;
	
st_list [Ranges r_in, boolean inwhile] returns [P3ListStatement p3stlist] @init{ Ranges r = $r_in; p3stlist = new P3ListStatement(r); }: 
						( st[r,inwhile]		       { p3stlist.add($st.p3st); r = $st.p3st.getRanges(); p3stlist.setRanges(r); }	
						)*
						;
	
st [Ranges r_in, boolean inwhile] returns [P3Statement p3st]:
						assignement_st[r_in,inwhile]		{ $p3st = $assignement_st.p3st; }
						| print_st[r_in]			{ $p3st = $print_st.p3st; }
						| return_st[r_in]			{ $p3st = $return_st.p3st; }
						| random_st[r_in,inwhile]		{ $p3st = $random_st.p3st; }
						| while_st[r_in]			{ $p3st = $while_st.p3stlist; }
						| if_st[r_in,inwhile]			{ $p3st = $if_st.p3stlist; }
						;
  
assignement_st [Ranges r_in, boolean inwhile] returns [P3Statement p3st]:
		'assign' var=variable[r_in] ':=' e=assign_expr[r_in] ';'		{ Ranges r_out = new Ranges(r_in);
											  for (Variable v : $variable.lvar) {
											    if (inwhile)
												r_out.modifyVariableRange(v,new BigIntegerInterval(BigInteger.ZERO,v.getMaxInt()));
											    else
												r_out.modifyVariableRange(v,$assign_expr.s.i);
											  }
											  $p3st = new P3Statement("assign " + $var.text + " := " + $e.text + ";\n", false, false, r_out);
											}
		;
		
print_st[Ranges r_in] returns [P3Statement p3st]:
		'print' ';' 	{ $p3st = new P3Statement("print;\n", false, false, $r_in); }	
		;
		
return_st [Ranges r_in]  returns [P3Statement p3st]:
		'return' ';' 	{ $p3st = new P3Statement("return;\n", false, false, $r_in); }
		| 'return' 'after' 'simulate' 
		  var1=variable[r_in] ':=' filename '(' var2=variable[r_in] ')' ';'	{ $p3st = new P3Statement("return after simulate " + $var1.text + " := " + $filename.text + " (" + $var2.text + ");\n", false, false, $r_in); }
		;

random_st [Ranges r_in, boolean inwhile] returns [P3Statement p3st] @init { String ret = ""; Ranges r_out = new Ranges(r_in); }:
		ran='random' var=variable[r_in] ':='					{ ret = "random " + $var.text + " := "; }
		( 'randombit' '(' float_number ')'					{ for (Variable v : $variable.lvar) {
											      r_out.modifyVariableRange(v,new BigIntegerInterval(BigInteger.ZERO, BigInteger.ONE));
											  }
											  ret += "randombit(" + $float_number.f + ");\n";
											}
		| 'random' '(' e=assign_expr[r_in] ',' f=assign_expr[r_in] ')'		{ for (Variable v : $variable.lvar) {
											    if (inwhile)
												r_out.modifyVariableRange(v,new BigIntegerInterval(BigInteger.ZERO,v.getMaxInt()));
											    else
												r_out.modifyVariableRange(v,new BigIntegerInterval($e.s.i.getMin(), $f.s.i.getMax()));
											  }
											  ret += "random(" + $e.text + "," + $f.text + ");\n"; 
											}
		) ';'									{ $p3st = new P3Statement(ret, false, false, r_out); }								
		;

while_st [Ranges r_in] returns [P3ListStatement p3stlist]:
		'while' '(' assign_expr[r_in] ')' 'do' st_list[r_in,true] 'od'
		{ p3stlist = $st_list.p3stlist;
		  P3Statement guard = new P3Statement("while(" + $assign_expr.text + ") do\n", true, false, $r_in);
		  if ($assign_expr.s.sec_dependent)
		      guard.setSecDep(true);
		  p3stlist.addFirst(guard);
		  P3Statement closing = new P3Statement("od\n", true, false, p3stlist.getRanges());
		  p3stlist.add(closing);
		}
		;

if_st [Ranges r_in, boolean inwhile] returns [P3ListStatement p3stlist] @init{ int nbVariables = r_in.getNbVariables(); }:	
		if_case[r_in, inwhile] 			{ p3stlist = $if_case.p3stlist; 
							  nbVariables = $if_case.p3stlist.getRanges().getNbVariables();
							}
		( elif_case[r_in, inwhile]		{ p3stlist.addAll($elif_case.p3stlist);
							  p3stlist.setRanges(p3stlist.getRanges().union_large($elif_case.p3stlist.getRanges())); 
							  if (nbVariables != $elif_case.p3stlist.getRanges().getNbVariables())
							      logger.log("WARNING: on line " + f.getLine() + ", some variable has not been initialised in one branch of an if condition.",1);
							}
		)*
		( e=else_case[r_in, inwhile]		{ p3stlist.addAll($else_case.p3stlist);
							  p3stlist.setRanges(p3stlist.getRanges().union_large($else_case.p3stlist.getRanges())); 
							  if (nbVariables != $else_case.p3stlist.getRanges().getNbVariables())
							      logger.log("WARNING: on line " + f.getLine() + ", some variable has not been initialised in one branch of an if condition.",1);
							}
		)? f='fi'				{ if (e == null) {
							      p3stlist.setRanges(p3stlist.getRanges().union_large($r_in));
							      if (nbVariables != $r_in.getNbVariables())
								  logger.log("WARNING: on line " + f.getLine() + ", some variable has not been initialised in one branch of an if condition.",1);
							  }
							  P3Statement closing = new P3Statement("fi\n", true, false, p3stlist.getRanges());
							  p3stlist.add(closing);
							}				
		;

if_case [Ranges r_in, boolean inwhile] returns [P3ListStatement p3stlist]:
		'if' '(' assign_expr[r_in] ')' 'then' st_list[r_in, inwhile]							
		{ p3stlist = $st_list.p3stlist;
		  P3Statement guard = new P3Statement("if(" + $assign_expr.text + ") then\n", true, false, $r_in);
		  if ($assign_expr.s.sec_dependent)
		      guard.setSecDep(true);
		  p3stlist.addFirst(guard);
		}
		;
      

else_case [Ranges r_in, boolean inwhile] returns [P3ListStatement p3stlist]:
		'else' st_list[r_in, inwhile]
		{ p3stlist = $st_list.p3stlist;
		  P3Statement guard = new P3Statement("else\n", true, true, $r_in);
		  p3stlist.addFirst(guard);
		}
		;

elif_case [Ranges r_in, boolean inwhile] returns [P3ListStatement p3stlist]:	
		'elif' '(' assign_expr[r_in] ')' 'then' st_list[r_in, inwhile]
		{ p3stlist = $st_list.p3stlist;
		  P3Statement guard = new P3Statement("elif(" + $assign_expr.text + ") then\n", true, false, $r_in);
		  if ($assign_expr.s.sec_dependent)
		      guard.setSecDep(true);
		  p3stlist.addFirst(guard);
		}
		;

assign_expr [Ranges r] returns [AssignExprReturn s] :	 
		a=and_expr[r] 					{ $s = $a.s; }
		( '||' b=and_expr[r] 				{ $s.sec_dependent = $s.sec_dependent || $b.s.sec_dependent;
								  if ($s.i.isZero() && $b.s.i.isZero())
								    $s.i = new BigIntegerInterval(BigInteger.ZERO,BigInteger.ZERO);
								  else if ($s.i.isStriclyPositive() || $b.s.i.isStriclyPositive()) 	// only work if results are non negative
								    $s.i = new BigIntegerInterval(BigInteger.ONE,BigInteger.ONE);
								  else
								    $s.i = new BigIntegerInterval(BigInteger.ZERO,BigInteger.ONE);								 
								}
		)* ;
	   	   
and_expr [Ranges r] returns [AssignExprReturn s] :	 
		a=xor_expr[r]					{ $s = $a.s; }
		( '&&' b=xor_expr[r] 				{ $s.sec_dependent = $s.sec_dependent || $b.s.sec_dependent;
								  if ($s.i.isZero() || $b.s.i.isZero())
								    $s.i = new BigIntegerInterval(BigInteger.ZERO,BigInteger.ZERO);
								  else if ($s.i.isStriclyPositive() && $b.s.i.isStriclyPositive()) 	// only work if results are non negative
								    $s.i = new BigIntegerInterval(BigInteger.ONE,BigInteger.ONE);
								  else
								    $s.i = new BigIntegerInterval(BigInteger.ZERO,BigInteger.ONE);								 
								}
		)* ;

xor_expr [Ranges r] returns [AssignExprReturn s] :
		a=comp_expr[r]					{ $s = $a.s; }
		( '^' b=comp_expr[r]				{ $s.sec_dependent = $s.sec_dependent || $b.s.sec_dependent;
								  if ( ($s.i.isZero() && $b.s.i.isZero())
								       || ($s.i.isStriclyPositive() && $b.s.i.isStriclyPositive())) 		// only work if results are non negative
								    $s.i = new BigIntegerInterval(BigInteger.ZERO,BigInteger.ZERO);
								  else if ( ($s.i.isZero() || $b.s.i.isZero())
								            && ($s.i.isStriclyPositive() || $b.s.i.isStriclyPositive())) 	// only work if results are non negative
								    $s.i = new BigIntegerInterval(BigInteger.ONE,BigInteger.ONE);
								  else
								    $s.i = new BigIntegerInterval(BigInteger.ZERO,BigInteger.ONE);
								}
		)* ;

        
comp_expr [Ranges r] returns [AssignExprReturn s] :
		a=sum_expr[r]					{ $s = $a.s; }
		( '<=' b=sum_expr[r]				{ $s.sec_dependent = $s.sec_dependent || $b.s.sec_dependent;
								  int comp = $a.s.i.compareTo($b.s.i);
								  if (comp < 0)
								    $s.i = new BigIntegerInterval(BigInteger.ONE,BigInteger.ONE);
								  else if (comp > 1)
								    $s.i = new BigIntegerInterval(BigInteger.ZERO,BigInteger.ZERO);
								  else
								    $s.i = new BigIntegerInterval(BigInteger.ZERO,BigInteger.ONE);
								}
		| '>=' b=sum_expr[r]				{ $s.sec_dependent = $s.sec_dependent || $b.s.sec_dependent;
								  int comp = $b.s.i.compareTo($a.s.i);
								  if (comp < 0)
								    $s.i = new BigIntegerInterval(BigInteger.ONE,BigInteger.ONE);
								  else if (comp > 1)
								    $s.i = new BigIntegerInterval(BigInteger.ZERO,BigInteger.ZERO);
								  else
								    $s.i = new BigIntegerInterval(BigInteger.ZERO,BigInteger.ONE);
								}
		| '<'  b=sum_expr[r]				{ $s.sec_dependent = $s.sec_dependent || $b.s.sec_dependent;
								  int comp = $a.s.i.compareTo($b.s.i);
								  if (comp < -1)
								    $s.i = new BigIntegerInterval(BigInteger.ONE,BigInteger.ONE);
								  else if (comp > 0)
								    $s.i = new BigIntegerInterval(BigInteger.ZERO,BigInteger.ZERO);
								  else
								    $s.i = new BigIntegerInterval(BigInteger.ZERO,BigInteger.ONE);
								}
		| '>'  b=sum_expr[r]				{ $s.sec_dependent = $s.sec_dependent || $b.s.sec_dependent;
								  int comp = $b.s.i.compareTo($a.s.i);
								  if (comp < -1)
								    $s.i = new BigIntegerInterval(BigInteger.ONE,BigInteger.ONE);
								  else if (comp > 0)
								    $s.i = new BigIntegerInterval(BigInteger.ZERO,BigInteger.ZERO);
								  else
								    $s.i = new BigIntegerInterval(BigInteger.ZERO,BigInteger.ONE);
								}
		| '==' b=sum_expr[r]				{ $s.sec_dependent = $s.sec_dependent || $b.s.sec_dependent;
								  if ($a.s.i.intersect($b.s.i)) {
								    if ($a.s.i.size().signum() == 0 && $b.s.i.size().signum() == 0)
								      $s.i = new BigIntegerInterval(BigInteger.ONE,BigInteger.ONE);
								    else
								      $s.i = new BigIntegerInterval(BigInteger.ZERO,BigInteger.ONE);
								  }
								  else
								    $s.i = new BigIntegerInterval(BigInteger.ZERO,BigInteger.ZERO);
								}
		| '!=' b=sum_expr[r]				{ $s.sec_dependent = $s.sec_dependent || $b.s.sec_dependent;
								  if ($a.s.i.intersect($b.s.i)) {
								    if ($a.s.i.size().signum() == 0 && $b.s.i.size().signum() == 0)
								      $s.i = new BigIntegerInterval(BigInteger.ZERO,BigInteger.ZERO);
								    else
								      $s.i = new BigIntegerInterval(BigInteger.ZERO,BigInteger.ONE);
								  }
								  else
								    $s.i = new BigIntegerInterval(BigInteger.ONE,BigInteger.ONE);
								}
		)?
		;

sum_expr [Ranges r] returns [AssignExprReturn s] :
		a=prod_expr[r]					{ $s = $a.s; }
		( '+' b=prod_expr[r]				{ $s.i = $s.i.add($b.s.i);
								  $s.sec_dependent = $s.sec_dependent || $b.s.sec_dependent;
								}
		| '-' b=prod_expr[r]				{ $s.i = $s.i.subtract($b.s.i); 
								  $s.sec_dependent = $s.sec_dependent || $b.s.sec_dependent;
								}
		)*
		;

prod_expr [Ranges r] returns [AssignExprReturn s] :
		a=unar_expr[r]					{ $s = $a.s; }
		( '*' b=unar_expr[r]				{ $s.i =$s.i.multiply($b.s.i);
								  $s.sec_dependent = $s.sec_dependent || $b.s.sec_dependent;
								}
		| '/' b=unar_expr[r]				{ $s.i = $s.i.divide($b.s.i); 
								  $s.sec_dependent = $s.sec_dependent || $b.s.sec_dependent;
								}
		| '\%' b=unar_expr[r]				{ int comp = $s.i.compareTo($b.s.i);  // only work if results are non negative
								  if (comp >= -1)
								     $s.i = new BigIntegerInterval(BigInteger.ZERO, $s.i.getMax().min($b.s.i.getMax()));
								  $s.sec_dependent = $s.sec_dependent || $b.s.sec_dependent;
								}
		)*
		;

unar_expr [Ranges r] returns [AssignExprReturn s] :
		term[r]						{ $s = $term.s; }
		| '!' term[r]					{ $s = new AssignExprReturn();
								  $s.sec_dependent = $term.s.sec_dependent;
								  if ($term.s.i.isStriclyPositive())
								    $s.i = new BigIntegerInterval(BigInteger.ZERO,BigInteger.ZERO);
								  else if ($term.s.i.isZero())
								    $s.i = new BigIntegerInterval(BigInteger.ONE,BigInteger.ONE);
  								  else
								    $s.i = new BigIntegerInterval(BigInteger.ZERO,BigInteger.ONE);
								}
		| '-' term[r]					{ $s = new AssignExprReturn();
								  $s.sec_dependent = $term.s.sec_dependent;
								  $s.i = $term.s.i.negate(); 
								}
		;

term [Ranges r] returns [AssignExprReturn s] :
		'(' assign_expr[r] ')'				{ $s = $assign_expr.s; }
		| leaf_term[r]					{ $s = $leaf_term.s; }
		;

leaf_term [Ranges r] returns [AssignExprReturn s] :
		NUMBER						{ BigInteger val = new BigInteger($NUMBER.text);
								  $s = new AssignExprReturn();
								  $s.i = new BigIntegerInterval(val,val);
								  $s.sec_dependent = false;
								}
		| variable[r]					{ $s = new AssignExprReturn();
								  $s.sec_dependent = false;
								  BigInteger min = null;
								  BigInteger max = BigInteger.ZERO;
								  for (Variable v : $variable.lvar) {
								      BigIntegerInterval vrange = r.getVariableRange(v);
								      if (vrange == null) {
								          logger.log("ERROR: variable " + v.getName() + " is used before being initialised.",1);
								          System.exit(1);
								      }
								      if (min == null || vrange.getMin().compareTo(min) < 0)
									  min = vrange.getMin();
								      if (vrange.getMax().compareTo(max) > 0)
									  max = vrange.getMax();
								      if (v.getPublicity() == -2) // secret
									  $s.sec_dependent = true;
								  }
								  $s.i = new BigIntegerInterval(min,max);
								}
		;
		
variable [Ranges r] returns [List<Variable> lvar] @init { $lvar = new LinkedList<Variable>(); } :
    IDENTIFIER 	indices[r]				{ Variable v = prog.getVariable($IDENTIFIER.text + $indices.text); 
							  if (v != null)
							      $lvar.add(v);
							  else {
							      v = prog.getVariable($IDENTIFIER.text);
							      if (v == null) {
								  logger.log("ERROR: on line " + $IDENTIFIER.getLine() + ", unknown variable " + $IDENTIFIER.text + ".",1);
								  System.exit(1);
							      }
							      List<String> string_indices = v.generateIndices();
							      for (String s : string_indices) {
								  v = prog.getVariable($IDENTIFIER.text + s);
								  if (v != null)
								      $lvar.add(v);
								  else {
								      logger.log("ERROR: on line " + $IDENTIFIER.getLine() + ", unknown variable " + $IDENTIFIER.text + s + ".",1);
								      System.exit(1);
								  }
							      }
							  }     
							}
    ;

indices [Ranges r]:	('[' assign_expr[r] ']')* ;
    
float_number returns [BigDecimal f]:
    n1=NUMBER '.' n2=NUMBER 		{ $f = new BigDecimal($n1.text + "." + $n2.text); }
    ;
    
filename:	('.' | IDENTIFIER | NUMBER)+ ; 

    	 
INT :		'int'('a'..'z'|'A'..'Z'|'0'..'9'|'_')+ ;
IDENTIFIER  :   ('a'..'z'|'A'..'Z'|'_')('a'..'z'|'A'..'Z'|'0'..'9'|'_')* ;
NUMBER :	'0'..'9'+ ;
//FILENAME :      ('a'..'z'|'A'..'Z'|'0'..'9'|'_'|'.')+ ;
// FLOAT :		'0''.'('0'..'9')+ ;
NEWLINE :	'\n' { skip(); } ;
SPACES :	(' '|'\t'|'\r')+ { skip(); } ;
COMMENT : 	'//' ~('\n'|'\r')* '\r'? '\n' { $channel=HIDDEN; } ;
