/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/

package preprocessor;

import java.io.PrintWriter;
import java.math.BigDecimal;
import org.antlr.runtime.*;
import support.Logger;

/** - Compute ranges with the grammar, from top to bottom:
      Returns a list statements and a list of ranges. Each statement is associated to a range. 
      The range contains for each variable its range and the total range of internal and observable variables.
      It also contains a Boolean to mark if it depends on secret variables.
 *  - Update the ranges from bottom to top:
      It computes for each variables the total range used in the program and it updates the range of this variable before it has been initialized.
 */
public class Preprocessor3 {
    
    public static P3ListStatement statements;

    public static void preprocess(String filename) throws Exception {
	Logger.getInstance().log("==== Preprocessor 3 ====",3);
      
    	Programme prog = new Programme(null);
        Preprocessor3Lexer lexer = new Preprocessor3Lexer(new ANTLRFileStream(filename));
        CommonTokenStream tokens = new CommonTokenStream(lexer);
        Preprocessor3Parser parser = new Preprocessor3Parser(tokens);
        parser.prog_grammar(prog);
        statements = parser.statements;
        
        // 1 compute the total ranges of the variables over all the program 
	Ranges global_ranges = statements.getGlobalRanges(new Ranges());
	
	// 2 update the ranges in which some variables have not been initialized
	statements.updateRanges(global_ranges);
        
        // 3 update the secret dependency from bottom to top 
        statements.updateSecDep(false);
        
        // 4 print the statements and ranges
        String print = statements.print();
        try ( PrintWriter out = new PrintWriter(filename.replace(".2.pp",".2.values")) ) {
	    prog.print(out);		// print declaration
	    out.println(print);		// print statements + ranges
	}
	
	// 5 generate code with simulate statements
        String code = statements.generateCode(false).getLeft();
	try(  PrintWriter out = new PrintWriter(filename.replace(".2",".3")) ) {
	    prog.print(out);		// print declaration
   	    out.println(code);		// print code
	}
    }
    
    public static void main(String[] args) throws Exception {
        preprocess(args[0]);
    }    
}
