/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/

package preprocessor;

import java.io.PrintWriter;
import java.math.BigInteger;
import java.util.Iterator;
import java.util.LinkedList;
import support.Logger;

public class PublicVariable extends Variable {
    private static final Logger logger = Logger.getInstance();

    private boolean publicity;  // true for PUBLIC, false for OBSERVABLE
    protected LinkedList<BigInteger> initVal;
    
    public PublicVariable(int t) {
        intType = t;
        minInt = BigInteger.ZERO;
        maxInt = new BigInteger("2").pow(t).subtract(BigInteger.ONE);
        arrays = new LinkedList<ArrayDimension>();
        publicity = true;
        initVal = new LinkedList<BigInteger>();
        lastUsed = null;
    }
        
    public PublicVariable(int t, BigInteger v) {
        this(t);
        initVal.add(v);
    }
    
    public void setPublicity(boolean b) {
        publicity = b;
    }
    
    // 1 for public
    // 2 for observable
    // -1 for private
    // -2 for secret
    public int getPublicity() {
        if (publicity) {
            return 1;
        }
        else {
            return 2;
        }
    }
            
    public int updateInitialValues(LinkedList<BigInteger> l) {
        int returnCode = 0;
        initVal = l;
        
        if (l.size() == 1 && getNbValues() > 1) {
	    // copy the value
	    for (int i = 1; i < getNbValues(); i++)
		initVal.add(initVal.getFirst());
	}
        else if (l.size() < getNbValues()) {
            logger.log("WARNING: array not entirely initialized,",1);
            returnCode = 1;
        }
        else if (l.size() > getNbValues()) {
            logger.log("WARNING: two many initial values provided,",1);
            returnCode = 1;
            while (l.size() > getNbValues()) {
                l.removeLast();
            } 
        }
        
        for (Iterator<BigInteger> it = l.iterator(); it.hasNext();) {
            BigInteger val = it.next();
            if (val.compareTo(minInt) < 0 || val.compareTo(maxInt) > 0) {
                logger.log("ERROR: initial value " + val.toString() + " out of range,",1);
                return 2;    
            }
        }
        return returnCode;    
    }
       
    public int print(PrintWriter out) {
        String decl = "";
        if (publicity) {
            decl = "public";
        }
        else {
            decl = "observable";
        }
        decl = decl.concat(" int"+intType+" "+name);
        
        int line=0;
        if (getDimensions() > 0) {
            Iterator<BigInteger> ar = initVal.iterator();
            for (Iterator<String> it = generateIndices().iterator(); it.hasNext();) {
                String s = it.next();
                String declit = decl.concat(s);

                if (ar.hasNext()) {
                    BigInteger val = ar.next();
                    declit = declit.concat(":=" + val.toString());
                }
                declit = declit.concat(";");
                out.println(declit);
                line++;
            }
        }
        else {
            if (!initVal.isEmpty()) {
                decl = decl.concat(":=" + initVal.getFirst().toString());
            }
            decl = decl.concat(";");
            out.println(decl);
            line++;
        }
        return line;
    }
}
