/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/

package preprocessor;

import java.math.BigDecimal;
import java.util.HashSet;

public class RandomStatement extends Statement {
    private String var;
    private BigDecimal proba;
    private String expr;
    private boolean type; // true  for random, false for randombit
    
    public RandomStatement() {
    	usedVariables = new HashSet<Variable>();
    	nbStatements = 1;
    }
    
    public RandomStatement(String v, String e) {
    	usedVariables = new HashSet<Variable>();
    	nbStatements = 1;
        var = v;
        expr = e;
        type = true;
    }
    
    public RandomStatement(String v, BigDecimal f) {
    	usedVariables = new HashSet<Variable>();
    	nbStatements = 1;
        var = v;
        proba = f;
        expr = "";
        type = false;
    }
    
    public void setVariable(String v) {
    	var = v;
    }
    
    public void setExpression(String e) {
    	expr = e;
    	type = true;
    }
    
    public void setProbability(BigDecimal f) {
    	proba = f;
        expr = "";
        type = false;
    }
   
    public Statement replace(String oldv, String newv) {
        String newVar = var.replace("("+oldv+")","("+newv+")");
        newVar = newVar.replace("("+oldv+"[","("+newv+"[");
        String newExpr = expr.replace("("+oldv+")","("+newv+")");
        newExpr = newExpr.replace("("+oldv+"[","("+newv+"[");
          
        RandomStatement ret;
        if (type) {
            ret = new RandomStatement(newVar,newExpr);
        }
        else {
            ret = new RandomStatement(newVar,proba);
        }
        ret.addUsedVariable(getUsedVariables());
        for (Variable v : ret.getUsedVariables()) {
        	v.setLastUsed(ret);
        }
        return ret;
    }

    public String toString(int line) {
        if (type) {
            return "random " + var + ":=random(" + expr + ");\n";
        }
        else {
            return "random " + var + ":=randombit(" + proba.toPlainString() + ");\n";
        }
    }
}
