/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
*/
    
package support;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.HashMap;

/**
 *
 * @author  
 */
public class Logger {
    
    private static Logger instance;
    private static int verbosity;
    private static int defaultVerbosity;
    private HashMap<String,Integer> exceptionMap;
    
    private Logger(){
        verbosity = 2;
        defaultVerbosity = 2;
        exceptionMap = null;
    }
    
    public static Logger getInstance(){
        if (instance == null){
            instance = new Logger();
        }
        return instance;
    }
    
    public void setVerbosity(int verb){
        verbosity = verb;
    }
    
    public int getVerbosity(){
        return verbosity;
    }

    public void setDefaultVerbosity(int defaultVerbosity) {
        Logger.defaultVerbosity = defaultVerbosity;
    }

    public void setExceptionMap(HashMap<String, Integer> exceptionMap) {
        this.exceptionMap = exceptionMap;
    }
    
    public void changeVerbosity(String name){
        if (exceptionMap == null){
            verbosity = defaultVerbosity;
            return;
        }
        verbosity = exceptionMap.getOrDefault(name, defaultVerbosity);
    }

    public void restoreVerbosity(){
        verbosity = defaultVerbosity;
    }

    public void log(String message, int level){
        if (level<=verbosity){
            System.out.println(message);
        }
    }
    
    // print arrays
    public void log(Object[] array, String name, int level) {
        if (array == null) {
            log("CAUTION(Logger62): the given array" + name + " is null.", level);
        } else {
            for (int i = 0; i < array.length; i++) {
                if (array[i] == null) {
                    log("  " + name + "[" + i + "] = null.", level);
                } else {
                    log("  " + name + "[" + i + "] = " + array[i] + ".", level);
                }
            }
        }
    }

    public void log(BigInteger[] array, String name, int level) {
        if (array == null) {
            log("CAUTION(Logger76): the given array" + name + " is null.", level);
        } else {
            for (int i = 0; i < array.length; i++) {
                if (array[i] == null) {
                    log("  " + name + "[" + i + "] = null.", level);
                } else {
                    log("  " + name + "[" + i + "] = " + array[i] + ".", level);
                }
            }
        }
    }

    public void log(BigDecimal[] array, String name, int level) {
        if (array == null) {
            log("CAUTION(Logger90): the given array" + name + " is null.", level);
        } else {
            for (int i = 0; i < array.length; i++) {
                if (array[i] == null) {
                    log("  " + name + "[" + i + "] = null.", level);
                } else {
                    log("  " + name + "[" + i + "] = " + array[i] + ".", level);
                }
            }
        }
    }
}
