package support;

import java.io.*;
import java.util.*;
import java.security.SecureRandom;
import distributions.JointFrequencyDistribution;
import bham.leakiest.*;

/**
 *
 * @author  
 */
public class SimulatorOfConditionalDistribution {
    private static final Logger logger = Logger.getInstance();
    
	// The number of experiments (the number of output files)
	private static int numFiles = 1000; // Change this for tests
	
	// Parameters for options
	private static final int loop = 1;
	private static final boolean outputfile = false;
	private static final boolean printFrequency = false;
 
	private static double sumLeak = 0;
	private static PrintWriter pw = null;
	
	/*
	 * TODO: To be completed
	 * @param args
	 */
	/*
	public static void main(String[] args) {
		int obsTotal = 1000;
		String inputDir5      = "./examples/estimate-MEL/chi-square/";
		String inputFileName  = inputDir5 + "matrix4-3d.txt";
		String outputFileNameBase = inputDir5 + "obs-cond4-3d/matrix4-3d" + "-" + obsTotal + "-";

	    // In case we have a random number of observations for each input
        log("// Executing SimulatorOfConditionalDistribution.class",1);

        // Commandline option
		if(args.length > 1) {
			inputFileName = args[0];
			outputFileNameBase = args[1] + "-" + obsTotal + "-";
		} else {
			//log("Put a file name as an argument of this program.");
			//System.exit(1);
		}
		
     // Read a channel file
     ReadFile channelFileReader = new ReadFile(inputFileName, logger.getVerbosity(), "");
     channelFileReader.readChannel();

     // Calculate a channel matrix from the given file
     Channel channel = channelFileReader.getChannel();
     double[][] matrix = channel.getMatrix();
		
     // Calculate a uniform input distribution
     String[] inputNames = channel.getInputNames();
     ProbDist prior = ProbDist.uniformProbDist(inputNames, true);

		// Output to a file
		int numExp = 0;
		String outputFileName = outputFileNameBase + String.format("%04d", numExp) + ".txt";				
		try {
			File outfile = new File(outputFileName);
			pw = new PrintWriter(new BufferedWriter(new FileWriter(outfile)));
		    log("Number of files = " + numFiles,4);
			JointFrequencyDistribution JFD = simulateConditionalDistribution(obsTotal, prior, channel);
		} catch(FileNotFoundException ex) {
			log(" file not found " + ex,1);
		} catch(Exception ex) {
			log(" error " + ex,1);
		}
	}
	*/

	/**
	 * Computes the cumulative distribution of the channel matrix.
	 * 
	 * @param channel channel
	 * @return the cumulative distribution of the channel matrix (conditional distribution)
	 */
	public static Channel computeCumulativeConditionalDistribution(Channel channel) {
		double[][] matrix = channel.getMatrix();
		int noOfInputs = matrix.length;
		int noOfOutputs = matrix[0].length;
		double[][] CDF = new double[noOfInputs][noOfOutputs];
		for(int numRow = 0; numRow < noOfInputs; numRow++) {
			// when numCol = 0
			CDF[numRow][0] = matrix[numRow][0];
			// when numCol > 0
			for(int numCol = 1; numCol < noOfOutputs; numCol++) {
				CDF[numRow][numCol] = CDF[numRow][numCol-1] + matrix[numRow][numCol];
				//log(" numRow = " + numRow + ", numCol = " + numCol);
				//log(" PDF[" + numRow + "," + numCol + "] = " + matrix[numRow][numCol]);
				log(" CDF[" + numRow + "," + numCol + "] = " + CDF[numRow][numCol] + "  " + channel.getInputNames()[numRow] + ", " + channel.getOutputNames()[numCol],4);
			}
		}
		Channel CDFchannel = new Channel(channel.getInputNames(), channel.getOutputNames(), CDF);
		//print
		if(logger.getVerbosity() >= 4) {
			log(" noOfInputs  = " + noOfInputs,4);
			log(" noOfOutputs = " + noOfOutputs,4);
			log(" output 0 = " + channel.getOutputNames()[0],4);
			log(" output 1 = " + channel.getOutputNames()[1],4);
			//log(" output 2 = " + channel.getOutputNames()[2],4);
			//log(" output 3 = " + channel.getOutputNames()[3],4);
			log("",4);
			channel.printChannel();
			log("CDF",4);
			CDFchannel.printChannel();
		}
		return CDFchannel;
	}

    /**
     * Randomly execute the conditional probability distribution to obtain a
     * conditional frequency distribution.
     *
     * @param obsTotal total sample size
     * @param prior prior
     * @param channel channel
     * @return conditional frequency distribution
     */
    public static JointFrequencyDistribution simulateConditionalDistribution(int obsTotal, ProbDist prior, Channel channel) {
        // Get the time of starting the program
        Date dateStart = new Date();

		// channel matrix
		final double[][] matrix = channel.getMatrix();

		// prior
		final double[] pmf = prior.probDistToPMFArray(channel.getInputNames());
		
		// Calculate a cumulative (conditional) distribution function
		final int noOfInputs = matrix.length;
		final int noOfOutputs = matrix[0].length;
		final Channel CDFchannel = computeCumulativeConditionalDistribution(channel);
		final double[][] CDF = CDFchannel.getMatrix();

		log("Number of observations = " + obsTotal,4);
		
		// Create a new freqency matrix by running the channel
		Observations obs = new Observations();

		for(int numRow = 0; numRow < noOfInputs; numRow++) {
			// Number of observations for the input
			int obsPerInput = (int)((double)obsTotal * pmf[numRow]);
			// Generate observation counts for the (numRow)-th row
			for(int numObs = 0; numObs < obsPerInput; numObs++) {
				// Generate a random seed
				SecureRandom random = new SecureRandom();

				// Generate a random observation following the channel matrix
				double ran = random.nextDouble();
				for(int numCol = 0; numCol < noOfOutputs; numCol++) {
					//log(" ran " + ran + "   CDF[" + numRow + ", " + numCol + "] =  " + CDF[numRow][numCol]);
					if(ran < CDF[numRow][numCol]) {
						if(outputfile) {
							pw.println("(" + channel.getInputNames()[numRow] + ", " + channel.getOutputNames()[numCol] + ")");
						}
						obs.addObservation(channel.getInputNames()[numRow], channel.getOutputNames()[numCol]);
						//log(" ran " + ran + " < CDF[" + numRow + ", " + numCol + "] =  " + CDF[numRow][numCol]);
						break;
					}
				}
			}
		}
		if(outputfile) {
			pw.close();
		}
  		JointFrequencyDistribution JFD = ConvertLeakiEstObjects.computeJointFrequencyDistribution(obs);

		// Get the time of terminating the program
		Date dateEnd = new Date();
		log("Time: " + (dateEnd.getTime() - dateStart.getTime()) + " [msec]",3);
		if(outputfile) {
			log("Mean: " + (sumLeak / loop),3);
		}
		
		if(printFrequency) {
			obs.printJointFrequencyMatrix();
		}
		
		return JFD;
	}
	
    /////////////////////////////////////////////////////////////
    // log
    private static void log(Object aObject, int level) {
        if (logger.getVerbosity() >= level) {
 	        Logger loggy = Logger.getInstance();
 	       loggy.log(String.valueOf(aObject), level);
        }
    }
}
