package support;

import bham.leakiest.*;
import java.io.*;
import java.util.*;
import java.security.SecureRandom;
import distributions.JointFrequencyDistribution;

/**
 *
 * @author  
 */
public class SimulatorOfJointDistribution {
    private static final Logger logger = Logger.getInstance();

	// The number of experiments (the number of output files)
	private static int numFiles = 1000; // Change this for tests
	
	// Parameters for options
	private static final boolean outputfile = false;
	private static final boolean printFrequency = false;
	private static final int loop = 1;

	private static double sumLeak = 0;
	private static PrintWriter pw = null;
	
	/*
	 * TODO: To be completed
	 * @param args
	 */
	/*
	public static void main(String[] args) {
		String inputDir5      = "./examples/estimate-MEL/chi-square/";
		String inputFileName  = inputDir5 + "matrix4-3d.txt";
		String outputFileNameBase = inputDir5 + "obs-cond4-3d/matrix4-3d" + "-" + obsTotal + "-";

        log("// Executing GenerateObsFromCondProb.class");

        // Commandline option
		if(args.length > 1) {
			inputFileName = args[0];
			outputFileNameBase = args[1] + "-" + obsTotal + "-";
		} else {
			//log("Put a file name as an argument of this program.");
			//System.exit(1);
		}
		
		// Read a channel file
		ReadFile channelFileReader = new ReadFile(inputFileName, logger.getVerbosity(), "");
		channelFileReader.readChannel();

		// Calculate a channel matrix from the given file
		Channel channel = channelFileReader.getChannel();
		double[][] matrix = channel.getMatrix();
		
		// Calculate a uniform input distribution
		String[] inputNames = channel.getInputNames();
		ProbDist prior = ProbDist.uniformProbDist(inputNames, true);

		// Output to a file
		int numExp = 0;
		String outputFileName = outputFileNameBase + String.format("%04d", numExp) + ".txt";				
		try {
			File outfile = new File(outputFileName);
			pw = new PrintWriter(new BufferedWriter(new FileWriter(outfile)));
		    log("Number of files = " + numFiles,4);
			JointFrequencyDistribution JFD = simulateJointDistribution(prior, channel);
		} catch(FileNotFoundException ex) {
			log(" file not found " + ex);
		} catch(Exception ex) {
			log(" error " + ex);
		}
	}
	*/

	/**
	 * Computes the cumulative distribution of the joint distribution.
	 * 
	 * @param prior prior
	 * @param channel channel
	 * @return the cumulative distribution of the channel matrix (joint distribution)
	 */
	public static Channel computeCumulativeJointDistribution(ProbDist prior, Channel channel) {
		Channel jointDist = channel.getJointDist(prior);
		return computeCumulativeJointDistribution(jointDist);
	}
	
	/**
	 * Computes the cumulative distribution of the joint distribution.
	 * 
	 * @param jointDist joint distribution
	 * @return the cumulative distribution of the channel matrix (joint distribution)
	 */
	public static Channel computeCumulativeJointDistribution(Channel jointDist) {
		double[][] jointMatrix = jointDist.getMatrix();
		int noOfInputs = jointMatrix.length;
		int noOfOutputs = jointMatrix[0].length;
		double[][] CDF = new double[noOfInputs][noOfOutputs];
		
		for(int numRow = 0; numRow < noOfInputs; numRow++) {
			// when numCol = 0
			if(numRow == 0) {
				CDF[numRow][0] = jointMatrix[numRow][0];
			} else {
				CDF[numRow][0] = CDF[numRow-1][noOfOutputs-1] + jointMatrix[numRow][0];
				log(" numRow = " + (numRow - 1) + ", numCol = " + (noOfOutputs-1),4);
			}
			// when numCol > 0
			for(int numCol = 1; numCol < noOfOutputs; numCol++) {
				CDF[numRow][numCol] = CDF[numRow][numCol-1] + jointMatrix[numRow][numCol];
				//log(" numRow = " + numRow + ", numCol = " + numCol,4);
				if(logger.getVerbosity() >= 4) {
					//log(" PDF[" + numRow + "," + numCol + "] = " + jointMatrix[numRow][numCol],4);
					log(" CDF[" + numRow + "," + numCol + "] = " + CDF[numRow][numCol] + "  " + jointDist.getInputNames()[numRow] + ", " + jointDist.getOutputNames()[numCol],4);
				}
			}
		}
		
		Channel CDFjoint = new Channel(jointDist.getInputNames(), jointDist.getOutputNames(), CDF);
		//print
		if(logger.getVerbosity() >= 4) {
			log(" noOfInputs  = " + noOfInputs,4);
			log(" noOfOutputs = " + noOfOutputs,4);
			log(" output 0 = " + jointDist.getOutputNames()[0],4);
			log(" output 1 = " + jointDist.getOutputNames()[1],4);
			//log(" output 2 = " + jointDist.getOutputNames()[2],4);
			//log(" output 3 = " + jointDist.getOutputNames()[3],4);
			log("Joint",4);
			jointDist.printChannel();
			log("CDF",4);
			CDFjoint.printChannel();
		}
		return CDFjoint;
	}

	/**
	 * Randomly execute the joint probability distribution
	 * to obtain a joint frequency distribution.
	 * 
	 * @param obsTotal total sample size
	 * @param prior prior
	 * @param channel channel
	 * @return joint frequency distribution
	 */
	public static JointFrequencyDistribution simulateJointDistribution(int obsTotal, ProbDist prior, Channel channel) {
		Channel jointDist = channel.getJointDist(prior);
		return simulateJointDistribution(obsTotal, jointDist);
	}
		
	/**
	 * Randomly execute the joint probability distribution
	 * to obtain a joint frequency distribution.
	 * 
	 * @param obsTotal total sample size
	 * @param jointDist joint distribution
	 * @return joint frequency distribution
	 */
	public static JointFrequencyDistribution simulateJointDistribution(int obsTotal, Channel jointDist) {
		// Get the time of starting the program
		Date dateStart = new Date();

		// Calculate a cumulative (joint) distribution function
		final Channel CDFjoint = computeCumulativeJointDistribution(jointDist);
		final double[][] CDFjointMatrix = CDFjoint.getMatrix();
		final int noOfInputs = CDFjointMatrix.length;
		final int noOfOutputs = CDFjointMatrix[0].length;

		log("Number of observations = " + obsTotal,4);
		
		// Create a new freqency matrix by running the joint
		Observations obs = new Observations();

		// Generate observation counts for the (numRow)-th row
		for(int numObs = 0; numObs < obsTotal; numObs++) {
			// Generate a random seed
			SecureRandom random = new SecureRandom();
			// Generate a random observation following the joint matrix
			double ran = random.nextDouble();
			for(int numRow = 0; numRow < noOfInputs; numRow++) {
				for(int numCol = 0; numCol < noOfOutputs; numCol++) {
					//log(" ran " + ran + "   CDF[" + numRow + ", " + numCol + "] =  " + CDF[numRow][numCol]);
					if(ran < CDFjointMatrix[numRow][numCol]) {
						if(outputfile) {
							pw.println("(" + jointDist.getInputNames()[numRow] + ", " + jointDist.getOutputNames()[numCol] + ")");
						}
						obs.addObservation(jointDist.getInputNames()[numRow], jointDist.getOutputNames()[numCol]);
						//log(" ran " + ran + " < CDF[" + numRow + ", " + numCol + "] =  " + CDF[numRow][numCol]);
						break;
					}
				}
				if(ran < CDFjointMatrix[numRow][noOfOutputs-1]) {
					//System.out.println(" ran " + ran + " < CDF[" + numRow + ", " + (noOfOutputs-1) + "] =  " + CDF[numRow][noOfOutputs-1]);
					break;
				}
			}
		}
		if(outputfile) {
			pw.close();
		}
  		JointFrequencyDistribution JFD = ConvertLeakiEstObjects.computeJointFrequencyDistribution(obs);

		// Get the time of terminating the program
		Date dateEnd = new Date();
		log("Time: " + (dateEnd.getTime() - dateStart.getTime()) + " [msec]",3);
		log("Mean: " + (sumLeak / loop),3);
		
		if(printFrequency) {
			obs.printJointFrequencyMatrix();
		}
		
		return JFD;
	}
	
    /////////////////////////////////////////////////////////////
    // log
    private static void log(Object aObject, int level){
        if (logger.getVerbosity() >= level) {
            Logger loggy = Logger.getInstance();
            loggy.log(String.valueOf(aObject),level);
        }
    }
}
