/*  HyLeak is a tool to evaluate the security of systems through a quantitative analysis of information leakage.
    Copyright (C) 2014-2017 - 

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
    
    Author: 
 */
package support;

import java.util.ArrayList;
import java.util.HashSet;

/**
 *
 * @author  
 */
public class Tarjan {

    private int index = 0;
    private final ArrayList<Node> stack = new ArrayList<>();
    private final HashSet<HashSet<String>> SCC = new HashSet<>();
    private NodeList nodes = new NodeList();
    private EdgeList edges = new EdgeList();

    public Tarjan (NodeList n, EdgeList e){
        nodes = n;
        edges = e;
    }
    
    public NodeList getNodes() {
        return this.nodes;
    }

    public EdgeList getEdges() {
        return this.edges;
    }

//   
    public HashSet<HashSet<String>> getSCC() {
        SCC.clear();
        index = 0;
        stack.clear();
        if (edges != null) {
            ArrayList<Node> nodeList = new ArrayList<>();
            nodeList.addAll(nodes);

            for (Node node : nodeList) {
                log("Selected node " + node.printAll(), 4);
                if (node.index == -1) {
                    log("Node " + node + " does not have an index; calling Tarjan's algorithm for node " + node.printAll() + " with edges " + edges.getEdgesFrom(node), 4);
                    tarjan(node);
                    log("Now it's node " + node.printAll(), 4);
                } else {
                    log("Already visited; moving on", 4);
                }
            }

        }
        return SCC;
    }

    private HashSet<HashSet<String>> tarjan(Node v) {
        v.index = index;
        v.lowlink = index;
        log("I gave the index " + index + " to node " + v, 4);
        index++;
        stack.add(0, v);
        EdgeList edgesfromv = edges.getEdgesFrom(v);
        log("The edges from " + v + " are " + edgesfromv, 4);
        for (Edge e : edgesfromv) {
            Node n = nodes.getNodeByName(e.to);
            if (n.index == -1) {
                log("Node " + n.printAll() + " successor of " + v.printAll() + " has no index, calling Tarjan's algorithm recursively for it", 4);
                //log("Here list has "+list.size()+" elements",4);
                tarjan(n);
                v.lowlink = Math.min(v.lowlink, n.lowlink);

            } else if (stack.contains(n)) {
                v.lowlink = Math.min(v.lowlink, n.index);
            }
            log("Now node " + v + " is " + v.printAll(), 4);
        }
        if (v.lowlink == v.index) {
            log("Node " + v + " is the root of a SCC!", 4);
            Node n;
            HashSet<String> component = new HashSet<>();
            do {
                n = stack.remove(0);
                component.add(n.getElement());
                //log("Adding "+n+" to the SCC of "+v);
            } while (n != v);
            SCC.add(component);
        }
        return SCC;
    }

    private static void log(Object aObject, int level) {
        Logger logger = Logger.getInstance();
        logger.log(String.valueOf(aObject), level);
    }
}
