/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package test;

import distributions.JointDistribution;
import distributions.JointFrequencyDistribution;
import distributions.SingleDistribution;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.math.MathContext;
import java.math.RoundingMode;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import hyleak.*;
import support.Contexter;
import support.Logger;

/**
 *
 * @author  
 */
public class testDecisionTree {

    private static final Contexter contexter = Contexter.getInstance();
    private static final Logger logger = Logger.getInstance();

    public static void main(String[] args) {
        logger.setVerbosity(2);
        contexter.setContext(new MathContext(5, RoundingMode.HALF_EVEN));

        int iterations = 1000;
        int samples = 1000;
        int pace = 10;

        BigDecimal[][] fixedResults = runFixedComputation(samples);
        BigDecimal[][] adaptiveResults = runAdaptiveComputation(samples);
        for (int i = 1; i < iterations; i++) {
            if (i % pace == 0) {
                log("Running iteration number " + i, 2);
            }
            BigDecimal[][] tempFixedResults = runFixedComputation(samples);
            BigDecimal[][] tempAdaptiveResults = runAdaptiveComputation(samples);
            for (int j = 0; j < 4; j++) {
                for (int k = 0; k < 4; k++) {
                    fixedResults[j][k] = fixedResults[j][k].add(tempFixedResults[j][k], contexter.getContext());
                    adaptiveResults[j][k] = adaptiveResults[j][k].add(tempAdaptiveResults[j][k], contexter.getContext());
                }
            }
        }
        log("Computation terminated", 2);
        for (int j = 0; j < 4; j++) {
            for (int k = 0; k < 4; k++) {
                fixedResults[j][k] = fixedResults[j][k].divide(new BigDecimal(iterations), contexter.getContext());
                adaptiveResults[j][k] = adaptiveResults[j][k].divide(new BigDecimal(iterations), contexter.getContext());
            }
        }
        String[] attributes = new String[]{"A", "B", "C", "D"};
        log("Results after " + iterations + " iterations of " + samples + " samples each:", 2);
        log("FIXED:", 2);
        for (int j = 0; j < 4; j++) {
            log("" + attributes[j], 2);
            log("Mutual information: " + fixedResults[j][0], 2);
            log("Lower bound: " + fixedResults[j][1], 2);
            log("Upper bound: " + fixedResults[j][2], 2);
            log("Difference: " + fixedResults[j][3], 2);
        }
        log("\n\nADAPTIVE:", 2);
        for (int j = 0; j < 4; j++) {
            log(attributes[j], 2);
            log("Mutual information: " + adaptiveResults[j][0], 2);
            log("Lower bound: " + adaptiveResults[j][1], 2);
            log("Upper bound: " + adaptiveResults[j][2], 2);
            log("Difference: " + adaptiveResults[j][3], 2);
        }
    }

    public static BigDecimal[][] runFixedComputation(int totSims) {
        String[] attributes = new String[]{"A", "B", "C", "D"};
        BigDecimal[][] toreturn = new BigDecimal[4][4];
        int count = 0;
        for (String attr : attributes) {

            int rounds = 10;
            SingleDistribution prior = null;
            JointPosteriorDistributionManager postDistMan = new JointPosteriorDistributionManager(prior, BigInteger.valueOf(totSims));

            HashMap<Signature, BigDecimal> mapExecProbs = new HashMap();

            MatrixSignature sig1 = new MatrixSignature("0.1", "dectree/db1" + attr + ".txt");

            mapExecProbs.put(sig1, sig1.getWeight());

            MatrixSignature sig2 = new MatrixSignature("0.3", "dectree/db2" + attr + ".txt");

            mapExecProbs.put(sig2, sig2.getWeight());

            MatrixSignature sig3 = new MatrixSignature("0.6", "dectree/db3" + attr + ".txt");

            mapExecProbs.put(sig3, sig3.getWeight());

            Map<Signature, BigInteger> simNumber = postDistMan.computeSampleSizesInitially(mapExecProbs, rounds);
            postDistMan.storeExactSubDistToManager(new JointDistribution());
            int sims = totSims / 3;
            JointFrequencyDistribution freqDist = support.SimulatorOfJointDistribution.simulateJointDistribution(sims, sig1.getChannel());
            postDistMan.storeFreqDistFromMatrixToManager(sig1, freqDist, sims, sig1.getWeight());

            //sims = totSims / 10 * 3;
            sims = totSims / 3;
            freqDist = support.SimulatorOfJointDistribution.simulateJointDistribution(sims, sig2.getChannel());
            postDistMan.storeFreqDistFromMatrixToManager(sig2, freqDist, sims, sig2.getWeight());

            sims = totSims / 3;
            freqDist = support.SimulatorOfJointDistribution.simulateJointDistribution(sims, sig3.getChannel());
            postDistMan.storeFreqDistFromMatrixToManager(sig3, freqDist, sims, sig3.getWeight());
            BigDecimal estimate = postDistMan.computeMIPointEstimate();
            BigDecimal[] confint = postDistMan.computeConfidenceIntervalOfMI();
            BigDecimal difference = confint[1].subtract(confint[0], contexter.getContext());
            log("finished fixed for " + attr, 3);
            toreturn[count][0] = estimate;
            toreturn[count][1] = confint[0];
            toreturn[count][2] = confint[1];
            toreturn[count][3] = difference;
            count++;
//            log("Mutual information of " + attr + " is " + postDistMan.computeMIPointEstimate().toString(), 1);
//            
//            log("Mutual information confidence interval of " + attr + " is [" + confint[0].toString() + " , " + confint[1].toString() + "]", 1);
        }
        return toreturn;
    }

    public static BigDecimal[][] runAdaptiveComputation(int totSims) {
        String[] attributes = new String[]{"A", "B", "C", "D"};
        BigDecimal[][] toreturn = new BigDecimal[4][4];
        int count = 0;
        for (String attr : attributes) {

            int rounds = 1;
            SingleDistribution prior = null;
            JointPosteriorDistributionManager postDistMan = new JointPosteriorDistributionManager(prior, BigInteger.valueOf(totSims));

            HashSet<MatrixSignature> signatures = new HashSet();
            HashMap<Signature, BigDecimal> mapExecProbs = new HashMap();
            HashMap<Signature,Integer> totSimsPerComp = new HashMap<>();
            
            MatrixSignature sig1 = new MatrixSignature("0.1", "dectree/db1" + attr + ".txt");
            mapExecProbs.put(sig1, sig1.getWeight());
            signatures.add(sig1);
            totSimsPerComp.put(sig1, 0);

            MatrixSignature sig2 = new MatrixSignature("0.3", "dectree/db2" + attr + ".txt");
            mapExecProbs.put(sig2, sig2.getWeight());
            signatures.add(sig2);
            totSimsPerComp.put(sig2, 0);

            MatrixSignature sig3 = new MatrixSignature("0.6", "dectree/db3" + attr + ".txt");
            signatures.add(sig3);
            mapExecProbs.put(sig3, sig3.getWeight());
            totSimsPerComp.put(sig3, 0);
            
            log("started adaptive for " + attr, 3);
            Map<Signature, BigInteger> simNumber = postDistMan.computeSampleSizesInitially(mapExecProbs, rounds);
            // TEST only: overwriting initial simulation numbers
//            for (Signature sig:signatures){
//                simNumber.put(sig, BigInteger.valueOf(totSims/30));
//            }
            
            postDistMan.storeExactSubDistToManager(new JointDistribution());
            for (int i = 0; i < rounds; i++) {
//            log("\n-------------------------------------------", 3);
                log("Starting simulation round " + (i + 1) + " of " + rounds, 3);
                for (MatrixSignature sig : signatures) {
                    int sims = simNumber.get(sig).intValue();
                    totSimsPerComp.put(sig, sims+totSimsPerComp.get(sig));
                    log("\nRunning " + sims + " simulations for this component", 3);
                    if (sims > 0) {
                        JointFrequencyDistribution freqDist = support.SimulatorOfJointDistribution.simulateJointDistribution(sims, sig.getChannel());
                        log("Simulations ran, storing...", 3);
                        postDistMan.storeFreqDistFromMatrixToManager(sig, freqDist, sims, sig.getWeight());
                    }

                }
                if (i < rounds - 1) {
                    int roundsLeft = rounds - i - 1;
                    log("Recomputing sample size...", 3);
                    simNumber = postDistMan.computeSampleSizesAdaptively(roundsLeft);
                    log("Sample size recomputed", 3);
                }
            }
            log("Total simulations per component:",3);
            for (MatrixSignature sig : signatures){
                log("Component with weight "+sig.getWeight()+" got "+totSimsPerComp.get(sig)+" samples",3);
            }

//            int sims = totSims / 10;
//            JointFrequencyDistribution freqDist = support.SimulatorOfJointDistribution.simulateJointDistribution(sims, sig1.getChannel());
//            postDistMan.storeFreqDistFromMatrixToManager(sig1, freqDist, sims, sig1.getWeight());
//
//            sims = totSims / 10 * 3;
//            freqDist = support.SimulatorOfJointDistribution.simulateJointDistribution(sims, sig2.getChannel());
//            postDistMan.storeFreqDistFromMatrixToManager(sig2, freqDist, sims, sig2.getWeight());
//
//            sims = totSims / 10 * 6;
//            freqDist = support.SimulatorOfJointDistribution.simulateJointDistribution(sims, sig3.getChannel());
//            postDistMan.storeFreqDistFromMatrixToManager(sig3, freqDist, sims, sig3.getWeight());
            log("finished adaptive for " + attr, 3);
            
            
            BigDecimal estimate = postDistMan.computeMIPointEstimate();
            BigDecimal[] confint = postDistMan.computeConfidenceIntervalOfMI();
            BigDecimal difference = confint[1].subtract(confint[0], contexter.getContext());
            toreturn[count][0] = estimate;
            toreturn[count][1] = confint[0];
            toreturn[count][2] = confint[1];
            toreturn[count][3] = difference;
            count++;
        }
        return toreturn;
    }

    private static void log(Object aObject, int level) {
        support.Logger.getInstance().log(String.valueOf(aObject), level);
    }

}
