/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package theory;

import java.math.BigDecimal;
import theory.BigFunctions;

/**
*
* @author  
*/
public class Statistics {
	////////////////////////////////////////////////////////////////
	// Chi-square statistics
	/**
	 * &chi;-squared distribution.
	 * chiSquareValues95[i] is the 95% value for the &chi;-square distribution with i+1 degrees of freedom.
	 */
	public static double[] chiSquareValues95 = {3.841,5.991,7.815,9.488,11.07,12.59,14.07,15.51,16.92,18.31,19.68,21.03,22.36,23.69,25.00,26.30,27.59,28.87,30.14,31.41,32.67,33.92,35.17,36.42,37.65,38.89,40.11,41.34,42.56,43.77,44.99,46.19,47.40,48.60,49.80,51.00,52.19,53.38,54.57,55.76,56.94,58.12,59.30,60.48,61.66,62.83,64.00,65.17,66.34,67.51,68.67,69.83,70.99,72.15,73.31,74.47,75.62,76.78,77.93,79.08,80.23,81.38,82.53,83.68,84.82,85.97,87.11,88.25,89.39,90.53,91.67,92.81,93.95,95.08,96.22,97.35,98.49,99.62,100.75,101.88,103.01,104.14,105.27,106.40,107.52,108.65,109.77,110.90,112.02,113.15,114.27,115.39,116.51,117.63,118.75,119.87,120.99,122.11,123.23,124.34}; 
	
	/**
	 * Returns the upper bound for the 95% confidence interval
	 * for the &chi;-squared distribution with n degrees of freedom.
	 * @param freedom freedom
	 * @return the upper bound for the 95% confidence interval
	 * @throws IllegalArgumentException If freedom &le; 0
	 */
	public static double chiSqu95Interval(int freedom) {
		if (freedom <= 0) {
			throw new IllegalArgumentException("Degree of freedom must be an integer greater than 0.");
		} else if (freedom < 101) {
			return chiSquareValues95[freedom-1];
		} else {
			// approximate chi-squared distribution with > 100 degrees of freedom as normal
			// chi ~ N(freedom,2freesom)
			return (1.644854*Math.sqrt(2*freedom)+freedom);
		}
	}
	
	////////////////////////////////////////////////////////////////
	// Normal distribution
	/**
	 * Calculates the upper bound for a 95% confidence interval
	 * for a normal distribution with given mean and variance
	 * i.e., the value below which 97.5% of sample take.
	 * 
	 * @param mean mean
	 * @param variance variance
     * @param scale the desired scale of the result
	 * @return the upper bound for a 95% confidence interval
	 */
	public static BigDecimal upperBoundNormal95(BigDecimal mean, BigDecimal variance, int scale) {
		// x = mean+Z.Math.sqrt(variance) 
		//return mean+1.96*Math.sqrt(variance);
		return mean.add( BigDecimal.valueOf(1.959964).multiply( BigFunctions.sqrt(variance, scale) ) );
	}

	/**
	 * Calculates the upper bound for a 95% confidence interval
	 * for a normal distribution with given mean and variance, lower only
	 * i.e., the value below which 95% of sample take.
	 * 
	 * @param mean mean
	 * @param variance variance
     * @param scale the desired scale of the result
	 * @return the upper bound for a (lower) 95% confidence interval
	 */
	public static BigDecimal upperBoundNormal95Upper(BigDecimal mean, BigDecimal variance, int scale) {
		// x = mean+Z.Math.sqrt(variance) 
		//return mean+1.6449*Math.sqrt(variance);
		return mean.add( BigDecimal.valueOf(1.644854).multiply( BigFunctions.sqrt(variance, scale) ) );
	}
	
	/**
	 * Calculates the lower bound for a 95% confidence interval
	 * for a normal distribution with given mean and variance
	 * i.e., the value above which 97.5% of sample take.
	 * 
	 * @param mean mean
	 * @param variance variance
     * @param scale the desired scale of the result
	 * @return the lower bound for a 95% confidence interval
	 */
	public static BigDecimal lowerBoundNormal95(BigDecimal mean, BigDecimal variance, int scale) { 
		// x = mean-Z.Math.sqrt(variance)
		//return mean-1.96*Math.sqrt(variance);
		return mean.subtract( BigDecimal.valueOf(1.959964).multiply( BigFunctions.sqrt(variance, scale) ) );
	}

	/**
	 * Calculates the lower bound for a 95% confidence interval
	 * for a normal distribution with given mean and variance, upper only
	 * i.e., the value above which 95% of sample take.
	 * 
	 * @param mean mean
	 * @param variance variance
     * @param scale the desired scale of the result
	 * @return the lower bound for a (upper) 95% confidence interval
	 */
	public static BigDecimal lowerBoundNormal95Lower(BigDecimal mean, BigDecimal variance, int scale) { 
		// x = mean-Z.Math.sqrt(variance)
		//return mean-1.6449*Math.sqrt(variance);
		return mean.subtract( BigDecimal.valueOf(1.644854).multiply( BigFunctions.sqrt(variance, scale) ) );
	}
}
